package net.drewke.tdme.engine;

import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.math.Vector3D;

/**
 * Camera
 * @author Andreas Drewke
 * @version $Id: e7a470ebcfe3163fc8f6250b32e2340e7e156462 $
 */
public class Camera {

	private int width;
	private int height;
	private float aspect;

	private float zNear;
	private float zFar;

	private Vector3D upVector;
	private Vector3D lookFrom;
	private Vector3D lookAt;

	private Frustum frustum;

	/**
	 * Public default constructor
	 */
	protected Camera() {
		width = 0;
		height = 0;
		aspect = 1;
		zNear = 10f;
		zFar = 4000f;
		upVector = new Vector3D(0f,1f,0f);
		lookFrom = new Vector3D(0f,50f,400f);
		lookAt = new Vector3D(0f,50f,0f);
		frustum = new Frustum();
	}

	/**
	 * @return float
	 */
	public float getZNear() {
		return zNear;
	}

	/**
	 * @param zNear
	 */
	public void setZNear(float zNear) {
		this.zNear = zNear;
	}

	/**
	 * @return float
	 */
	public float getZFar() {
		return zFar;
	}

	/**
	 * @param zFar
	 */
	public void setZFar(float zFar) {
		this.zFar = zFar;
	}

	/**
	 * @return up vector
	 */
	public Vector3D getUpVector() {
		return upVector;
	}

	/**
	 * @return look from vector
	 */
	public Vector3D getLookFrom() {
		return lookFrom;
	}

	/**
	 * @return look at vector
	 */
	public Vector3D getLookAt() {
		return lookAt;
	}

	/**
	 * @return frustum
	 */
	protected Frustum getFrustum() {
		return frustum;
	}

	/**
	 * Computes the up vector for given look from and look at vectors
	 * @param look from
	 * @param look at
	 * @return up vector
	 */
	public static Vector3D computeUpVector(Vector3D lookFrom, Vector3D lookAt) {
		Vector3D forward = lookAt.clone().sub(lookFrom).normalize();
		if (Math.abs(forward.getY()) > 0.99f) {
			Vector3D upVector = new Vector3D(0f, 0f, forward.getY()).normalize();
			return upVector;
		}
		Vector3D right = Vector3D.computeCrossProduct(forward, new Vector3D(0f, 1f, 0f)).normalize();
		Vector3D upVector = Vector3D.computeCrossProduct(right, forward).normalize();
		return upVector;
	}

 	/**
	 * Computes the projection matrix
	 * 	based on: http://www.songho.ca/opengl/gl_transform.html
	 * @param field of view
	 * @param aspect
	 * @param z near
	 * @param z far
	 * @return projection matrix
	 */
	private static Matrix3D computeProjectionMatrix(float fieldOfView, float aspect, float zNear, float zFar) {
		float tangent = (float)Math.tan(fieldOfView / 2.0f * 3.14159265f / 180.0f);
		float height = zNear * tangent;
		float width = height * aspect;

		//
		return computeFrustumMatrix(-width, width, -height, height, zNear, zFar);
	}

	/**
	 * Computes frustum matrix
	 * 	based on: http://www.songho.ca/opengl/gl_transform.html
	 * @param left
	 * @param right
	 * @param top
	 * @param bottom
	 * @param near
	 * @param far
	 * @return
	 */
	private static Matrix3D computeFrustumMatrix(float left, float right, float bottom, float top, float near, float far) {
		return new Matrix3D(
			2.0f * near / (right - left),
			0.0f,
			0.0f,
			0.0f,
			0.0f,
			2.0f * near / (top - bottom),
			0.0f,
			0.0f,
			(right + left) / (right - left),
			(top + bottom) / (top - bottom),
			-(far + near) / (far - near),
			-1.0f,
			0.0f,
			0.0f,
			-(2.0f * far * near) / (far - near),
			1.0f
		);
	}

	/**
	 * Computes projection matrix for given look from, look at and up vector
	 * @param look from
	 * @param look at
	 * @param up vector
	 * @return model view matrix
	 */
	private static Matrix3D computeModelViewMatrix(Vector3D lookFrom, Vector3D lookAt, Vector3D upVector) {
		Vector3D forward = lookAt.clone().sub(lookFrom).normalize();
		Vector3D side = Vector3D.computeCrossProduct(forward, upVector).normalize();
		Vector3D up = Vector3D.computeCrossProduct(side, forward);
		float[] sideXYZ = side.getArray();
		float[] forwardXYZ = forward.getArray();
		float[] upXYZ = up.getArray();

		//
		return
			new Matrix3D().
			identity().
			translate(lookFrom.clone().scale(-1f)).multiply(
				new Matrix3D(
					sideXYZ[0], upXYZ[0], -forwardXYZ[0], 0.0f,
					sideXYZ[1], upXYZ[1], -forwardXYZ[1], 0.0f,
					sideXYZ[2], upXYZ[2], -forwardXYZ[2], 0.0f,
					0f, 0f, 0f, 1.0f
				)
			);
	}

	/**
	 * Sets up camera while resizing the view port
	 * @param gl
	 * @param width
	 * @param height
	 */
	protected void updateCamera(int width, int height) {
		// setup new view port if required
		if (this.width != width || this.height != height) {
			if (height <= 0) height = 1;
			aspect = (float) width / (float) height;
			Engine.renderer.setViewPort(0, 0, width, height);
			Engine.renderer.updateViewPort();
		} else {
			aspect = (float) width / (float) height;
		}

		// setup projection and model view
		Engine.renderer.getProjectionMatrix().set(computeProjectionMatrix(45.0f, aspect, zNear, zFar));
		Engine.renderer.updateProjectionMatrix();
		Engine.renderer.getModelViewMatrix().set(computeModelViewMatrix(lookFrom, lookAt, upVector));
		Engine.renderer.updateModelViewMatrix();

		// update frustum
		frustum.updateFrustum();
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "Camera [width=" + width + ", height=" + height + ", aspect="
				+ aspect + ", upVector=" + upVector + ", lookFrom=" + lookFrom
				+ ", lookAt=" + lookAt + ", frustum=" + frustum + "]";
	}

}