package net.drewke.tdme.engine;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import javax.media.opengl.DebugGL2;
import javax.media.opengl.DebugGL3;
import javax.media.opengl.GL2;
import javax.media.opengl.GL3;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLProfile;
import javax.media.opengl.glu.GLU;

import net.drewke.tdme.model.Model;
import net.drewke.tdme.primitives.BoundingBox;

import com.jogamp.opengl.util.gl2.GLUT;

/**
 * Engine main class
 * @author Andreas Drewke
 * @version $Id: c0026d215ec34568a95dd1e94319e22cc3a1fe5b $
 */
public class Engine {

	private static GLU glu;
	private static GLUT glut;
	protected static Engine instance;
	protected static GLRenderer renderer;

	private int width;
	private int height;

	private Timing timing;
	private Camera camera;
	private Light[] lights;

	private TextureManager textureManager;
	private VBOManager vboManager;
	private MeshManager meshManager;

	private Hashtable<String, BoundingBox> boundingBoxCache;

	private Hashtable<String,Object3D> objects;
	private Hashtable<String,Object3D> objectsToRemove;

	private boolean shadowMappingEnabled = true;
	private ShadowMapping shadowMapping;

	private GL3LightingShader gl3LightingShader;

	/**
	 * @return GLU
	 */
	public static GLU getGLU() {
		return glu;
	}

	/**
	 * @return GLUT
	 */
	public static GLUT getGLUT() {
		return glut;
	}

	/**
	 * Returns engine instance
	 * 	Constructor has to be called first
	 * @return
	 */
	public static Engine getInstance() {
		return instance;
	}

	/**
	 * @return supported GL profile
	 */
	public static GLProfile getProfile() {
		GLProfile glp = null;
		if (GLProfile.isAvailable(GLProfile.GL3)) {
			System.out.println("Engine::Using GL3");
			glp = GLProfile.get(GLProfile.GL3);
		} else
		if (GLProfile.isAvailable(GLProfile.GL3bc)) {
			System.out.println("Engine::Using GL3bc");
			glp = GLProfile.get(GLProfile.GL3bc);
		} else
		if (GLProfile.isAvailable(GLProfile.GL2)) {
			System.out.println("Engine::GL2");
			glp = GLProfile.get(GLProfile.GL2);
		} else {
			System.out.println("Engine::No suiting OpenGL profile available!");
			return null;
		}
		System.out.println("Engine::Using " + glp + ", GL2 = " + glp.isGL2() + ", GLES2 = " + glp.isGLES2() + ", GL3 = " + glp.isGL3());
		return glp;
	}

	/**
	 * Updates the renderer with given drawable
	 * @param drawable
	 */
	private void updateRenderer(GLAutoDrawable drawable) {
		if (drawable.getGL().isGL3()) {
			GL3 gl = (GL3)drawable.getGL().getGL3();
			// notify gl context to renderer
			renderer.setGL(gl);
		} else
		if (drawable.getGL().isGL2()) {
			GL2 gl = (GL2)drawable.getGL().getGL2();
			// notify gl context to renderer
			renderer.setGL(gl);
		} else {
			System.out.println("Engine::updateRenderer(): unsupported GL!");
		}
	}

	/**
	 * Public default constructor
	 */
	public Engine() {
		instance = this;
		renderer = null;
		width = 0;
		height = 0;
		timing = new Timing();
		camera = new Camera();
		lights = new Light[8];
		for (int i = 0; i < lights.length; i++) lights[i] = new Light(i);
		textureManager = new TextureManager();
		vboManager = new VBOManager();
		meshManager = new MeshManager();
		boundingBoxCache = new Hashtable<String, BoundingBox>();
		objects = new Hashtable<String,Object3D>();
		objectsToRemove = new Hashtable<String,Object3D>();

		// shadow mapping
		shadowMapping = null;

		// gl 3 stuff
		gl3LightingShader = null;
	}

	/**
	 * @return shadow mapping or null if disabled
	 */
	protected ShadowMapping getShadowMapping() {
		return shadowMapping;
	}

	/**
	 * @return Timing
	 */
	public Timing getTiming() {
		return timing;
	}

	/**
	 * @return Camera
	 */
	public Camera getCamera() {
		return camera;
	}

	/**
	 * @return lights
	 */
	public Light[] getLights() {
		return lights;
	}

	/**
	 * Returns light at idx (0 <= idx < 8)
	 * @param idx
	 * @return Light
	 */
	public Light getLightAt(int idx) {
		assert(idx >= 0 && idx < 8);
		return lights[idx];
	}

	/**
	 * @return texture manager
	 */
	protected TextureManager getTextureManager() {
		return textureManager;
	}

	/**
	 * @return vertex buffer object manager
	 */
	protected VBOManager getVBOManager() {
		return vboManager;
	}

	/**
	 * @return mesh manager
	 */
	protected MeshManager getMeshManager() {
		return meshManager;
	}

	/**
	 * Gets a models bounding box
	 * @param model
	 * @return bounding box
	 */
	protected BoundingBox getModelBoundingBox(Model model) {
		BoundingBox boundingBox = boundingBoxCache.get(model.getId());
		if (boundingBox == null) {
			boundingBox = ModelUtilities.createBoundingBox(new Object3DModel(model));
			boundingBoxCache.put(model.getId(), boundingBox);
		}
		return boundingBox;
	}

	/**
	 * @return a copy of hashtable containing current objects indexed by id
	 */
	public Hashtable<String, Object3D> getObjects() {
		return (Hashtable<String, Object3D>)objects.clone();
	}

	/**
	 * Returns a object by given id
	 * @param id
	 * @return object3d or null
	 */
	public Object3D getObject(String id) {
		return objects.get(id);
	}

	/**
	 * Adds an object by id
	 * @param id
	 * @return
	 */
	public void addObject(Object3D object) {
		Object3D oldObject = objects.put(object.getId(), object);
		// unload old object
		if (oldObject != null) {
			objectsToRemove.put(object.getId(), object);
		}
	}

	/**
	 * Removes an object
	 * @param id
	 */
	public void removeObject(String id) {
		Object3D object = objects.get(id);
		if (object != null) {
			objectsToRemove.put(object.getId(), object);
		}
	}

	/**
	 * Removes all objects and caches
	 */
	public void reset() {
		Iterator<Map.Entry<String, Object3D>> i = objects.entrySet().iterator();
		while (i.hasNext()) {
			Map.Entry<String, Object3D> entry = i.next();
			Object3D object = entry.getValue();
			objectsToRemove.put(object.getId(), object);
		}
		boundingBoxCache.clear();
	}

	/**
	 * Initialize render engine
	 * @param drawable
	 */
	public void init(GLAutoDrawable drawable) {
		init(drawable, false);
	}

	/**
	 * Initialize render engine
	 * @param drawable
	 * @param debug
	 */
	public void init(GLAutoDrawable drawable, boolean debug) {
		if (drawable.getGL().isGL3()) {
			GL3 gl = (GL3)drawable.getGL().getGL3();
			if (debug == true) {
				drawable.setGL(new DebugGL3(gl));
			}
			// use gl3 renderer
			renderer = new GL3Renderer() {
				protected void updateProjectionMatrix() {
					if (gl3LightingShader != null) gl3LightingShader.updateMatrices(this);
					if (shadowMapping != null) shadowMapping.updateMVPMatrices(this);
				}
				protected void updateModelViewMatrix() {
					if (gl3LightingShader != null) gl3LightingShader.updateMatrices(this);
					if (shadowMapping != null) shadowMapping.updateMVPMatrices(this);
				}
				protected void bindTexture(int textureId) {
					super.bindTexture(textureId);
					if (gl3LightingShader != null) gl3LightingShader.bindTexture(this, textureId);
				}
				protected void updateTextureMatrix() {
					// no op
				}
				protected void updateLight(int lightId) {
					if (gl3LightingShader != null) gl3LightingShader.updateLight(this, lightId);
				}
				protected void updateMaterial() {
					if (gl3LightingShader != null) gl3LightingShader.updateMaterial(this);
				}
			};
			// notify gl context to renderer
			renderer.setGL(gl);
		} else
		if (drawable.getGL().isGL2()) {
			GL2 gl = (GL2)drawable.getGL().getGL2();
			if (debug == true) {
				drawable.setGL(new DebugGL2(gl));
			}
			// use gl2 renderer
			renderer = new GL2Renderer() {
				protected void updateProjectionMatrix() {
					super.updateProjectionMatrix();
					if (shadowMapping != null) shadowMapping.updateMVPMatrices(this);
				}
				protected void updateModelViewMatrix() {
					super.updateModelViewMatrix();
					if (shadowMapping != null) shadowMapping.updateMVPMatrices(this);
				}
			};
			// notify gl context to renderer
			renderer.setGL(gl);
		} else {
			System.out.println("Engine::init(): unsupported GL!");
		}

		// create opengl 3 specific stuff
		if (drawable.getGL().isGL3()) {
			gl3LightingShader = new GL3LightingShader();
			gl3LightingShader.init();
		}

		// initialize shadow mapping
		if (shadowMappingEnabled == true) {
			shadowMapping = new ShadowMapping();
			shadowMapping.init();
		}

		// init
		renderer.init();
	}

	/**
	 * Reshape
	 * @param drawable
	 * @param x
	 * @param y
	 * @param width
	 * @param height
	 */
	public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
		// update our width and height
		this.width = width;
		this.height = height;

		// update renderer
		updateRenderer(drawable);

		// update shadow mapping
		if (shadowMapping != null) {
			shadowMapping.reshape(width, height);
		}
	}

	/**
	 * Renders the scene
	 * @param drawable
	 */
	public void display(GLAutoDrawable drawable) {
		// update renderer
		updateRenderer(drawable);

		// update timing
		timing.updateTiming();

		// do requested object removals
		Iterator<Map.Entry<String, Object3D>> it = objectsToRemove.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry<String, Object3D> entry = it.next();
			Object3D object = entry.getValue();
			object.dispose();
			objects.remove(entry.getKey());
		}
		//	clear objects to remove
		objectsToRemove.clear();

		// update camera
		camera.updateCamera(width, height);

		// compute transformations and check if visible 
		Vector<Object3D> visibleObjects = new Vector<Object3D>();
		it = objects.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry<String, Object3D> entry = it.next();
			Object3D object = entry.getValue();

			// check if disabled
			if (object.isEnabled() == false) {
				continue;
			}

			// draw if visible
			if (camera.getFrustum().isVisible(object.getBoundingBoxTransformed())) {
				object.computeTransformations();
				visibleObjects.add(object);
			}
		}

		// render shadow maps
		if (shadowMapping != null) shadowMapping.createShadowMaps();

		// clear previous frame values
		renderer.clear(renderer.CLEAR_DEPTH_BUFFER_BIT | renderer.CLEAR_COLOR_BUFFER_BIT);

		// enable materials
		renderer.setMaterialEnabled();

		// setup up gl3 stuff
		if (gl3LightingShader != null) {
			gl3LightingShader.useProgram();
		}

		// update lights
		for (int j = 0; j < lights.length; j++) {
			lights[j].update();
		}

		// first render pass, draw color buffer and depth buffer
		it = objects.entrySet().iterator();
		for (Object3D object: visibleObjects) {
			// check if disabled
			if (object.isEnabled() == false) {
				continue;
			}

			// draw if visible
			object.render();
		}

		// setup up gl3 stuff
		if (gl3LightingShader != null) {
			gl3LightingShader.unUseProgram();
		}

		// render shadows if required
		if (shadowMapping != null) shadowMapping.renderShadowMaps(visibleObjects);

		// disable materials
		renderer.setMaterialDisabled();
	}

	/**
	 * Shutdown the engine
	 * @param drawable
	 */
	public void dispose(GLAutoDrawable drawable) {
		// update renderer
		updateRenderer(drawable);

		// dispose objects
		Iterator<Map.Entry<String, Object3D>> it = objects.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry<String, Object3D> entry = it.next();
			Object3D object = entry.getValue();
			object.dispose();
		}

		//
		if (shadowMapping != null) shadowMapping.dispose();
	}

	/**
	 * Set up GUI mode rendering
	 * @param drawable
	 */
	public void initGUIMode(GLAutoDrawable drawable) {
		renderer.initGuiMode(drawable.getWidth(), drawable.getHeight());
	}

	/**
	 * Set up GUI mode rendering
	 * @param drawable
	 */
	public void doneGUIMode(GLAutoDrawable drawable) {
		renderer.doneGuiMode();
	}

	/**
	 * Static class initialization
	 */
	static {
		glu = new GLU();
		glut = new GLUT();
	}

}