package net.drewke.tdme.engine;

import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.primitives.BoundingBox;
import net.drewke.tdme.primitives.Plane;
import net.drewke.tdme.primitives.Sphere;

/**
 * Frustum class
 * 	based on http://www.crownandcutlass.com/features/technicaldetails/frustum.html
 * @author Mark Morley, Andreas Drewke
 * @version $Id: a631918a0535d34e0edb972f4f41bddf7b160ab4 $
 */
public class Frustum {

	// right, left, bottom, top, far, near
	protected static final int PLANE_RIGHT = 0; 
	protected static final int PLANE_LEFT = 1;
	protected static final int PLANE_BOTTOM = 2;
	protected static final int PLANE_TOP = 3;
	protected static final int PLANE_FAR = 4;
	protected static final int PLANE_NEAR = 5;

	private Plane[] planes;

	/**
	 * Public default constructor
	 */
	protected Frustum() {
		planes = new Plane[6];
		for (int i = 0; i < 6; i++) {
			planes[i] = new Plane();
		}
	}

	/**
	 * @return planes
	 */
	protected Plane[] getPlanes() {
		return planes;
	}

	/**
	 * Setups frustum, should be called if frustum did change 
	 * @param gl
	 */
	protected void updateFrustum() {
		Matrix3D frustumMatrix =
			new Matrix3D(Engine.renderer.getProjectionMatrix()).transpose().multiply(
				new Matrix3D(Engine.renderer.getModelViewMatrix()).transpose()
			);

		float[] data = frustumMatrix.getArray();

		float x,y,z,d,t;

		// right plane
		x = data[12] - data[0];
		y = data[13] - data[1];
		z = data[14] - data[2];
		d = data[15] - data[3];

		// 	normalize
		t = (float)Math.sqrt((x * x) + (y * y) + (z * z));
		x /= t; y /= t; z /= t; d /= t;

		//	setup
		planes[0].getNormal().set(x, y, z);
		planes[0].setDistance(d);

		// left plane
		x = data[12] + data[0];
		y = data[13] + data[1];
		z = data[14] + data[2];
		d = data[15] + data[3];

		// 	normalize
		t = (float)Math.sqrt((x * x) + (y * y) + (z * z));
		x /= t; y /= t; z /= t; d /= t;

		//	setup
		planes[1].getNormal().set(x, y, z);
		planes[1].setDistance(d);

		// bottom plane
		x = data[12] + data[4];
		y = data[13] + data[5];
		z = data[14] + data[6];
		d = data[15] + data[7];

		// 	normalize
		t = (float)Math.sqrt((x * x) + (y * y) + (z * z));
		x /= t; y /= t; z /= t; d /= t;

		//	setup
		planes[2].getNormal().set(x, y, z);
		planes[2].setDistance(d);

		// top plane
		x = data[12] - data[4];
		y = data[13] - data[5];
		z = data[14] - data[6];
		d = data[15] - data[7];

		// 	normalize
		t = (float)Math.sqrt((x * x) + (y * y) + (z * z));
		x /= t; y /= t; z /= t; d /= t;

		//	setup
		planes[3].getNormal().set(x, y, z);
		planes[3].setDistance(d);

		// far plane
		x = data[12] - data[8];
		y = data[13] - data[9];
		z = data[14] - data[10];
		d = data[15] - data[11];

		// 	normalize
		t = (float)Math.sqrt((x * x) + (y * y) + (z * z));
		x /= t; y /= t; z /= t; d /= t;

		//	setup
		planes[4].getNormal().set(x, y, z);
		planes[4].setDistance(d);

		// near plane
		x = data[12] + data[8];
		y = data[13] + data[9];
		z = data[14] + data[10];
		d = data[15] + data[11];

		// 	normalize
		t = (float)Math.sqrt((x * x) + (y * y) + (z * z));
		x /= t; y /= t; z /= t; d /= t;

		//	setup
		planes[5].getNormal().set(x, y, z);
		planes[5].setDistance(d);
	}

	/**
	 * Checks if given vector is in frustum
	 * @param v
	 * @return visibility
	 */
	protected boolean isVisible(Vector3D v) {
		float[] vector = v.getArray();
		for (Plane p: planes) {
			// plane normal
			float[] normal = p.getNormal().getArray();
			if ((normal[0] * vector[0]) +
				(normal[1] * vector[1]) +
				(normal[2] * vector[2]) +
				p.getDistance() <= 0) {
				//
				return false;
			}
		}
		return true;
	}

	/**
	 * Checks if sphere is in frustum
	 * @param s
	 * @return visibility
	 */
	protected boolean isVisible(Sphere s) {
		// sphere center
		float[] center = s.getCenter().getArray();
		float radius = s.getRadius();
		for (Plane p: planes) {
			// plane normal
			float[] normal = p.getNormal().getArray();
			if ((normal[0] * center[0]) +
				(normal[1] * center[1]) +
				(normal[2] * center[2]) +
				p.getDistance() <= -radius) {
				//
				return false;
			}
		}
		return true;
	}

	/**
	 * Checks if bounding box is in frustum
	 * @param s
	 * @return visibility
	 */
	protected boolean isVisible(BoundingBox b) {
		float min[] = b.getMin().getArray();
		float max[] = b.getMax().getArray();
		float minX = min[0];
		float minY = min[1];
		float minZ = min[2];
		float maxX = max[0];
		float maxY = max[1];
		float maxZ = max[2];		
		for (Plane p: planes) {
			float[] normal = p.getNormal().getArray();
			float distance = p.getDistance();
			if ((normal[0] * minX) +
				(normal[1] * minY) +
				(normal[2] * minZ) +
				distance > 0) {
				//
				continue;
			}
			if ((normal[0] * maxX) +
				(normal[1] * minY) +
				(normal[2] * minZ) +
				distance > 0) {
				//
				continue;
			}
			if ((normal[0] * minX) +
				(normal[1] * maxY) +
				(normal[2] * minZ) +
				distance > 0) {
				//
				continue;
			}
			if ((normal[0] * maxX) +
				(normal[1] * maxY) +
				(normal[2] * minZ) +
				distance > 0) {
				//
				continue;
			}
			if ((normal[0] * minX) +
				(normal[1] * minY) +
				(normal[2] * maxZ) +
				distance > 0) {
				//
				continue;
			}
			if ((normal[0] * maxX) +
				(normal[1] * minY) +
				(normal[2] * maxZ) +
				distance > 0) {
				//
				continue;
			}
			if ((normal[0] * minX) +
				(normal[1] * maxY) +
				(normal[2] * maxZ) +
				distance > 0) {
				//
				continue;
			}
			if ((normal[0] * maxX) +
				(normal[1] * maxY) +
				(normal[2] * maxZ) +
				distance > 0) {
				//
				continue;
			}
			//
			return false;
		}
		return true;
	}

}