package net.drewke.tdme.engine;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import javax.media.opengl.GL;
import javax.media.opengl.GL2;
import javax.media.opengl.GL3;

import net.drewke.tdme.assets.textures.Texture;
import net.drewke.tdme.math.Matrix3D;

import com.jogamp.common.nio.Buffers;

/**
 * OpenGL 2 renderer
 * @author Andreas Drewke
 * @version $Id: 364f45157aa977df24744da609e48b740a2b5ae7 $
 */
public class GL2Renderer extends GLRenderer {

	private GL2 gl;
	private boolean bufferObjectsAvailable;
	private int viewPortX;
	private int viewPortY;
	private int viewPortWidth;
	private int viewPortHeight;
	private Matrix3D projectionMatrix;
	private Matrix3D modelViewMatrix;
	private Matrix3D[] textureMatrix;
	private int activeTextureUnit;

	/**
	 * Public constructor
	 */
	protected GL2Renderer() {
		// setup GL2 consts
		ID_NONE = 0;

		CLEAR_DEPTH_BUFFER_BIT = GL2.GL_DEPTH_BUFFER_BIT;
		CLEAR_COLOR_BUFFER_BIT = GL2.GL_COLOR_BUFFER_BIT;
		CULLFACE_FRONT = GL2.GL_FRONT;

		CULLFACE_BACK = GL2.GL_BACK;

		CLIENTSTATE_TEXTURECOORD_ARRAY = GL2.GL_TEXTURE_COORD_ARRAY;
		CLIENTSTATE_VERTEX_ARRAY = GL2.GL_VERTEX_ARRAY;
		CLIENTSTATE_NORMAL_ARRAY = GL2.GL_NORMAL_ARRAY;

		TEXTUREUNITS_MAX = 2;

		SHADER_FRAGMENT_SHADER = GL2.GL_FRAGMENT_SHADER;
		SHADER_VERTEX_SHADER = GL2.GL_VERTEX_SHADER;

		DEPTHFUNCTION_LESSEQUAL = GL2.GL_LEQUAL;
		DEPTHFUNCTION_EQUAL = GL2.GL_EQUAL;

		//
		gl = null;
		bufferObjectsAvailable = false;
		viewPortX = 0;
		viewPortY = 0;
		viewPortWidth = 0;
		viewPortHeight = 0;
		projectionMatrix = new Matrix3D().identity();
		modelViewMatrix = new Matrix3D().identity();
		textureMatrix = new Matrix3D[TEXTUREUNITS_MAX];
		for (int i = 0; i < textureMatrix.length; i++) {
			textureMatrix[i] = new Matrix3D().identity();
		}
		activeTextureUnit = 0;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setGL(javax.media.opengl.GL2)
	 */
	protected void setGL(GL gl) {
		this.gl = (GL2)gl;
	}

	/**
	 * Checks if VBO is available
	 * 	based on http://www.felixgers.de/teaching/jogl/vertexBufferObject.html
	 * @param gl
	 * @return vbo availability
	 */
	private boolean checkBufferObjectsAvailable() {
		// Check if extension is available.
		boolean extensionOK = gl.isExtensionAvailable("GL_ARB_vertex_buffer_object");

		// Check for VBO functions.
		boolean functionsOK =
			gl.isFunctionAvailable("glGenBuffersARB") &&
			gl.isFunctionAvailable("glBindBufferARB") &&
			gl.isFunctionAvailable("glBufferDataARB") &&
			gl.isFunctionAvailable("glDeleteBuffersARB");

		return extensionOK == true && functionsOK == true; 
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#init()
	 */
	protected void init() {
		// check VBO
		bufferObjectsAvailable = checkBufferObjectsAvailable();
	
		// setup open gl
		gl.glShadeModel(GL2.GL_SMOOTH);									// Enable Smooth Shading
		gl.glClearColor(0.0f, 0.0f, 0.0f, 0.5f);						// Black Background
		gl.glClearDepth(1.0f); 											// Depth Buffer Setup
		gl.glEnable(GL2.GL_DEPTH_TEST); 								// Enables Depth Testing
		gl.glFrontFace(GL2.GL_CCW);										// Triangles are in counter clockwise order
		gl.glEnable(GL2.GL_CULL_FACE);									// The Type Of Depth Testing To Do
		gl.glDepthFunc(GL2.GL_LEQUAL);
		gl.glHint(GL2.GL_PERSPECTIVE_CORRECTION_HINT, GL2.GL_NICEST);	// Really Nice Perspective Calculations
		gl.glEnable(GL2.GL_TEXTURE_2D);									// enable texture 2d
		gl.glEnable(GL2.GL_LIGHTING);									// enable lighting
		gl.glBlendFunc(GL2.GL_SRC_ALPHA, GL2.GL_ONE_MINUS_SRC_ALPHA);  // enable alpha transparency
		gl.glBlendEquation(GL2.GL_FUNC_ADD);
		gl.glDisable(GL2.GL_BLEND);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#isBufferObjectsAvailable()
	 */
	protected boolean isBufferObjectsAvailable() {
		return bufferObjectsAvailable;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#getTextureUnits()
	 */
	protected int getTextureUnits() {
		// check for number of texture units
		IntBuffer tmp = IntBuffer.allocate(1);
		gl.glGetIntegerv(GL2.GL_MAX_TEXTURE_UNITS, tmp);
		while (tmp.remaining() > 0) {
			int textureUnits = tmp.get();
			return textureUnits;
		}
		return -1;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#loadShader(int, java.lang.String, java.lang.String)
	 */
	protected int loadShader(int type, String pathName, String fileName) {
		// create shader
		int handle = gl.glCreateShaderObjectARB(type);

		// exit if no handle returned
		if (handle == 0) return 0;

		// shader source
		String[] shaderSources = new String[1];
		shaderSources[0] = new String();

		// read shader source
		DataInputStream sourceInputStream = null;
		try {
			sourceInputStream = new DataInputStream(new FileInputStream(pathName + "/" + fileName));
			BufferedReader reader = new BufferedReader(new InputStreamReader(sourceInputStream));
			String line;
			while ((line = reader.readLine()) != null) {
				shaderSources[0]+= line + "\n";
			}
			sourceInputStream.close();
		} catch (IOException ioe) {
			return 0;
		} finally {
			try { sourceInputStream.close(); } catch (IOException ioe) {}
		}

		// load source
		gl.glShaderSourceARB(
			handle,
			1,
			shaderSources,
			null
		);

		// compile
		gl.glCompileShaderARB(handle);

		// check state
		IntBuffer compileStatus = IntBuffer.allocate(1);
		gl.glGetObjectParameterivARB(handle, GL2.GL_OBJECT_COMPILE_STATUS_ARB, compileStatus);
		while (compileStatus.remaining() > 0) {
			int result = compileStatus.get();
			if (result == 0) {
				// get error
				IntBuffer infoLogLengthBuffer = Buffers.newDirectIntBuffer(1);
				ByteBuffer infoLogBuffer = Buffers.newDirectByteBuffer(2048); 
		        gl.glGetShaderInfoLog(handle, infoLogBuffer.limit(), infoLogLengthBuffer, infoLogBuffer); 
		        final byte[] infoLogBytes = new byte[infoLogLengthBuffer.get()]; 
		        infoLogBuffer.get(infoLogBytes); 
		        String infoLogString = new String(infoLogBytes);

		        // be verbose
				System.out.println("[" + handle + "]" + pathName + "/" + fileName + ": failed: " + infoLogString);
			}
		}

		//
		return handle;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#useProgram(int)
	 */
	protected void useProgram(int programId) {
		gl.glUseProgramObjectARB(programId);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#createProgram()
	 */
	protected int createProgram() {
		return gl.glCreateProgramObjectARB();
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#attachShaderToProgram(int, int)
	 */
	protected void attachShaderToProgram(int programId, int shaderId) {
		gl.glAttachObjectARB(programId, shaderId);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#linkProgram(int)
	 */
	protected void linkProgram(int programId) {
		gl.glLinkProgramARB(programId);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#getProgramUniformLocation(int, java.lang.String)
	 */
	protected int getProgramUniformLocation(int programId, String name) {
		return gl.glGetUniformLocationARB(programId, name);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setProgramUniformInteger(int, int)
	 */
	protected void setProgramUniformInteger(int uniformId, int value) {
		gl.glUniform1iARB(uniformId, value);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setProgramUniformFloat(int, float)
	 */
	protected void setProgramUniformFloat(int uniformId, float value) {
		gl.glUniform1fARB(uniformId, value);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setProgramUniformFloatMatrix4x4(int, float[])
	 */
	protected void setProgramUniformFloatMatrix4x4(int uniformId, float[] data) {
		gl.glUniformMatrix4fv(uniformId, 1, false, data, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setProgramUniformFloatVec4(int, float[])
	 */
	protected void setProgramUniformFloatVec4(int uniformId, float[] data) {
		gl.glUniform4fv(uniformId, 1, data, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setProgramUniformFloatVec3(int, float[])
	 */
	protected void setProgramUniformFloatVec3(int uniformId, float[] data) {
		gl.glUniform3fv(uniformId, 1, data, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setViewPort(int, int, int, int)
	 */
	protected void setViewPort(int x, int y, int width, int height) {
		this.viewPortX = x;
		this.viewPortY = x;
		this.viewPortWidth = width;
		this.viewPortHeight = height;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#updateViewPort()
	 */
	protected void updateViewPort() {
		gl.glViewport(viewPortX, viewPortY, viewPortWidth, viewPortHeight);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#getProjectionMatrix()
	 */
	protected Matrix3D getProjectionMatrix() {
		return projectionMatrix;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#updateProjectionMatrix()
	 */
	protected void updateProjectionMatrix() {
		gl.glMatrixMode(GL2.GL_PROJECTION);
		gl.glLoadMatrixf(projectionMatrix.getArray(), 0);		
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#getModelViewMatrix()
	 */
	protected Matrix3D getModelViewMatrix() {
		return modelViewMatrix;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#updateModelViewMatrix()
	 */
	protected void updateModelViewMatrix() {
		gl.glMatrixMode(GL2.GL_MODELVIEW);
		gl.glLoadMatrixf(modelViewMatrix.getArray(), 0);		
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#getTextureMatrix()
	 */
	protected Matrix3D getTextureMatrix() {
		return textureMatrix[activeTextureUnit];
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#updateTextureMatrix()
	 */
	protected void updateTextureMatrix() {
		gl.glMatrixMode(GL2.GL_TEXTURE);
		gl.glLoadMatrixf(textureMatrix[activeTextureUnit].getArray(), 0);		
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#enableBlending()
	 */
	protected void enableBlending() {
		gl.glEnable(GL2.GL_BLEND);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#disableBlending()
	 */
	protected void disableBlending() {
		gl.glDisable(GL2.GL_BLEND);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#enableDepthBuffer()
	 */
	protected void enableDepthBuffer() {
		gl.glDepthMask(true);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#disableDepthBuffer()
	 */
	protected void disableDepthBuffer() {
		gl.glDepthMask(false);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setDepthFunction(int)
	 */
	protected void setDepthFunction(int depthFunction) {
		gl.glDepthFunc(depthFunction);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setColorMask(boolean, boolean, boolean, boolean)
	 */
	protected void setColorMask(boolean red, boolean green, boolean blue, boolean alpha) {
		gl.glColorMask(red, green, blue, alpha);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#clear(int)
	 */
	protected void clear(int mask) {
		gl.glClear(mask);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setCullFace(int)
	 */
	protected void setCullFace(int cullFace) {
		gl.glCullFace(cullFace);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#createTexture()
	 */
	protected int createTexture() {
		// generate open gl texture
		final int[] tmp = new int[1];
		gl.glGenTextures(1, tmp, 0);
		int textureId = tmp[0]; // texture id
		return textureId;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#createDepthTexture(int, int)
	 */
	protected int createDepthTexture(int width, int height) {
		int depthTextureGlId;
		// create depth texture
		final int[] tmp = new int[1];
		gl.glGenTextures(1, tmp, 0);
		depthTextureGlId = tmp[0]; // texture id
		gl.glBindTexture(GL2.GL_TEXTURE_2D, depthTextureGlId);

		// GL_LINEAR does not make sense for depth texture. However, next tutorial shows usage of GL_LINEAR and PCF
		gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MIN_FILTER, GL2.GL_NEAREST);
		gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MAG_FILTER, GL2.GL_NEAREST);

		// Remove artifacts on the edges of the shadow map
		gl.glTexParameterf(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_WRAP_S, GL2.GL_CLAMP );
		gl.glTexParameterf(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_WRAP_T, GL2.GL_CLAMP );

		// No need to force GL_DEPTH_COMPONENT24, drivers usually give you the max precision if available 
		gl.glTexImage2D(
			GL2.GL_TEXTURE_2D,
			0,
			GL2.GL_DEPTH_COMPONENT,
			width,
			height,
			0,
			GL2.GL_DEPTH_COMPONENT,
			GL2.GL_FLOAT,
			null
		);

		gl.glBindTexture(GL2.GL_TEXTURE_2D, 0);

		return depthTextureGlId;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#uploadTexture(net.drewke.tdme.assets.textures.Texture)
	 */
	protected void uploadTexture(Texture texture) {
		//
		gl.glTexImage2D(
			GL2.GL_TEXTURE_2D,
			0,
			texture.getDepth() == 32?GL2.GL_RGBA:GL2.GL_RGB,
			texture.getTextureWidth(),
			texture.getTextureHeight(),
			0,
			texture.getDepth() == 32?GL2.GL_RGBA:GL2.GL_RGB,
			GL2.GL_UNSIGNED_BYTE,
			texture.getTextureData()
		);

		//
		gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MIN_FILTER, GL2.GL_LINEAR_MIPMAP_LINEAR);
		gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MAG_FILTER, GL2.GL_LINEAR);

		//
		gl.glGenerateMipmap(GL2.GL_TEXTURE_2D);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#resizeDepthTexture(int, int, int)
	 */
	protected void resizeDepthTexture(int textureId, int width, int height) {
		gl.glBindTexture(GL2.GL_TEXTURE_2D, textureId);
		gl.glTexImage2D(
			GL2.GL_TEXTURE_2D,
			0,
			GL2.GL_DEPTH_COMPONENT,
			width,
			height,
			0,
			GL2.GL_DEPTH_COMPONENT,
			GL2.GL_FLOAT,
			null
		);
		gl.glBindTexture(GL2.GL_TEXTURE_2D, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#bindTexture(int)
	 */
	protected void bindTexture(int textureId) {
		// bind depth texture
		gl.glBindTexture(GL2.GL_TEXTURE_2D, textureId);		
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#disposeTexture(int)
	 */
	protected void disposeTexture(int textureId) {
		// delete texture
		IntBuffer ids = IntBuffer.allocate(1);
		ids.put(textureId);
		ids.flip();
		gl.glDeleteTextures(ids.remaining(), ids);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#createDepthFramebufferObject(int)
	 */
	protected int createDepthFramebufferObject(int depthTextureGlId) {
		int frameBufferGlId;

		// create a frame buffer object
		int fboIds[] = new int[1];
		gl.glGenFramebuffers(1, fboIds, 0);
		frameBufferGlId = fboIds[0];
		gl.glBindFramebuffer(GL2.GL_FRAMEBUFFER, frameBufferGlId);

		// Instruct open gl that we won't bind a color texture with the currently binded FBO
		gl.glDrawBuffer(GL2.GL_NONE);
		gl.glReadBuffer(GL2.GL_NONE);

		// attach the texture to FBO depth attachment point
		gl.glFramebufferTexture2D(
			GL2.GL_FRAMEBUFFER,
			GL2.GL_DEPTH_ATTACHMENT,
			GL2.GL_TEXTURE_2D,
			depthTextureGlId,
			0
		);

		// check FBO status
		int fboStatus = gl.glCheckFramebufferStatus(GL2.GL_FRAMEBUFFER);
		if (fboStatus != GL2.GL_FRAMEBUFFER_COMPLETE) {
			System.out.println("GL_FRAMEBUFFER_COMPLETE_EXT failed, CANNOT use FBO\n");
		}

		// switch back to window-system-provided framebuffer
		gl.glBindFramebuffer(GL2.GL_FRAMEBUFFER, 0);

		//
		return frameBufferGlId;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#bindFrameBuffer(int)
	 */
	protected void bindFrameBuffer(int frameBufferId) {
		gl.glBindFramebuffer(GL2.GL_FRAMEBUFFER, frameBufferId);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#disposeFrameBufferObject(int)
	 */
	protected void disposeFrameBufferObject(int frameBufferId) {
		// delete fbo
		IntBuffer ids = IntBuffer.allocate(1);
		ids.put(frameBufferId);
		ids.flip();
		gl.glDeleteBuffers(ids.remaining(), ids);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#createBufferObjects(int)
	 */
	protected int[] createBufferObjects(int buffers) {
		int[] bufferObjectIds = new int[buffers];
		gl.glGenBuffers(buffers, bufferObjectIds, 0);
		return bufferObjectIds;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#uploadBufferObject(int, int, java.nio.FloatBuffer)
	 */
	protected void uploadBufferObject(int bufferObjectId, int size, FloatBuffer data) {
		gl.glBindBuffer(GL2.GL_ARRAY_BUFFER, bufferObjectId);
		gl.glBufferData(GL2.GL_ARRAY_BUFFER, size, data, GL2.GL_STATIC_DRAW);
		gl.glBindBuffer(GL2.GL_ARRAY_BUFFER, ID_NONE);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#bindTextureCoordinatesBufferObject(int, int)
	 */
	protected void bindTextureCoordinatesBufferObject(int bufferObjectId, int trianglesOffset) {
		gl.glBindBuffer(GL2.GL_ARRAY_BUFFER, bufferObjectId);
		gl.glTexCoordPointer(2, GL2.GL_FLOAT, 0, (long)trianglesOffset * 3L * 2L * 4L);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#bindVerticesBufferObject(int, int)
	 */
	protected void bindVerticesBufferObject(int bufferObjectId, int trianglesOffset) {
		gl.glBindBuffer(GL2.GL_ARRAY_BUFFER, bufferObjectId);
		gl.glVertexPointer(3, GL2.GL_FLOAT, 0, (long)trianglesOffset * 3L * 3L * 4L);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#bindNormalsBufferObject(int, int)
	 */
	protected void bindNormalsBufferObject(int bufferObjectId, int trianglesOffset) {
		gl.glBindBuffer(GL2.GL_ARRAY_BUFFER, bufferObjectId);
		gl.glNormalPointer(GL2.GL_FLOAT, 0, (long)trianglesOffset * 3L * 3L * 4L);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#drawTrianglesFromBufferObjects(int)
	 */
	protected void drawTrianglesFromBufferObjects(int triangles) {
		gl.glDrawArrays(GL2.GL_TRIANGLES, 0, triangles * 3);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#unbindBufferObjects()
	 */
	protected void unbindBufferObjects() {
		// unbind buffers
		gl.glBindBuffer(GL2.GL_ARRAY_BUFFER, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#disposeBufferObjects(int[])
	 */
	protected void disposeBufferObjects(int[] bufferObjectIds) {
		IntBuffer ids = IntBuffer.allocate(3);
		ids.put(bufferObjectIds);
		ids.flip();
		gl.glDeleteBuffers(ids.remaining(), ids);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setTextureCoordinatePointer(java.nio.FloatBuffer)
	 */
	protected void setTextureCoordinatePointer(FloatBuffer buffer) {
		gl.glTexCoordPointer(2, GL2.GL_FLOAT, 0, buffer);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setVerticesPointer(java.nio.FloatBuffer)
	 */
	protected void setVerticesPointer(FloatBuffer buffer) {
		gl.glVertexPointer(3, GL2.GL_FLOAT, 0, buffer);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setNormalsPointer(java.nio.FloatBuffer)
	 */
	protected void setNormalsPointer(FloatBuffer buffer) {
		gl.glNormalPointer(GL2.GL_FLOAT, 0, buffer);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#drawTrianglesFromBufferPointers(int)
	 */
	protected void drawTrianglesFromBufferPointers(int triangles) {
		gl.glDrawArrays(GL2.GL_TRIANGLES, 0, triangles * 3);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#getTextureUnit()
	 */
	protected int getTextureUnit() {
		return activeTextureUnit;
	}

	protected void setTextureUnit(int textureUnit) {
		this.activeTextureUnit = textureUnit;
		gl.glActiveTexture(GL2.GL_TEXTURE0 + textureUnit);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#enableClientState(int)
	 */
	protected void enableClientState(int clientState) {
		gl.glEnableClientState(clientState);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#disableClientState(int)
	 */
	protected void disableClientState(int clientState) {
		gl.glDisableClientState(clientState);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightEnabled(int)
	 */
	protected void setLightEnabled(int lightId) {
		gl.glEnable(lightId + GL2.GL_LIGHT0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightDisabled(int)
	 */
	protected void setLightDisabled(int lightId) {
		gl.glDisable(lightId + GL2.GL_LIGHT0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightAmbient(int, float[])
	 */
	protected void setLightAmbient(int lightId, float[] ambient) {
		gl.glLightfv(lightId + GL2.GL_LIGHT0, GL2.GL_AMBIENT, ambient, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightDiffuse(int, float[])
	 */
	protected void setLightDiffuse(int lightId, float[] diffuse) {
		gl.glLightfv(lightId + GL2.GL_LIGHT0, GL2.GL_DIFFUSE, diffuse, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightPosition(int, float[])
	 */
	protected void setLightPosition(int lightId, float[] position) {
		gl.glLightfv(lightId + GL2.GL_LIGHT0, GL2.GL_POSITION, position, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightSpotDirection(int, float[])
	 */
	protected void setLightSpotDirection(int lightId, float[] spotDirection) {
		gl.glLightfv(lightId + GL2.GL_LIGHT0, GL2.GL_SPOT_DIRECTION, spotDirection, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightSpotExponent(int, float)
	 */
	protected void setLightSpotExponent(int lightId, float spotExponent) {
		gl.glLightf(lightId + GL2.GL_LIGHT0, GL2.GL_SPOT_EXPONENT, spotExponent);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightSpotCutOff(int, float)
	 */
	protected void setLightSpotCutOff(int lightId, float spotCutOff) {
		gl.glLightf(lightId + GL2.GL_LIGHT0, GL2.GL_SPOT_CUTOFF, spotCutOff);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightConstantAttenuation(int, float)
	 */
	protected void setLightConstantAttenuation(int lightId, float constantAttenuation) {
		gl.glLightf(lightId + GL2.GL_LIGHT0, GL2.GL_CONSTANT_ATTENUATION, constantAttenuation);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightLinearAttenuation(int, float)
	 */
	protected void setLightLinearAttenuation(int lightId, float linearAttenuation) {
		gl.glLightf(lightId + GL2.GL_LIGHT0, GL2.GL_LINEAR_ATTENUATION, linearAttenuation);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightQuadraticAttenuation(int, float)
	 */
	protected void setLightQuadraticAttenuation(int lightId, float quadraticAttenuation) {
		gl.glLightf(lightId + GL2.GL_LIGHT0, GL2.GL_QUADRATIC_ATTENUATION, quadraticAttenuation);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#updateLight(int)
	 */
	protected void updateLight(int lightId) {
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialEnabled()
	 */
	protected void setMaterialEnabled() {
		gl.glEnable(GL2.GL_COLOR_MATERIAL);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialDisabled()
	 */
	protected void setMaterialDisabled() {
		gl.glDisable(GL2.GL_COLOR_MATERIAL);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialAmbient(float[])
	 */
	protected void setMaterialAmbient(float[] ambient) {
		gl.glMaterialfv(GL2.GL_FRONT_AND_BACK, GL2.GL_AMBIENT, ambient, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialDiffuse(float[])
	 */
	protected void setMaterialDiffuse(float[] diffuse) {
		gl.glMaterialfv(GL2.GL_FRONT_AND_BACK, GL2.GL_DIFFUSE, diffuse, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialSpecular(float[])
	 */
	protected void setMaterialSpecular(float[] specular) {
		gl.glMaterialfv(GL2.GL_FRONT_AND_BACK, GL2.GL_SPECULAR, specular, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialEmission(float[])
	 */
	protected void setMaterialEmission(float[] emission) {
		gl.glMaterialfv(GL2.GL_FRONT_AND_BACK, GL2.GL_EMISSION, emission, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialShininess(float)
	 */
	protected void setMaterialShininess(float shininess) {
		gl.glMaterialf(GL2.GL_FRONT_AND_BACK, GL2.GL_SHININESS, shininess);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#updateMaterial()
	 */
	protected void updateMaterial() {
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#initGuiMode()
	 */
	protected void initGuiMode(int width, int height) {
		gl.glMatrixMode(GL2.GL_PROJECTION);
		gl.glLoadIdentity();
		gl.glOrtho(0, width, height, 0, -9999, 9999);

		gl.glMatrixMode(GL2.GL_MODELVIEW);
		gl.glLoadIdentity();

		gl.glMatrixMode(GL2.GL_TEXTURE_MATRIX);
		gl.glLoadIdentity();

		gl.glEnable(GL2.GL_BLEND);
		gl.glDisable(GL2.GL_DEPTH_TEST);
		gl.glDisable(GL2.GL_CULL_FACE);
		gl.glDisable(GL2.GL_LIGHTING);

		// save attributes
		gl.glPushAttrib(GL2.GL_ALL_ATTRIB_BITS);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#doneGuiMode()
	 */
	protected void doneGuiMode() {
		gl.glPopAttrib();
		gl.glDisable(GL2.GL_BLEND);
		gl.glEnable(GL2.GL_DEPTH_TEST);
		gl.glEnable(GL2.GL_CULL_FACE);
		gl.glEnable(GL2.GL_LIGHTING);
	}

}