package net.drewke.tdme.engine;

import net.drewke.tdme.math.Matrix3D;

/**
 * Interface to gl3 lighting shader program
 * @author Andreas Drewke
 * @version $Id: 0c3aa5f77aee3fda8fb82319eafe09650036ed2f $
 */
public class GL3LightingShader {

	public final static int MAX_LIGHTS = 8;

	private int gl3RenderLightingProgramId;
	private int gl3RenderLightingFragmentShaderId;
	private int gl3RenderLightingVertexShaderId;

	/*

		We have the following data in shader program:

			int diffuseTextureUnit
			mat4 objectToScreenMatrix
			mat4 modelViewMatrix
			mat4 normalMatrix
			vec4 sceneColor
			
			struct Material {
				vec4 ambient;
				vec4 diffuse;
				vec4 specular;
				float shininess;
			};

			struct Light {
				int enabled;
				vec4 ambient;
				vec4 diffuse;
				vec4 specular;
				vec4 position;
				vec3 spotDirection;
				float spotExponent;
				float spotCosCutoff;
				float constantAttenuation;
				float linearAttenuation;
				float quadraticAttenuation;
			};

			int diffuseTextureUnit
			mat4 objectToScreenMatrix
			mat4 modelViewMatrix
			mat4 normalMatrix
			vec4 sceneColor

			uniform Material material;
			uniform Light lights[MAX_LIGHTS];
	*/

	private int uniformDiffuseTextureUnit;
	private int uniformDiffuseTextureAvailable;
	private int uniformObjectToScreenMatrix;
	private int uniformModelViewMatrix;
	private int uniformNormalMatrix;
	private int uniformSceneColor;
	
	private int uniformMaterialAmbient;
	private int uniformMaterialDiffuse;
	private int uniformMaterialSpecular;
	private int uniformMaterialShininess;

	private int[] uniformLightEnabled;
	private int[] uniformLightAmbient;
	private int[] uniformLightDiffuse;
	private int[] uniformLightSpecular;
	private int[] uniformLightPosition;
	private int[] uniformLightSpotDirection;
	private int[] uniformLightSpotExponent;
	private int[] uniformLightSpotCosCutoff;
	private int[] uniformLightConstantAttenuation;
	private int[] uniformLightLinearAttenuation;
	private int[] uniformLightQuadraticAttenuation;

	private boolean isRunning;

	/**
	 * Protected constructor
	 */
	protected GL3LightingShader() {
		isRunning = false;
	}

	/**
	 * Initialize renderer
	 */
	protected void init() {
		GLRenderer renderer = Engine.getInstance().renderer;
		// lighting
		//	fragment shader
		gl3RenderLightingFragmentShaderId = renderer.loadShader(
			renderer.SHADER_FRAGMENT_SHADER,
			"shader/gl3/lighting",
			"render_fragmentshader.c"
		);

		//	vertex shader
		gl3RenderLightingVertexShaderId = renderer.loadShader(
			renderer.SHADER_VERTEX_SHADER,
			"shader/gl3/lighting",
			"render_vertexshader.c"
		);

		// create, attach and link program
		gl3RenderLightingProgramId = renderer.createProgram();
		renderer.attachShaderToProgram(gl3RenderLightingProgramId, gl3RenderLightingVertexShaderId);
		renderer.attachShaderToProgram(gl3RenderLightingProgramId, gl3RenderLightingFragmentShaderId);
		renderer.linkProgram(gl3RenderLightingProgramId);

		// get uniforms
		//	globals
		uniformDiffuseTextureUnit = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "diffuseTextureUnit");
		uniformDiffuseTextureAvailable = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "diffuseTextureAvailable");
		uniformObjectToScreenMatrix =  renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "objectToScreenMatrix");
		uniformModelViewMatrix =  renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "modelViewMatrix");
		uniformNormalMatrix = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "normalMatrix");
		uniformSceneColor = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "sceneColor");

		//	material
		uniformMaterialAmbient =  renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "material.ambient");
		uniformMaterialDiffuse =  renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "material.diffuse");
		uniformMaterialSpecular =  renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "material.specular");
		uniformMaterialShininess =  renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "material.shininess");

		//	lights
		uniformLightEnabled = new int[MAX_LIGHTS];
		uniformLightAmbient = new int[MAX_LIGHTS];
		uniformLightDiffuse = new int[MAX_LIGHTS];
		uniformLightSpecular = new int[MAX_LIGHTS];
		uniformLightPosition = new int[MAX_LIGHTS];
		uniformLightSpotDirection = new int[MAX_LIGHTS];
		uniformLightSpotExponent = new int[MAX_LIGHTS];
		uniformLightSpotCosCutoff = new int[MAX_LIGHTS];
		uniformLightConstantAttenuation = new int[MAX_LIGHTS];
		uniformLightLinearAttenuation = new int[MAX_LIGHTS];
		uniformLightQuadraticAttenuation = new int[MAX_LIGHTS];
		for (int i = 0; i < MAX_LIGHTS; i++) {
			uniformLightEnabled[i] = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "lights[" + i + "].enabled");
			uniformLightAmbient[i] = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "lights[" + i + "].ambient");
			uniformLightDiffuse[i] = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "lights[" + i + "].diffuse");
			uniformLightSpecular[i] = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "lights[" + i + "].specular");
			uniformLightPosition[i] = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "lights[" + i + "].position");
			uniformLightSpotDirection[i] = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "lights[" + i + "].spotDirection");
			uniformLightSpotExponent[i] = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "lights[" + i + "].spotExponent");
			uniformLightSpotCosCutoff[i] = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "lights[" + i + "].spotCosCutoff");
			uniformLightConstantAttenuation[i] = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "lights[" + i + "].constantAttenuation");
			uniformLightLinearAttenuation[i] = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "lights[" + i + "].linearAttenuation");
			uniformLightQuadraticAttenuation[i] = renderer.getProgramUniformLocation(gl3RenderLightingProgramId, "lights[" + i + "].quadraticAttenuation");
		}
	}

	/**
	 * Use lighting program
	 */
	protected void useProgram() {
		isRunning = true;
		GLRenderer renderer = Engine.getInstance().renderer; 
		renderer.useProgram(gl3RenderLightingProgramId);
		// initialize static uniforms
		renderer.setProgramUniformInteger(uniformDiffuseTextureUnit, 0);
		renderer.setProgramUniformFloatVec4(uniformSceneColor, new float[] {0f,0f,0f,0f});
	}

	/**
	 * Unuse lighting program
	 */
	protected void unUseProgram() {
		Engine.getInstance().renderer.useProgram(Engine.getInstance().renderer.ID_NONE);
		isRunning = false;
	}

	/**
	 * Update material to program
	 * @param gl3 renderer
	 */
	protected void updateMaterial(GL3Renderer renderer) {
		// skip if not running
		if (isRunning == false) return;

		//
		renderer.setProgramUniformFloatVec4(uniformMaterialAmbient, renderer.material.ambient);
		renderer.setProgramUniformFloatVec4(uniformMaterialDiffuse, renderer.material.diffuse);
		renderer.setProgramUniformFloatVec4(uniformMaterialSpecular, renderer.material.specular);
		renderer.setProgramUniformFloat(uniformMaterialShininess, renderer.material.shininess);
	}

	/**
	 * Update light to program
	 * @param renderer
	 * @param light id
	 */
	protected void updateLight(GL3Renderer renderer, int lightId) {
		// skip if not running
		if (isRunning == false) return;

		//
		renderer.setProgramUniformInteger(uniformLightEnabled[lightId], renderer.lights[lightId].enabled);
		renderer.setProgramUniformFloatVec4(uniformLightAmbient[lightId], renderer.lights[lightId].ambient);
		renderer.setProgramUniformFloatVec4(uniformLightDiffuse[lightId], renderer.lights[lightId].diffuse);
		renderer.setProgramUniformFloatVec4(uniformLightSpecular[lightId], renderer.lights[lightId].specular);
		renderer.setProgramUniformFloatVec4(uniformLightPosition[lightId], renderer.lights[lightId].position);
		renderer.setProgramUniformFloatVec3(uniformLightSpotDirection[lightId], renderer.lights[lightId].spotDirection);
		renderer.setProgramUniformFloat(uniformLightSpotExponent[lightId], renderer.lights[lightId].spotExponent);
		renderer.setProgramUniformFloat(uniformLightSpotCosCutoff[lightId], renderer.lights[lightId].spotCosCutoff);
		renderer.setProgramUniformFloat(uniformLightConstantAttenuation[lightId], renderer.lights[lightId].constantAttenuation);
		renderer.setProgramUniformFloat(uniformLightLinearAttenuation[lightId], renderer.lights[lightId].linearAttenuation);
		renderer.setProgramUniformFloat(uniformLightQuadraticAttenuation[lightId], renderer.lights[lightId].quadraticAttenuation);
	}

	/**
	 * Update matrices to program
	 * @param renderer
	 */
	protected void updateMatrices(GL3Renderer renderer) {
		// skip if not running
		if (isRunning == false) return;

		// object to screen matrix
		Matrix3D objectToScreenMatrix = 
			renderer.getModelViewMatrix().clone().
			multiply(
				renderer.getProjectionMatrix()
			);
		// model view matrix
		Matrix3D modelViewMatrix = renderer.getModelViewMatrix().clone();
		// normal matrix
		Matrix3D normalMatrix =
			modelViewMatrix.cloneNoTranslation().
			invert();

		// upload matrices
		renderer.setProgramUniformFloatMatrix4x4(uniformObjectToScreenMatrix, objectToScreenMatrix.getArray());
		renderer.setProgramUniformFloatMatrix4x4(uniformModelViewMatrix, modelViewMatrix.getArray());
		renderer.setProgramUniformFloatMatrix4x4(uniformNormalMatrix, normalMatrix.getArray());
	}

	/**
	 * Bind texture
	 * @param renderer
	 * @param texture id
	 */
	protected void bindTexture(GL3Renderer renderer, int textureId) {
		// skip if not running
		if (isRunning == false) return;

		// check diffuse texture
		if (renderer.getTextureUnit() == 0) {
			renderer.setProgramUniformInteger(uniformDiffuseTextureAvailable, textureId == 0?0:1);			
		}
	}

}
