package net.drewke.tdme.engine;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import javax.media.opengl.GL;
import javax.media.opengl.GL2;
import javax.media.opengl.GL3;

import net.drewke.tdme.assets.textures.Texture;
import net.drewke.tdme.math.Matrix3D;

import com.jogamp.common.nio.Buffers;

/**
 * OpenGL 2 renderer
 * @author Andreas Drewke
 * @version $Id: 37ba7cbd65a67ab743e594471b5a92b6b3572ac6 $
 */
public abstract class GL3Renderer extends GLRenderer {

	/**
	 * Bean holding material data
	 */
	protected class Material {
		float[] ambient;
		float[] diffuse;
		float[] specular;
		float[] emission;
		float shininess;
	}

	/**
	 * Bean holding light data
	 */
	protected class Light {
		int enabled;
		float[] ambient;
		float[] diffuse;
		float[] specular;
		float[] position;
		float[] spotDirection;
		float spotExponent;
		float spotCosCutoff;
		float constantAttenuation;
		float linearAttenuation;
		float quadraticAttenuation;		
	}

	protected GL3 gl;
	private int viewPortX;
	private int viewPortY;
	private int viewPortWidth;
	private int viewPortHeight;
	private Matrix3D projectionMatrix;
	private Matrix3D modelViewMatrix;
	private Matrix3D[] textureMatrix;
	private int activeTextureUnit;

	protected Material material;
	protected Light[] lights;

	/**
	 * Public constructor
	 */
	protected GL3Renderer() {
		// setup GL3 consts
		ID_NONE = 0;

		CLEAR_DEPTH_BUFFER_BIT = GL3.GL_DEPTH_BUFFER_BIT;
		CLEAR_COLOR_BUFFER_BIT = GL3.GL_COLOR_BUFFER_BIT;
		CULLFACE_FRONT = GL3.GL_FRONT;

		CULLFACE_BACK = GL3.GL_BACK;

		CLIENTSTATE_TEXTURECOORD_ARRAY = 2;
		CLIENTSTATE_VERTEX_ARRAY = 0;
		CLIENTSTATE_NORMAL_ARRAY = 1;

		TEXTUREUNITS_MAX = 2;

		SHADER_FRAGMENT_SHADER = GL3.GL_FRAGMENT_SHADER;
		SHADER_VERTEX_SHADER = GL3.GL_VERTEX_SHADER;

		DEPTHFUNCTION_LESSEQUAL = GL3.GL_LEQUAL;
		DEPTHFUNCTION_EQUAL = GL3.GL_EQUAL;

		//
		gl = null;
		viewPortX = 0;
		viewPortY = 0;
		viewPortWidth = 0;
		viewPortHeight = 0;
		projectionMatrix = new Matrix3D().identity();
		modelViewMatrix = new Matrix3D().identity();
		textureMatrix = new Matrix3D[TEXTUREUNITS_MAX];
		for (int i = 0; i < textureMatrix.length; i++) {
			textureMatrix[i] = new Matrix3D().identity();
		}
		activeTextureUnit = 0;

		// set up material
		material = new Material();
		material.ambient = new float[] {0.2f, 0.2f, 0.2f, 1.0f};
		material.diffuse = new float[] {0.8f, 0.8f, 0.8f, 1.0f};
		material.specular = new float[] {0.0f, 0.0f, 0.0f, 1.0f};
		material.emission = new float[] {0.0f, 0.0f, 0.0f, 1.0f};
		material.shininess = 0.0f;

		// set up lights
		lights = new Light[8];
		for (int i = 0; i < lights.length; i++) {
			lights[i] = new Light();
			lights[i].enabled = 0;
			lights[i].ambient = new float[] {0f,0f,0f,1f};
			lights[i].diffuse = new float[] {1f,1f,1f,1f};
			lights[i].specular = new float[] {1f,1f,1f,1f};
			lights[i].position = new float[] {0f,0f,0f,0f};
			lights[i].spotDirection = new float[] {0f,0f,-1f};
			lights[i].spotExponent = 0f;
			lights[i].spotCosCutoff = (float)Math.cos(Math.PI / 180f * 180f);
			lights[i].constantAttenuation = 1f;
			lights[i].linearAttenuation = 0f;
			lights[i].quadraticAttenuation = 0f;	
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setGL(javax.media.opengl.GL3)
	 */
	protected void setGL(GL gl) {
		this.gl = (GL3)gl;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#init()
	 */
	protected void init() {
		// setup open gl
		gl.glClearColor(0.0f, 0.0f, 0.0f, 0.5f);						// Black Background
		gl.glClearDepth(1.0f); 											// Depth Buffer Setup
		gl.glEnable(GL3.GL_DEPTH_TEST); 								// Enables Depth Testing
		gl.glFrontFace(GL3.GL_CCW);										// Triangles are in counter clockwise order
		gl.glEnable(GL3.GL_CULL_FACE);									// The Type Of Depth Testing To Do
		gl.glDepthFunc(GL3.GL_LEQUAL);
		gl.glEnable(GL3.GL_TEXTURE_2D);									// enable texture 2d
		gl.glBlendFunc(GL2.GL_SRC_ALPHA, GL2.GL_ONE_MINUS_SRC_ALPHA);  // enable alpha transparency
		gl.glBlendEquation(GL3.GL_FUNC_ADD);
		gl.glDisable(GL3.GL_BLEND);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#isBufferObjectsAvailable()
	 */
	protected boolean isBufferObjectsAvailable() {
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#getTextureUnits()
	 */
	protected int getTextureUnits() {
		// check for number of texture units
		return -1;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#loadShader(int, java.lang.String, java.lang.String)
	 */
	protected int loadShader(int type, String pathName, String fileName) {
		// create shader
		int handle = gl.glCreateShader(type);

		// exit if no handle returned
		if (handle == 0) return 0;

		// shader source
		String[] shaderSources = new String[1];
		shaderSources[0] = new String();

		// read shader source
		DataInputStream sourceInputStream = null;
		try {
			sourceInputStream = new DataInputStream(new FileInputStream(pathName + "/" + fileName));
			BufferedReader reader = new BufferedReader(new InputStreamReader(sourceInputStream));
			String line;
			while ((line = reader.readLine()) != null) {
				shaderSources[0]+= line + "\n";
			}
			sourceInputStream.close();
		} catch (IOException ioe) {
			return 0;
		} finally {
			try { sourceInputStream.close(); } catch (IOException ioe) {}
		}

		// load source
		gl.glShaderSource(
			handle,
			1,
			shaderSources,
			null
		);

		// compile
		gl.glCompileShader(handle);

		// check state
		IntBuffer compileStatus = IntBuffer.allocate(1);
		gl.glGetShaderiv(handle, GL3.GL_COMPILE_STATUS, compileStatus);
		while (compileStatus.remaining() > 0) {
			int result = compileStatus.get();
			if (result == 0) {
				// get error
				IntBuffer infoLogLengthBuffer = Buffers.newDirectIntBuffer(1);
				ByteBuffer infoLogBuffer = Buffers.newDirectByteBuffer(2048); 
		        gl.glGetShaderInfoLog(handle, infoLogBuffer.limit(), infoLogLengthBuffer, infoLogBuffer); 
		        final byte[] infoLogBytes = new byte[infoLogLengthBuffer.get()]; 
		        infoLogBuffer.get(infoLogBytes); 
		        String infoLogString = new String(infoLogBytes);

		        // be verbose
				System.out.println("[" + handle + "]" + pathName + "/" + fileName + ": failed: " + infoLogString);
			}
		}

		//
		return handle;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#useProgram(int)
	 */
	protected void useProgram(int programId) {
		gl.glUseProgram(programId);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#createProgram()
	 */
	protected int createProgram() {
		return gl.glCreateProgram();
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#attachShaderToProgram(int, int)
	 */
	protected void attachShaderToProgram(int programId, int shaderId) {
		gl.glAttachShader(programId, shaderId);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#linkProgram(int)
	 */
	protected void linkProgram(int programId) {
		gl.glLinkProgram(programId);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#getProgramUniformLocation(int, java.lang.String)
	 */
	protected int getProgramUniformLocation(int programId, String name) {
		return gl.glGetUniformLocation(programId, name);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setProgramUniformInteger(int, int)
	 */
	protected void setProgramUniformInteger(int uniformId, int value) {
		gl.glUniform1i(uniformId, value);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setProgramUniformFloat(int, float)
	 */
	protected void setProgramUniformFloat(int uniformId, float value) {
		gl.glUniform1f(uniformId, value);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setProgramUniformFloatMatrix4x4(int, float[])
	 */
	protected void setProgramUniformFloatMatrix4x4(int uniformId, float[] data) {
		gl.glUniformMatrix4fv(uniformId, 1, false, data, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setProgramUniformFloatVec4(int, float[])
	 */
	protected void setProgramUniformFloatVec4(int uniformId, float[] data) {
		gl.glUniform4fv(uniformId, 1, data, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setProgramUniformFloatVec3(int, float[])
	 */
	protected void setProgramUniformFloatVec3(int uniformId, float[] data) {
		gl.glUniform3fv(uniformId, 1, data, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setViewPort(int, int, int, int)
	 */
	protected void setViewPort(int x, int y, int width, int height) {
		this.viewPortX = x;
		this.viewPortY = x;
		this.viewPortWidth = width;
		this.viewPortHeight = height;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#updateViewPort()
	 */
	protected void updateViewPort() {
		gl.glViewport(viewPortX, viewPortY, viewPortWidth, viewPortHeight);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#getProjectionMatrix()
	 */
	protected Matrix3D getProjectionMatrix() {
		return projectionMatrix;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#getModelViewMatrix()
	 */
	protected Matrix3D getModelViewMatrix() {
		return modelViewMatrix;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#getTextureMatrix()
	 */
	protected Matrix3D getTextureMatrix() {
		return textureMatrix[activeTextureUnit];
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#enableBlending()
	 */
	protected void enableBlending() {
		gl.glEnable(GL3.GL_BLEND);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#disableBlending()
	 */
	protected void disableBlending() {
		gl.glDisable(GL3.GL_BLEND);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#enableDepthBuffer()
	 */
	protected void enableDepthBuffer() {
		gl.glDepthMask(true);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#disableDepthBuffer()
	 */
	protected void disableDepthBuffer() {
		gl.glDepthMask(false);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setDepthFunction(int)
	 */
	protected void setDepthFunction(int depthFunction) {
		gl.glDepthFunc(depthFunction);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setColorMask(boolean, boolean, boolean, boolean)
	 */
	protected void setColorMask(boolean red, boolean green, boolean blue, boolean alpha) {
		gl.glColorMask(red, green, blue, alpha);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#clear(int)
	 */
	protected void clear(int mask) {
		gl.glClear(mask);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setCullFace(int)
	 */
	protected void setCullFace(int cullFace) {
		gl.glCullFace(cullFace);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#createTexture()
	 */
	protected int createTexture() {
		// generate open gl texture
		final int[] tmp = new int[1];
		gl.glGenTextures(1, tmp, 0);
		int textureId = tmp[0]; // texture id
		return textureId;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#createDepthTexture(int, int)
	 */
	protected int createDepthTexture(int width, int height) {
		int depthTextureGlId;
		// create depth texture
		final int[] tmp = new int[1];
		gl.glGenTextures(1, tmp, 0);
		depthTextureGlId = tmp[0]; // texture id
		gl.glBindTexture(GL3.GL_TEXTURE_2D, depthTextureGlId);

		gl.glTexImage2D(
			GL3.GL_TEXTURE_2D,
			0,
			GL3.GL_DEPTH_COMPONENT,
			width,
			height, 
			0,
			GL3.GL_DEPTH_COMPONENT,
			GL3.GL_FLOAT,
			null
		);

		//
		gl.glTexParameteri(GL3.GL_TEXTURE_2D, GL3.GL_TEXTURE_MIN_FILTER, GL3.GL_NEAREST);
		gl.glTexParameteri(GL3.GL_TEXTURE_2D, GL3.GL_TEXTURE_MAG_FILTER, GL3.GL_NEAREST);
		gl.glTexParameteri(GL3.GL_TEXTURE_2D, GL3.GL_TEXTURE_WRAP_S, GL3.GL_CLAMP_TO_EDGE);
		gl.glTexParameteri(GL3.GL_TEXTURE_2D, GL3.GL_TEXTURE_WRAP_T, GL3.GL_CLAMP_TO_EDGE);

		//
		gl.glBindTexture(GL3.GL_TEXTURE_2D, ID_NONE);

		//
		return depthTextureGlId;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#uploadTexture(net.drewke.tdme.assets.textures.Texture)
	 */
	protected void uploadTexture(Texture texture) {
		//
		gl.glTexImage2D(
			GL3.GL_TEXTURE_2D,
			0,
			texture.getDepth() == 32?GL3.GL_RGBA:GL3.GL_RGB,
			texture.getTextureWidth(),
			texture.getTextureHeight(),
			0,
			texture.getDepth() == 32?GL3.GL_RGBA:GL3.GL_RGB,
			GL3.GL_UNSIGNED_BYTE,
			texture.getTextureData()
		);

		//
		gl.glTexParameteri(GL3.GL_TEXTURE_2D, GL3.GL_TEXTURE_MIN_FILTER, GL3.GL_LINEAR_MIPMAP_LINEAR);
		gl.glTexParameteri(GL3.GL_TEXTURE_2D, GL3.GL_TEXTURE_MAG_FILTER, GL3.GL_LINEAR);

		//
		gl.glGenerateMipmap(GL3.GL_TEXTURE_2D);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#resizeDepthTexture(int, int, int)
	 */
	protected void resizeDepthTexture(int textureId, int width, int height) {
		gl.glBindTexture(GL3.GL_TEXTURE_2D, textureId);
		gl.glTexImage2D(
			GL3.GL_TEXTURE_2D,
			0,
			GL3.GL_DEPTH_COMPONENT,
			width,
			height, 
			0,
			GL3.GL_DEPTH_COMPONENT,
			GL3.GL_FLOAT,
			null
		);
		gl.glBindTexture(GL3.GL_TEXTURE_2D, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#bindTexture(int)
	 */
	protected void bindTexture(int textureId) {
		// bind depth texture
		gl.glBindTexture(GL3.GL_TEXTURE_2D, textureId);		
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#disposeTexture(int)
	 */
	protected void disposeTexture(int textureId) {
		// delete texture
		IntBuffer ids = IntBuffer.allocate(1);
		ids.put(textureId);
		ids.flip();
		gl.glDeleteTextures(ids.remaining(), ids);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#createDepthFramebufferObject(int)
	 */
	protected int createDepthFramebufferObject(int depthTextureGlId) {
		int frameBufferGlId;

		// create a frame buffer object
		int fboIds[] = new int[1];
		gl.glGenFramebuffers(1, fboIds, 0);
		frameBufferGlId = fboIds[0];
		gl.glBindFramebuffer(GL3.GL_FRAMEBUFFER, frameBufferGlId);

		// Instruct open gl that we won't bind a color texture with the currently binded FBO
		gl.glDrawBuffer(GL3.GL_NONE);
		gl.glReadBuffer(GL3.GL_NONE);

		// attach the texture to FBO depth attachment point
		gl.glFramebufferTexture(
			GL3.GL_FRAMEBUFFER,
			GL3.GL_DEPTH_ATTACHMENT,
			depthTextureGlId,
			0
		);

		// check FBO status
		int fboStatus = gl.glCheckFramebufferStatus(GL3.GL_FRAMEBUFFER);
		if (fboStatus != GL3.GL_FRAMEBUFFER_COMPLETE) {
			System.out.println("GL_FRAMEBUFFER_COMPLETE_EXT failed, CANNOT use FBO\n");
		}

		// switch back to window-system-provided framebuffer
		gl.glBindFramebuffer(GL3.GL_FRAMEBUFFER, 0);

		//
		return frameBufferGlId;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#bindFrameBuffer(int)
	 */
	protected void bindFrameBuffer(int frameBufferId) {
		gl.glBindFramebuffer(GL3.GL_FRAMEBUFFER, frameBufferId);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#disposeFrameBufferObject(int)
	 */
	protected void disposeFrameBufferObject(int frameBufferId) {
		// delete fbo
		IntBuffer ids = IntBuffer.allocate(1);
		ids.put(frameBufferId);
		ids.flip();
		gl.glDeleteBuffers(ids.remaining(), ids);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#createBufferObjects(int)
	 */
	protected int[] createBufferObjects(int buffers) {
		int[] bufferObjectIds = new int[buffers];
		gl.glGenBuffers(buffers, bufferObjectIds, 0);
		return bufferObjectIds;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#uploadBufferObject(int, int, java.nio.FloatBuffer)
	 */
	protected void uploadBufferObject(int bufferObjectId, int size, FloatBuffer data) {
		gl.glBindBuffer(GL3.GL_ARRAY_BUFFER, bufferObjectId);
		gl.glBufferData(GL3.GL_ARRAY_BUFFER, size, data, GL3.GL_STATIC_DRAW);
		gl.glBindBuffer(GL3.GL_ARRAY_BUFFER, ID_NONE);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#bindTextureCoordinatesBufferObject(int, int)
	 */
	protected void bindTextureCoordinatesBufferObject(int bufferObjectId, int trianglesOffset) {
		gl.glBindBuffer(GL3.GL_ARRAY_BUFFER, bufferObjectId);
		gl.glVertexAttribPointer(CLIENTSTATE_TEXTURECOORD_ARRAY, 2, GL3.GL_FLOAT, false, 0, (long)trianglesOffset * 3L * 2L * 4L);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#bindVerticesBufferObject(int, int)
	 */
	protected void bindVerticesBufferObject(int bufferObjectId, int trianglesOffset) {
		gl.glBindBuffer(GL3.GL_ARRAY_BUFFER, bufferObjectId);
		gl.glVertexAttribPointer(CLIENTSTATE_VERTEX_ARRAY, 3, GL3.GL_FLOAT, false, 0, (long)trianglesOffset * 3L * 3L * 4L);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#bindNormalsBufferObject(int, int)
	 */
	protected void bindNormalsBufferObject(int bufferObjectId, int trianglesOffset) {
		gl.glBindBuffer(GL3.GL_ARRAY_BUFFER, bufferObjectId);
		gl.glVertexAttribPointer(CLIENTSTATE_NORMAL_ARRAY, 3, GL3.GL_FLOAT, false, 0, (long)trianglesOffset * 3L * 3L * 4L);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#drawTrianglesFromBufferObjects(int)
	 */
	protected void drawTrianglesFromBufferObjects(int triangles) {
		gl.glDrawArrays(GL3.GL_TRIANGLES, 0, triangles * 3);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#unbindBufferObjects()
	 */
	protected void unbindBufferObjects() {
		// unbind buffers
		gl.glBindBuffer(GL3.GL_ARRAY_BUFFER, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#disposeBufferObjects(int[])
	 */
	protected void disposeBufferObjects(int[] bufferObjectIds) {
		IntBuffer ids = IntBuffer.allocate(3);
		ids.put(bufferObjectIds);
		ids.flip();
		gl.glDeleteBuffers(ids.remaining(), ids);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setTextureCoordinatePointer(java.nio.FloatBuffer)
	 */
	protected void setTextureCoordinatePointer(FloatBuffer buffer) {
		System.out.println("GL3Renderer::setTextureCoordinatePointer(): not supported!");
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setVerticesPointer(java.nio.FloatBuffer)
	 */
	protected void setVerticesPointer(FloatBuffer buffer) {
		System.out.println("GL3Renderer::setVerticesPointer(): not supported!");
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setNormalsPointer(java.nio.FloatBuffer)
	 */
	protected void setNormalsPointer(FloatBuffer buffer) {
		System.out.println("GL3Renderer::setNormalsPointer(): not supported!");
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#drawTrianglesFromBufferPointers(int)
	 */
	protected void drawTrianglesFromBufferPointers(int triangles) {
		System.out.println("GL3Renderer::drawTrianglesFromBufferPointers(): not supported!");
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#getTextureUnit()
	 */
	protected int getTextureUnit() {
		return activeTextureUnit;
	}

	protected void setTextureUnit(int textureUnit) {
		this.activeTextureUnit = textureUnit;
		gl.glActiveTexture(GL3.GL_TEXTURE0 + textureUnit);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#enableClientState(int)
	 */
	protected void enableClientState(int clientState) {
		gl.glEnableVertexAttribArray(clientState);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#disableClientState(int)
	 */
	protected void disableClientState(int clientState) {
		gl.glDisableVertexAttribArray(clientState);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightEnabled(int)
	 */
	protected void setLightEnabled(int lightId) {
		lights[lightId].enabled = 1;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightDisabled(int)
	 */
	protected void setLightDisabled(int lightId) {
		lights[lightId].enabled = 0;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightAmbient(int, float[])
	 */
	protected void setLightAmbient(int lightId, float[] ambient) {
		lights[lightId].ambient = ambient.clone();
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightDiffuse(int, float[])
	 */
	protected void setLightDiffuse(int lightId, float[] diffuse) {
		lights[lightId].diffuse = diffuse.clone();
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightPosition(int, float[])
	 */
	protected void setLightPosition(int lightId, float[] position) {
		lights[lightId].position = position.clone();
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightSpotDirection(int, float[])
	 */
	protected void setLightSpotDirection(int lightId, float[] spotDirection) {
		lights[lightId].spotDirection = spotDirection.clone();
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightSpotExponent(int, float)
	 */
	protected void setLightSpotExponent(int lightId, float spotExponent) {
		lights[lightId].spotExponent = spotExponent;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightSpotCutOff(int, float)
	 */
	protected void setLightSpotCutOff(int lightId, float spotCutOff) {
		lights[lightId].spotCosCutoff = (float)Math.cos(Math.PI / 180f * spotCutOff);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightConstantAttenuation(int, float)
	 */
	protected void setLightConstantAttenuation(int lightId, float constantAttenuation) {
		lights[lightId].constantAttenuation = constantAttenuation;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightLinearAttenuation(int, float)
	 */
	protected void setLightLinearAttenuation(int lightId, float linearAttenuation) {
		lights[lightId].linearAttenuation = linearAttenuation;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setLightQuadraticAttenuation(int, float)
	 */
	protected void setLightQuadraticAttenuation(int lightId, float QuadraticAttenuation) {
		lights[lightId].quadraticAttenuation = QuadraticAttenuation;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialEnabled()
	 */
	protected void setMaterialEnabled() {
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialDisabled()
	 */
	protected void setMaterialDisabled() {
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialAmbient(float[])
	 */
	protected void setMaterialAmbient(float[] ambient) {
		material.ambient = ambient.clone();
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialDiffuse(float[])
	 */
	protected void setMaterialDiffuse(float[] diffuse) {
		material.diffuse = diffuse.clone();
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialSpecular(float[])
	 */
	protected void setMaterialSpecular(float[] specular) {
		material.specular = specular.clone();
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialEmission(float[])
	 */
	protected void setMaterialEmission(float[] emission) {
		material.emission = emission.clone();
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#setMaterialShininess(float)
	 */
	protected void setMaterialShininess(float shininess) {
		material.shininess = shininess;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#initGuiMode()
	 */
	protected void initGuiMode(int width, int height) {
		gl.glEnable(GL3.GL_BLEND);
		gl.glDisable(GL3.GL_DEPTH_TEST);
		gl.glDisable(GL3.GL_CULL_FACE);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.GLRenderer#doneGuiMode()
	 */
	protected void doneGuiMode() {
		gl.glDisable(GL2.GL_BLEND);
		gl.glEnable(GL2.GL_DEPTH_TEST);
		gl.glEnable(GL2.GL_CULL_FACE);
		gl.glEnable(GL2.GL_LIGHTING);
	}

}
