package net.drewke.tdme.engine;

import java.nio.FloatBuffer;

import javax.media.opengl.GL;

import net.drewke.tdme.assets.textures.Texture;
import net.drewke.tdme.math.Matrix3D;

/**
 * OpenGL renderer interface
 * @author Andreas Drewke
 * @ersion $Id: 8a89d2dde870ac7c9fa126d32d063ac0dcd19413 $
 */
abstract public class GLRenderer {

	protected int ID_NONE = -1;

	protected int CLEAR_DEPTH_BUFFER_BIT = -1;
	protected int CLEAR_COLOR_BUFFER_BIT = -1;

	protected int CULLFACE_FRONT = -1;
	protected int CULLFACE_BACK = -1;

	protected int CLIENTSTATE_TEXTURECOORD_ARRAY = -1;
	protected int CLIENTSTATE_VERTEX_ARRAY = -1;
	protected int CLIENTSTATE_NORMAL_ARRAY = -1;

	protected int TEXTUREUNITS_MAX = -1;

	protected int SHADER_FRAGMENT_SHADER = -1;
	protected int SHADER_VERTEX_SHADER = -1;

	protected int DEPTHFUNCTION_LESSEQUAL = -1;
	protected int DEPTHFUNCTION_EQUAL = -1;

	/**
	 * Updates gl
	 * @param gl
	 */
	abstract protected void setGL(GL gl);

	/**
	 * Initialize renderer
	 */
	abstract protected void init();

	/**
	 * Checks if buffer objects is available
	 * @return buffer objects availability
	 */
	abstract protected boolean isBufferObjectsAvailable();

	/**
	 * @return number of texture units
	 */
	abstract protected int getTextureUnits();

	/**
	 * Loads a shader into open gl stack
	 * @param gl
	 * @param type
	 * @param pathName
	 * @param fileName
	 * @return shader handle
	 */
	abstract protected int loadShader(int type, String pathName, String fileName);

	/**
	 * Use shader program
	 * @param programId
	 */
	abstract protected void useProgram(int programId);

	/**
	 * Creates a shader program
	 * @return int
	 */
	abstract protected int createProgram();

	/**
	 * Attaches a shader to a program
	 * @param program id
	 * @param shader id
	 */
	abstract protected void attachShaderToProgram(int programId, int shaderId);

	/**
	 * Links attached shaders to a program
	 * @param programId
	 */
	abstract protected void linkProgram(int programId);

	/**
	 * Returns location of given uniform variable
	 * @param program id
	 * @param uniform name
	 * @return
	 */
	abstract protected int getProgramUniformLocation(int programId, String name);

	/**
	 * Set up a integer uniform value
	 * @param uniform id
	 * @param value
	 */
	abstract protected void setProgramUniformInteger(int uniformId, int value);

	/**
	 * Set up a float uniform value
	 * @param uniform id
	 * @param value
	 */
	abstract protected void setProgramUniformFloat(int uniformId, float value);

	/**
	 * Set up a float matrix 4x4 uniform value
	 * @param uniform id
	 * @param value
	 */
	abstract protected void setProgramUniformFloatMatrix4x4(int uniformId, float[] value);

	/**
	 * Set up a float vec4 uniform value
	 * @param uniform id
	 * @param data
	 */
	abstract protected void setProgramUniformFloatVec4(int uniformId, float[] data);

	/**
	 * Set up a float vec3 uniform value
	 * @param uniform id
	 * @param data
	 */
	abstract protected void setProgramUniformFloatVec3(int uniformId, float[] data);

	/**
	 * Set up viewport parameter
	 * @param x
	 * @param y
	 * @param width
	 * @param height
	 */
	abstract protected void setViewPort(int x, int y, int width, int height);

	/**
	 * Update viewport
	 */
	abstract protected void updateViewPort();

	/**
	 * @return projection matrix
	 */
	abstract protected Matrix3D getProjectionMatrix();

	/**
	 * Update projection matrix
	 */
	abstract protected void updateProjectionMatrix();

	/**
	 * @return model view matrix
	 */
	abstract protected Matrix3D getModelViewMatrix();

	/**
	 * Update model view matrix
	 */
	abstract protected void updateModelViewMatrix();

	/**
	 * @return texture matrix
	 */
	abstract protected Matrix3D getTextureMatrix();

	/**
	 * Update texture matrix for active texture unit
	 */
	abstract protected void updateTextureMatrix();

	/**
	 * Enables blending
	 */
	abstract protected void enableBlending();

	/**
	 * Disables blending
	 */
	abstract protected void disableBlending();

	/**
	 * Enable depth buffer
	 */
	abstract protected void enableDepthBuffer();

	/**
	 * Disable depth buffer
	 */
	abstract protected void disableDepthBuffer();

	/**
	 * Set up depth function
	 * @param depth function
	 */
	abstract protected void setDepthFunction(int depthFunction);

	/**
	 * Set up GL rendering colormask
	 * @param red
	 * @param green
	 * @param blue
	 * @param alpha
	 */
	abstract protected void setColorMask(boolean red, boolean green, boolean blue, boolean alpha);

	/**
	 * Clear render buffer with given mask
	 * @param mask
	 */
	abstract protected void clear(int mask);

	/**
	 * Sets up which face will be culled
	 * @param cull face
	 */
	abstract protected void setCullFace(int cullFace);

	/**
	 * Creates a texture
	 * @return texture id
	 */
	abstract protected int createTexture();

	/**
	 * Creates a depth texture
	 * @param width
	 * @param height
	 * @return depth texture id
	 */
	abstract protected int createDepthTexture(int width, int height);

	/**
	 * Uploads texture data to current bound texture
	 * @param texture
	 */
	abstract protected void uploadTexture(Texture texture);

	/**
	 * Resizes a depth texture
	 * @param texture id
	 * @param width
	 * @param height
	 */
	abstract protected void resizeDepthTexture(int textureId, int width, int height);

	/**
	 * Binds a texture with given id or unbinds when using ID_NONE
	 * @param textureId
	 */
	abstract protected void bindTexture(int textureId);

	/**
	 * Dispose a texture
	 * @param texture id
	 */
	abstract protected void disposeTexture(int textureId);

	/**
	 * Creates a frame buffer object with depth texture attached
	 * @param depth texture gl id
	 * @return frame buffer object id
	 */
	abstract protected int createDepthFramebufferObject(int depthTextureGlId);

	/**
	 * Enables a framebuffer to be rendered
	 * @param frameBufferId
	 */
	abstract protected void bindFrameBuffer(int frameBufferId);

	/**
	 * Disposes a frame buffer object
	 * @param frame buffer id
	 */
	abstract protected void disposeFrameBufferObject(int frameBufferId);

	/**
	 * Generate buffer objects for vertex data and such
	 * @param buffers
	 * @return ids
	 */
	abstract protected int[] createBufferObjects(int buffers);

	/**
	 * Uploads buffer data to buffer object
	 * @param buffer object id
	 * @param size
	 * @param data
	 */
	abstract protected void uploadBufferObject(int bufferObjectId, int size, FloatBuffer data);

	/**
	 * Bind texture coordinates buffer object
	 * @param buffer object id
	 * @param triangles offset
	 */
	abstract protected void bindTextureCoordinatesBufferObject(int bufferObjectId, int trianglesOffset);

	/**
	 * Bind vertices buffer object
	 * @param buffer object id
	 * @param triangles offset
	 */
	abstract protected void bindVerticesBufferObject(int bufferObjectId, int trianglesOffset);

	/**
	 * Bind normals buffer object
	 * @param buffer object id
	 * @param triangles offset
	 */
	abstract protected void bindNormalsBufferObject(int bufferObjectId, int trianglesOffset);

	/**
	 * Draw triangles from buffer objects
	 * @param triangles
	 */
	abstract protected void drawTrianglesFromBufferObjects(int triangles);

	/**
	 * Unbind buffer objects
	 */
	abstract protected void unbindBufferObjects();

	/**
	 * Disposes a frame buffer object
	 * @param frame buffer id
	 */
	abstract protected void disposeBufferObjects(int[] bufferObjectIds);

	/**
	 * Set up a texture coordinates buffer pointer
	 * @param size
	 * @param buffer
	 */
	abstract protected void setTextureCoordinatePointer(FloatBuffer buffer);

	/**
	 * Set up a vertices buffer pointer
	 * @param buffer
	 */
	abstract protected void setVerticesPointer(FloatBuffer buffer);

	/**
	 * Set up a normals buffer pointer
	 * @param buffer
	 */
	abstract protected void setNormalsPointer(FloatBuffer buffer);

	/**
	 * Draw triangles from buffer pointers
	 * @param triangles
	 */
	abstract protected void drawTrianglesFromBufferPointers(int triangles);

	/**
	 * @return active texture unit
	 */
	abstract protected int getTextureUnit();

	/**
	 * Sets up texture unit
	 * @param texture unit
	 */
	abstract protected void setTextureUnit(int textureUnit);

	/**
	 * Enable a client state / capability
	 * @param client state
	 */
	abstract protected void enableClientState(int clientState);

	/**
	 * Disable a client state / capability
	 * @param client state
	 */
	abstract protected void disableClientState(int clientState);

	/**
	 * Enable light
	 * @param lightId
	 */
	abstract protected void setLightEnabled(int lightId);

	/**
	 * Disable light
	 * @param lightId
	 */
	abstract protected void setLightDisabled(int lightId);

	/**
	 * Set light ambient color
	 * @param light id
	 * @param ambient
	 */
	abstract protected void setLightAmbient(int lightId, float[] ambient);

	/**
	 * Set light diffuse color
	 * @param light id
	 * @param ambient
	 */
	abstract protected void setLightDiffuse(int lightId, float[] diffuse);

	/**
	 * Set light position
	 * @param light id
	 * @param position
	 */
	abstract protected void setLightPosition(int lightId, float[] position);

	/**
	 * Set spot direction
	 * @param light id
	 * @param spot direction
	 */
	abstract protected void setLightSpotDirection(int lightId, float[] spotDirection);

	/**
	 * Set spot exponent
	 * @param light id
	 * @param spot exponent
	 */
	abstract protected void setLightSpotExponent(int lightId, float spotExponent);

	/**
	 * Set spot cut off angle
	 * @param light id
	 * @param spot cut off
	 */
	abstract protected void setLightSpotCutOff(int lightId, float spotCutOff);

	/**
	 * Set up constant attenuation
	 * @param light id
	 * @param constant attenuation
	 */
	abstract protected void setLightConstantAttenuation(int lightId, float constantAttenuation);

	/**
	 * Set up linear attenuation
	 * @param light id
	 * @param linear attenuation
	 */
	abstract protected void setLightLinearAttenuation(int lightId, float linearAttenuation);

	/**
	 * Set up quadratic attenuation
	 * @param light id
	 * @param quadratic attenuation
	 */
	abstract protected void setLightQuadraticAttenuation(int lightId, float QuadraticAttenuation);

	/**
	 * Update light
	 * @param light id
	 */
	abstract protected void updateLight(int lightId);

	/**
	 * Enable material
	 */
	abstract protected void setMaterialEnabled();

	/**
	 * Enable material
	 */
	abstract protected void setMaterialDisabled();

	/**
	 * Set material ambient color
	 * @param ambient
	 */
	abstract protected void setMaterialAmbient(float[] ambient);

	/**
	 * Set material diffuse color
	 * @param diffuse
	 */
	abstract protected void setMaterialDiffuse(float[] diffuse);

	/**
	 * Set material specular color
	 * @param specular
	 */
	abstract protected void setMaterialSpecular(float[] specular);

	/**
	 * Set material emission color
	 * @param emission
	 */
	abstract protected void setMaterialEmission(float[] emission);

	/**
	 * Set material emission color
	 * @param emission
	 */
	abstract protected void setMaterialShininess(float shininess);

	/**
	 * Update material to renderer
	 */
	abstract protected void updateMaterial();

	/**
	 * Set up renderer for GUI rendering
	 * @param width
	 * @param height
	 */
	abstract protected void initGuiMode(int width, int height);

	/**
	 * Set up renderer for 3d rendering
	 */
	abstract protected void doneGuiMode();

}
