package net.drewke.tdme.engine;

import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.math.Vector3D4;

/**
 * Light 
 * @author Andreas Drewke
 * @version $Id: 5824d6689bd880d0dad7e5c22f016744b9bf5117 $
 */
public class Light {

	private int id;
	private boolean enabled = false;
	private float[] ambient = {0f,0f,0f,1f};
	private float[] diffuse = {1f,1f,1f,1f};
	private float[] specular = {1f,1f,1f,1f};
	private Vector3D4 position = new Vector3D4(0f,0f,0f,0f);
	private Vector3D spotDirection = new Vector3D(0f,0f,-1f);
	private float spotExponent = 0f;
	private float spotCutOff = 180f;
	private float constantAttenuation = 1f;
	private float linearAttenuation = 0f;
	private float quadraticAttenuation = 0f;

	/**
	 * Public default constructor
	 */
	protected Light(int id) {
		this.id = id;
	}

	/**
	 * @return light id
	 */
	public int getId() {
		return id;
	}

	/**
	 * @return enabled
	 */
	public boolean isEnabled() {
		return enabled;
	}

	/**
	 * Set enabled
	 * @param enabled
	 */
	public void setEnabled(boolean enabled) {
		this.enabled = enabled;
	}

	/**
	 * @return ambient light RGBA
	 */
	public float[] getAmbient() {
		return ambient;
	}

	/**
	 * Set ambient light
	 * @param r
	 * @param g
	 * @param b
	 * @param a
	 */
	public void setAmbient(float r, float g, float b, float a) {
		this.ambient[0] = r;
		this.ambient[1] = g;
		this.ambient[2] = b;
		this.ambient[3] = a;
	}

	/**
	 * Diffuse light 
	 * @return diffuse light RGBA
	 */
	public float[] getDiffuse() {
		return diffuse;
	}

	/**
	 * Sets up diffuse light
	 * @param r
	 * @param g
	 * @param b
	 * @param a
	 */
	public void setDiffuse(float r, float g, float b, float a) {
		this.diffuse[0] = r;
		this.diffuse[1] = g;
		this.diffuse[2] = b;
		this.diffuse[3] = a;
	}

	/**
	 * @return specular light RGBA
	 */
	public float[] getSpecular() {
		return specular;
	}

	/**
	 * Set specular light
	 * @param r
	 * @param g
	 * @param b
	 * @param a
	 */
	public void setSpecular(float r, float g, float b, float a) {
		this.specular[0] = r;
		this.specular[1] = g;
		this.specular[2] = b;
		this.specular[3] = a;
	}

	/**
	 * @return position of light
	 */
	public Vector3D4 getPosition() {
		return position;
	}

	/**
	 * @return spot direction
	 */
	public Vector3D getSpotDirection() {
		return spotDirection;
	}

	/**
	 * @return spot exponent
	 */
	public float getSpotExponent() {
		return spotExponent;
	}

	/**
	 * set up spot exponent 
	 * @param spot exponent
	 */
	public void setSpotExponent(float spotExponent) {
		this.spotExponent = spotExponent;
	}

	/**
	 * @return spot cutoff
	 */
	public float getSpotCutOff() {
		return spotCutOff;
	}

	/**
	 * set spot cut off
	 * @param spot cut off
	 */
	public void setSpotCutOff(float spotCutOff) {
		this.spotCutOff = spotCutOff;
	}

	/**
	 * @return constant attenuation
	 */
	public float getConstantAttenuation() {
		return constantAttenuation;
	}

	/**
	 * set up constant attenuation
	 * @param constant attenuation
	 */
	public void setConstantAttenuation(float constantAttenuation) {
		this.constantAttenuation = constantAttenuation;
	}

	/**
	 * @return linear attenuation
	 */
	public float getLinearAttenuation() {
		return linearAttenuation;
	}

	/**
	 * set up linear attenuation
	 * @param linarAttenuation
	 */
	public void setLinearAttenuation(float linarAttenuation) {
		this.linearAttenuation = linarAttenuation;
	}

	/**
	 * @return quadratic attenuation
	 */
	public float getQuadraticAttenuation() {
		return quadraticAttenuation;
	}

	/**
	 * set up quadratic attenuation
	 * @param quadraticAttenuation
	 */
	public void setQuadraticAttenuation(float quadraticAttenuation) {
		this.quadraticAttenuation = quadraticAttenuation;
	}

	/**
	 * Update light
	 * @param gl
	 */
	protected void update() {
		if (enabled) {
			Engine.renderer.setLightEnabled(id);
			Engine.renderer.setLightAmbient(id, ambient);
			Engine.renderer.setLightDiffuse(id, diffuse);
			Engine.renderer.setLightPosition(id, position.getArray());
			Engine.renderer.setLightSpotDirection(id, spotDirection.getArray());
			Engine.renderer.setLightSpotExponent(id, spotExponent);
			Engine.renderer.setLightSpotCutOff(id, spotCutOff);
			Engine.renderer.setLightConstantAttenuation(id, constantAttenuation);
			Engine.renderer.setLightLinearAttenuation(id, linearAttenuation);
			Engine.renderer.setLightQuadraticAttenuation(id, quadraticAttenuation);
			Engine.renderer.updateLight(id);
		} else {
			Engine.renderer.setLightDisabled(id);
			Engine.renderer.updateLight(id);
		}
	}

}