package net.drewke.tdme.engine;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.model.AnimationSetup;
import net.drewke.tdme.model.Group;
import net.drewke.tdme.model.Model;
import net.drewke.tdme.primitives.BoundingBox;

/**
 * Model utilities
 * @author Andreas Drewke
 * @version $Id: a2d50219b452636d8be7fcd37f70283f77e2f779 $
 */
public class ModelUtilities {

	/**
	 * Creates a bounding box from given model
	 * @param model
	 * @return axis aligned bounding box
	 */
	public static BoundingBox createBoundingBox(Model model) {
		return ModelUtilities.createBoundingBox(new Object3DModel(model));
	}

	/**
	 * Creates a bounding box from given object3d model
	 * @param model
	 * @return axis aligned bounding box
	 */
	public static BoundingBox createBoundingBox(Object3DModel object3DModel) {
		Model model = object3DModel.getModel();
		AnimationSetup defaultAnimation = model.getAnimationSetup(Model.ANIMATIONSETUP_DEFAULT);

		//
		float minX = 0f, minY = 0f, minZ = 0f;
		float maxX = 0f, maxY = 0f, maxZ = 0f;
		boolean firstVertex = true;

		// create bounding box for whole animation at 60fps
		for (float t = 0.0f;
			t <= (defaultAnimation != null?defaultAnimation.getFrames():0.0f) / model.getFPS();
			t+= 1f / model.getFPS()) {
			//

			// calculate transformations matrices without world transformations
			object3DModel.computeTransformationsMatrices(
				model.getSubGroups(),
				object3DModel.getModel().getImportTransformationsMatrix(),
				(int)(t * 1000f),
				(int)(t * 1000f)
			);

			Object3DGroup.computeTransformations(
				object3DModel.object3dGroups,
				object3DModel.transformationsMatrices
			);

			// parse through object groups to determine min, max
			for(Object3DGroup object3DGroup: object3DModel.object3dGroups) {
				for(Vector3D vertex: object3DGroup.mesh.transformedVertices) {
					// vertex xyz array
					float[] vertexXYZ = vertex.getArray();
					// determine min, max
					if (firstVertex == true) {
						minX = vertexXYZ[0];
						minY = vertexXYZ[1];
						minZ = vertexXYZ[2];
						maxX = vertexXYZ[0];
						maxY = vertexXYZ[1];
						maxZ = vertexXYZ[2];
						firstVertex = false;
					} else {
						if (vertexXYZ[0] < minX) minX = vertexXYZ[0];
						if (vertexXYZ[1] < minY) minY = vertexXYZ[1];
						if (vertexXYZ[2] < minZ) minZ = vertexXYZ[2];
						if (vertexXYZ[0] > maxX) maxX = vertexXYZ[0];
						if (vertexXYZ[1] > maxY) maxY = vertexXYZ[1];
						if (vertexXYZ[2] > maxZ) maxZ = vertexXYZ[2];
					}
				}
			}
		}

		// skip on models without meshes to be rendered
		if (firstVertex == true) return null;

		// otherwise go with bounding box
		return new BoundingBox(minX, minY, minZ, maxX, maxY, maxZ);
	}

	/**
	 * Invert normals of a model
	 * @param model
	 */
	public static void invertNormals(Model model) {
		invertNormals(model.getSubGroups());
	}

	/**
	 * Invert normals recursive
	 * @param groups
	 */
	private static void invertNormals(Hashtable<String, Group>groups) {
		Iterator<Map.Entry<String, Group>> i = groups.entrySet().iterator();
		while (i.hasNext()) {
			Map.Entry<String, Group> entry = i.next();
			Group group = entry.getValue();

			// invert
			for(Vector3D normal: group.getNormals()) {
				normal.scale(-1f);
			}

			// process sub groups
			invertNormals(group.getSubGroups());
		}
	}

}