package net.drewke.tdme.engine;

import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.model.FacesEntity;
import net.drewke.tdme.model.Material;
import net.drewke.tdme.model.Model;
import net.drewke.tdme.primitives.BoundingBox;

/**
 * TDME Object 3D
 * @author Andreas Drewke
 * @version $Id: d5ec920bd2d78b40d8d35c4f3384c1240369ee61 $
 */
public class Object3D extends Object3DBase {

	private String id;
	private boolean enabled;
	private boolean dynamicShadowing;
	private boolean showVisibilityVolume;
	private BoundingBox boundingBox;
	private BoundingBox boundingBoxTransformed;

	/**
	 * Public constructor
	 * @param model
	 */
	public Object3D(String id, Model model) {
		super(model, true);
		this.id = id;
		enabled = true;
		showVisibilityVolume = false;
		dynamicShadowing = false;
		boundingBox = Engine.getInstance().getModelBoundingBox(model);
		boundingBoxTransformed = new BoundingBox();
		boundingBoxTransformed.fromBoundingVolumeWithTransformations(boundingBox, this);
	}

	/**
	 * @return object id
	 */
	public String getId() {
		return id;
	}

	/**
	 * @return true if enabled to be rendered
	 */
	public boolean isEnabled() {
		return enabled;
	}

	/**
	 * Enable/disable rendering
	 * @param enabled
	 */
	public void setEnabled(boolean enabled) {
		this.enabled = enabled;
	}

	/**
	 * @return dynamic shadowing enabled
	 */
	public boolean isDynamicShadowingEnabled() {
		return dynamicShadowing;
	}

	/**
	 * Enable/disable dynamic shadowing
	 * @param dynamicShadowing
	 */
	public void setDynamicShadowingEnabled(boolean dynamicShadowing) {
		this.dynamicShadowing = dynamicShadowing;
	}

	/**
	 * @return show visibility volumes
	 */
	public boolean isShowVisibilityVolume() {
		return showVisibilityVolume;
	}

	/**
	 * Set up show visibility volumes
	 * @param showVisibilityVolumes
	 */
	public void setShowVisibilityVolume(boolean showVisibilityVolume) {
		this.showVisibilityVolume = showVisibilityVolume;
	}

	/**
	 * @return bounding box
	 */
	public BoundingBox getBoundingBox() {
		return boundingBox;
	}

	/**
	 * Retrieves bounding sphere with transformations applied
	 * @return bounding sphere
	 */
	public BoundingBox getBoundingBoxTransformed() {
		return boundingBoxTransformed;
	}

	/**
	 * Dispose render data
	 * @param gl
	 */
	public void dispose() {
		Engine engine = Engine.getInstance();
		TextureManager textureManager = engine.getTextureManager();
		MeshManager meshManager = engine.getMeshManager();
		// delete textures
		for(int i = 0; i < object3dGroups.length; i++) {
			Object3DGroup object3DGroup = object3dGroups[i];

			// dispose mesh
			meshManager.removeMesh(object3DGroup.id);

			// dispose renderer
			object3DGroup.renderer.dispose();

			// dispose textures
			FacesEntity[] facesEntities = object3DGroup.group.getFacesEntities();
			for (int j = 0; j < facesEntities.length; j++) {
				// get entity's material
				Material material = facesEntities[j].getMaterial();

				//	skip if no material was set up
				if (material == null) continue;

				// gl texture id
				int glTextureId = object3DGroup.textureGlIdsByEntities[j];

				// 	skip if no texture was set up
				if (glTextureId == Object3DGroup.GLTEXTUREID_NONE ||
					glTextureId == Object3DGroup.GLTEXTUREID_NOTUSED) continue;

				// remove texture from texture manager
				textureManager.removeTexture(material.getDiffuseTexture().getId());

				// mark as removed
				object3DGroup.textureGlIdsByEntities[j] = Object3DGroup.GLTEXTUREID_NONE;
			}
		}
	}

	/**
	 * Pre render step, computes transformations
	 */
	public void computeTransformations() {
		// do transformations if we have a animation
		if (animationSetup != null) {
			Engine engine = Engine.getInstance();

			// animation timing
			Timing timing = engine.getTiming();
			animationLastAtTime = animationCurrentAtTime;
			long currentFrameAtTime = timing.getCurrentFrameAtTime();
			long lastFrameAtTime = timing.getLastFrameAtTime();
			if (lastFrameAtTime != Timing.UNDEFINED) {
				animationCurrentAtTime+= currentFrameAtTime - lastFrameAtTime; 
			}
	
			// calculate transformations matrices
			computeTransformationsMatrices(
				model.getSubGroups(),
				model.getImportTransformationsMatrix(),
				animationLastAtTime,
				animationCurrentAtTime
			);

			// do transformations in group render data
			Object3DGroup.computeTransformations(object3dGroups, transformationsMatrices);
		}
	}

	/**
	 * Renders this object
	 */
	public void render() {
		//
		ShadowMapping shadowMapping = Engine.getInstance().getShadowMapping();

		// set up local->world transformations matrix
		Matrix3D modelViewMatrix = Engine.renderer.getModelViewMatrix().clone();
		Engine.renderer.getModelViewMatrix().set(
			transformationsMatrix.clone().multiply(modelViewMatrix)
		);
		Engine.renderer.updateModelViewMatrix();

		// do transformation start to shadow mapping
		if (shadowMapping != null) {
			shadowMapping.startObjectTransformations(transformationsMatrix);
		}

		// render model groups recursively
		Engine.renderer.enableClientState(Engine.renderer.CLIENTSTATE_VERTEX_ARRAY);
		Engine.renderer.enableClientState(Engine.renderer.CLIENTSTATE_NORMAL_ARRAY);
		Object3DGroup.renderGroups(object3dGroups);
		Engine.renderer.disableClientState(Engine.renderer.CLIENTSTATE_NORMAL_ARRAY);
		Engine.renderer.disableClientState(Engine.renderer.CLIENTSTATE_VERTEX_ARRAY);

		// do transformations end to shadow mapping
		if (shadowMapping != null) {
			shadowMapping.endObjectTransformations();
		}

		// finish local -> world transformations
		Engine.renderer.getModelViewMatrix().set(
			modelViewMatrix
		);
		Engine.renderer.updateModelViewMatrix();

		//
		/*
		if (showVisibilityVolume) {
			Primitives.doRenderBoundingBox(gl, boundingBoxTransformed);
		}
		*/
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.Transformations#fromTransformations(net.drewke.tdme.engine.Transformations)
	 */
	public void fromTransformations(Transformations transformations) {
		//
		super.fromTransformations(transformations);

		//
		boundingBoxTransformed.fromBoundingVolumeWithTransformations(boundingBox, this);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.engine.Transformations#update()
	 */
	public void update() {
		//
		super.update();

		//
		boundingBoxTransformed.fromBoundingVolumeWithTransformations(boundingBox, this);
	}

}