package net.drewke.tdme.engine;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.model.Animation;
import net.drewke.tdme.model.AnimationSetup;
import net.drewke.tdme.model.Face;
import net.drewke.tdme.model.FacesEntity;
import net.drewke.tdme.model.Group;
import net.drewke.tdme.model.Model;
import net.drewke.tdme.primitives.Triangle;

/**
 * Object3D base class, contains 
 * @author Andreas Drewke
 */
public class Object3DBase extends Transformations {

	protected Model model;
	protected Hashtable<String, Matrix3D> transformationsMatrices;

	protected AnimationSetup animationSetup = null;

	protected long animationCurrentAtTime = 0;
	protected long animationLastAtTime = 0;
	protected boolean animationFinished = false;
	protected float animationTime = 0.0f;

	protected Object3DGroup[] object3dGroups = null;

	protected boolean recreateBuffers;

	/**
	 * Public constructor
	 * @param model
	 */
	protected Object3DBase(Model model, boolean useMeshManager) {
		this.model = model;

		//
		transformationsMatrices = new Hashtable<String, Matrix3D>();

		// animation
		setAnimation(Model.ANIMATIONSETUP_DEFAULT);

		// object 3d groups
		object3dGroups = Object3DGroup.createGroups(this, useMeshManager);

		// calculate transformations matrices
		computeTransformationsMatrices(
			model.getSubGroups(),
			model.getImportTransformationsMatrix(),
			animationLastAtTime,
			animationCurrentAtTime
		);

		// do initial transformations
		Object3DGroup.computeTransformations(object3dGroups, transformationsMatrices);	
	}

	/**
	 * @return model
	 */
	public Model getModel() {
		return model;
	}

	/**
	 * Sets up animation to play
	 * @param id
	 */
	public void setAnimation(String id) {
		AnimationSetup _animationActiveSetup = model.getAnimationSetup(id);
		// only switch animation if we have one
		if (_animationActiveSetup != null) {
			animationSetup = _animationActiveSetup;
			animationLastAtTime = Timing.UNDEFINED;
			animationCurrentAtTime = 0L;
			animationTime = 0.0f;
			animationFinished = false;
		}
	}

	/**
	 * @return active animation setup id
	 */
	public String getAnimation() {
		return animationSetup == null?"none":animationSetup.getId();
	}

	/**
	 * Returns current animation time 
	 * @return 0.0 <= time <= 1.0
	 */
	public float getAnimationTime() {
		return animationTime;
	}

	/**
	 * Returns transformation matrix for given group
	 * @param group id
	 * @return transformation matrix or null
	 */
	public Matrix3D getTransformationsMatrix(String id) {
		return transformationsMatrices.get(id);
	}

	/**
	 * Calculates all groups transformation matrices
	 * @param groups
	 * @param parentTransformationsMatrix
	 */
	protected void computeTransformationsMatrices(Hashtable<String, Group> groups, Matrix3D parentTransformationsMatrix, long animationLastAtTime, long animationCurrentAtTime) {
		// iterate through groups
		Iterator<Map.Entry<String, Group>> i = groups.entrySet().iterator();
		while (i.hasNext()) {
			Map.Entry<String, Group> entry = i.next();
			Group group = entry.getValue();

			// group transformation matrix
			Matrix3D transformationsMatrix = null;

			// compute animation matrix if animation setups exist
			Animation animation = group.getAnimation();
			if (animation != null && animationFinished == false) {
				Matrix3D[] animationMatrices = animation.getTransformationsMatrices();
				int frames = animationSetup.getFrames();
				float fps = animation.getFPS();

				// determine current and last matrix
				float frameAtLast = (animationLastAtTime / 1000f) * fps;
				float frameAtCurrent = (animationCurrentAtTime / 1000f) * fps;

				// check if looping is disabled
				if (animationSetup.isLoop() == false && frameAtCurrent >= frames) {
					// set frame at current to last frame
					frameAtLast = frames - 1;
					frameAtCurrent = frames - 1;
					animationFinished = true;
				}

				//
				int matrixAtLast = ((int)frameAtLast % frames);
				int matrixAtCurrent = ((int)frameAtCurrent % frames);
				animationTime = frames <= 1?0.0f:(float)matrixAtCurrent / (float)(frames - 1); 

				// compute animation transformations matrix
				float t = frameAtCurrent - (float)Math.floor(frameAtLast);
				if (t < 1f) {
					if (matrixAtLast == matrixAtCurrent) {
						matrixAtCurrent = ((matrixAtCurrent + 1) % frames);
					}
					transformationsMatrix = Matrix3D.interpolateLinear(
						animationMatrices[matrixAtLast + animationSetup.getStartFrame()],
						animationMatrices[matrixAtCurrent + animationSetup.getStartFrame()],
						t					
					);
				} else {
					transformationsMatrix = animationMatrices[matrixAtCurrent + animationSetup.getStartFrame()].clone();
				}
			}

			// do we have no animation matrix?
			if (transformationsMatrix == null) {
				// no animation matrix, set up local transformation matrix up as group matrix
				transformationsMatrix = group.getTransformationsMatrix().clone();
			} else {
				// we have animation matrix, so multiply it with group transformation matrix
				transformationsMatrix.multiply(group.getTransformationsMatrix());
			}

			// apply parent transformation matrix 
			if (parentTransformationsMatrix != null) {
				transformationsMatrix.multiply(parentTransformationsMatrix);
			}

			// put and associate transformation matrices with group
			transformationsMatrices.put(group.getId(), transformationsMatrix);

			// calculate sub groups
			Hashtable<String,Group> subGroups = group.getSubGroups(); 
			if (subGroups.size() > 0) {
				computeTransformationsMatrices(
					subGroups,
					transformationsMatrix,
					animationLastAtTime,
					animationCurrentAtTime
				);
			}
		}
	}

	/**
	 * Retrieves complete list of face triangles for all render groups 
	 * @return faces
	 */
	public Triangle[] getFaceTriangles() {
		Vector<Triangle> triangles = new Vector<Triangle>();
		for (Object3DGroup object3DGroup: object3dGroups) {
			Vector3D[] groupVerticesTransformed = object3DGroup.mesh.transformedVertices;
			for (FacesEntity facesEntity: object3DGroup.group.getFacesEntities())
			for (Face face: facesEntity.getFaces()) {
				int[] faceVertexIndices = face.getVertexIndices();
				triangles.add(
					new Triangle(
						groupVerticesTransformed[faceVertexIndices[0]].clone(),
						groupVerticesTransformed[faceVertexIndices[1]].clone(),
						groupVerticesTransformed[faceVertexIndices[2]].clone()
					)
				);
			}
		}
		Triangle[] triangleArray = new Triangle[triangles.size()];
		triangles.copyInto(triangleArray);
		return triangleArray;
	}

	/**
	 * Returns object3d group mesh object
	 * @param group id
	 * @return object3d group mesh object
	 */
	public Object3DGroupMesh getMesh(String groupId) {
		// TODO: maybe rather use a hashtable than an array to have a faster access
		for (Object3DGroup object3DGroup: object3dGroups) {
			if (object3DGroup.group.getId().equals(groupId)) {
				return object3DGroup.mesh;
			}
		}
		return null;
	}

}
