package net.drewke.tdme.engine;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import net.drewke.tdme.assets.textures.Texture;
import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.model.FacesEntity;
import net.drewke.tdme.model.Group;
import net.drewke.tdme.model.Material;
import net.drewke.tdme.model.Model;

/**
 * Object 3d render group 
 * @author Andreas Drewke
 * @version $Id: 06c7635f681b9d55b5f0dcfa84330420f13f3da4 $
 */
public class Object3DGroup {

	private static long counter = 0;
	final protected static int GLTEXTUREID_NONE = -1; 
	final protected static int GLTEXTUREID_NOTUSED = 0;

	protected String id;
	protected Group group;
	protected boolean animated;
	protected String textureId = null;
	protected int[] textureGlIdsByEntities = null;
	protected Object3DGroupRenderer renderer;
	protected Object3DGroupMesh mesh;

	/**
	 * Creates object 3d groups from given object3d base object
	 * @param object 3d base
	 * @return object 3d group array
	 */
	protected static Object3DGroup[] createGroups(Object3DBase object, boolean useMeshManager) {
		Vector<Object3DGroup> object3DGroups = new Vector<Object3DGroup>();
		Model model = object.getModel();
		createGroups(object3DGroups, model.getSubGroups(), model.hasAnimations(), useMeshManager);
		Object3DGroup[] object3DGroupArray = new Object3DGroup[object3DGroups.size()];
		object3DGroups.toArray(object3DGroupArray);
		return object3DGroupArray;
	}

	/**
	 * Creates a object 3d groups recursively for given group and it sub groups 
	 * @param object 3d group vector
	 * @param groups
	 * @param animated
	 */
	private static void createGroups(Vector<Object3DGroup> object3DGroups, Hashtable<String, Group> groups, boolean animated, boolean useMeshManager) {
		Engine engine = Engine.getInstance();
		Iterator<Map.Entry<String, Group>> i = groups.entrySet().iterator();
		while (i.hasNext()) {
			Map.Entry<String, Group> entry = i.next();
			Group group = entry.getValue();

			// skip on joints
			if (group.isJoint() == true) {
				continue;
			}

			// determine face count
			int faceCount = group.getFaceCount();

			// skip on groups without faces
			if (faceCount > 0) {

				// create group render data
				Object3DGroup object3DGroup = new Object3DGroup();
	
				// add it to group render data list
				object3DGroups.addElement(object3DGroup);
	
				//
				object3DGroup.id = group.getModel().getId() + ":" + group.getId();
				if (animated == true) object3DGroup.id+= ":" + (counter++);
	
				object3DGroup.group = group;
				object3DGroup.animated = animated;
				if (useMeshManager == true) {
					MeshManager meshManager = Engine.getInstance().getMeshManager();
					object3DGroup.mesh = meshManager.getMesh(object3DGroup.id);
					if (object3DGroup.mesh == null) {
						object3DGroup.mesh = Object3DGroupMesh.createMesh(group);
						meshManager.addMesh(object3DGroup.id, object3DGroup.mesh);
					}
				} else {
					object3DGroup.mesh = Object3DGroupMesh.createMesh(group);
				}
	
				object3DGroup.textureGlIdsByEntities = new int[group.getFacesEntities().length];
				for (int j = 0; j < object3DGroup.textureGlIdsByEntities.length; j++) {
					object3DGroup.textureGlIdsByEntities[j] = GLTEXTUREID_NONE;
				}
	
				// determine renderer
				if (engine.renderer.isBufferObjectsAvailable() == true) {
					object3DGroup.renderer = new Object3DGroupVBORenderer(object3DGroup);				
				} else {
					object3DGroup.renderer = new Object3DGroupDrawArrayRenderer(object3DGroup);
				}
			}

			// but still check sub groups
			createGroups(object3DGroups, group.getSubGroups(), animated, useMeshManager);
		}
	}

	/**
	 * Renders given object 3d groups
	 * @param gl
	 * @param object 3d groups
	 * @param recreate buffers
	 */
	protected static void renderGroups(Object3DGroup[] object3DGroups) {
		for (Object3DGroup object3DGroup: object3DGroups) {
			//
			object3DGroup.renderer.render();
		}
	}

	/**
	 * Applies transformations to meshes for given object 3d groups
	 * @param group render data list
	 * @param transformation matrices
	 */
	protected static void computeTransformations(Object3DGroup[] object3DGroups, Hashtable<String, Matrix3D> transformationMatrices) {
		for (Object3DGroup object3DGroup: object3DGroups) {
			object3DGroup.mesh.computeTransformations(
				object3DGroup.group,
				transformationMatrices
			);
		}
	}

	/**
	 * Set ups a material for rendering
	 * @param gl
	 * @param material
	 */
	protected static void setupMaterial(Object3DGroup object3DGroup, int facesEntityIdx) {
		FacesEntity[] facesEntities = object3DGroup.group.getFacesEntities(); 
		Material material = facesEntities[facesEntityIdx].getMaterial();

		// get material or use default
		if (material == null) material = Material.getDefaultMaterial();

		// load texture
		if (object3DGroup.textureGlIdsByEntities[facesEntityIdx] == GLTEXTUREID_NONE) {
			if (material.getDiffuseTexture() != null) {
				//
				Texture diffuseTexture = material.getDiffuseTexture();
				object3DGroup.textureId = diffuseTexture.getId(); 
				object3DGroup.textureGlIdsByEntities[facesEntityIdx] = Engine.getInstance().getTextureManager().addTexture(diffuseTexture);
			} else {
				object3DGroup.textureId = null; 
				object3DGroup.textureGlIdsByEntities[facesEntityIdx] = GLTEXTUREID_NOTUSED;
			}
		}

		// apply materials
		Engine.renderer.setMaterialAmbient(material.getAmbientColor().getArray());
		Engine.renderer.setMaterialDiffuse(material.getDiffuseColor().getArray());
		Engine.renderer.setMaterialSpecular(material.getSpecularColor().getArray());
		Engine.renderer.setMaterialEmission(material.getEmissionColor().getArray());
		Engine.renderer.setMaterialShininess(material.getShininess());
		Engine.renderer.updateMaterial();

		// bind texture
		Engine.renderer.bindTexture(object3DGroup.textureGlIdsByEntities[facesEntityIdx]);
	}

	/**
	 * Clear material for rendering
	 * @param gl
	 * @param material
	 */
	protected static void clearMaterial() {
		// unbind texture
		Engine.renderer.bindTexture(Engine.renderer.ID_NONE);
	}

}