package net.drewke.tdme.engine;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.util.Arrays;
import java.util.Hashtable;

import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.model.Face;
import net.drewke.tdme.model.FacesEntity;
import net.drewke.tdme.model.Group;
import net.drewke.tdme.model.Joint;
import net.drewke.tdme.model.JointWeight;
import net.drewke.tdme.model.Skinning;
import net.drewke.tdme.model.TextureCoordinate;

public class Object3DGroupMesh {

	private int faces;

	protected Vector3D transformedVertices[] = null;
	protected Vector3D transformedNormals[] = null;

	private boolean recreatedBuffers;
	protected FloatBuffer fbVertices = null;
	protected FloatBuffer fbNormals = null;
	protected FloatBuffer fbTextureCoordinates = null;

	/**
	 * Creates a object3d group mesh from group 
	 * @param group
	 */
	protected static Object3DGroupMesh createMesh(Group group) {
		Object3DGroupMesh mesh = new Object3DGroupMesh();

		// group data
		Vector3D[] groupVertices = group.getVertices();
		Vector3D[] groupNormals = group.getNormals();
		TextureCoordinate[] groupTextureCoordinates = group.getTextureCoordinates();

		// determine face count
		int faceCount = group.getFaceCount();

		// set up face count
		mesh.faces = faceCount;

		// transformed mesh vertices
		mesh.transformedVertices = new Vector3D[groupVertices.length];
		for(int j = 0; j < mesh.transformedVertices.length; j++) {
			mesh.transformedVertices[j] = new Vector3D().set(groupVertices[j]);
		}

		// transformed mesh normals
		mesh.transformedNormals = new Vector3D[groupNormals.length];
		for(int j = 0; j < mesh.transformedNormals.length; j++) {
			mesh.transformedNormals[j] = new Vector3D().set(groupNormals[j]);
		}

		mesh.recreatedBuffers = false;

		mesh.fbVertices = ByteBuffer.allocateDirect(mesh.faces * 3 * 3 * Float.SIZE / Byte.SIZE).order(ByteOrder.nativeOrder()).asFloatBuffer();
		mesh.fbNormals = ByteBuffer.allocateDirect(mesh.faces * 3 * 3 * Float.SIZE / Byte.SIZE).order(ByteOrder.nativeOrder()).asFloatBuffer();
		mesh.fbTextureCoordinates = groupTextureCoordinates == null?null:ByteBuffer.allocateDirect(mesh.faces * 3 * 2 * Float.SIZE / Byte.SIZE).order(ByteOrder.nativeOrder()).asFloatBuffer();

		// create texture coordinates buffer, will never be changed in engine
		if (mesh.fbTextureCoordinates != null) {
			// construct texture coordinates byte buffer as this will not change usually
			for (FacesEntity facesEntity: group.getFacesEntities()) {
				for (Face face: facesEntity.getFaces()) {
					int[] vertexIndices = face.getVertexIndices();
					int[] textureCoordinateIndices = face.getTextureCoordinateIndices();
					if (textureCoordinateIndices != null) {
						for (int j = 0; j < vertexIndices.length; j++) {
							mesh.fbTextureCoordinates.put(groupTextureCoordinates[textureCoordinateIndices[j]].getArray());
						}
					}
				}
			}
			mesh.fbTextureCoordinates.flip();
		}

		mesh.recreateBuffers(group);

		//
		return mesh;
	}

	/**
	 * Computes mesh transformations
	 * @param group
	 * @param transformation matrices
	 */
	protected void computeTransformations(Group group, Hashtable<String, Matrix3D> transformationMatrices) {
		Vector3D[] groupVertices = group.getVertices();
		int[][] groupVerticesNormalIndices = group.getVerticesNormalIndices();
		Vector3D[] groupNormals = group.getNormals();

		// transformations for skinned meshes
		Skinning skinning = group.getSkinning();
		if (skinning != null) {
			for (int vertexIndex = 0; vertexIndex < groupVertices.length; vertexIndex++) {
				// do vertices
				Vector3D vertex = groupVertices[vertexIndex];
				Vector3D _vertex = transformedVertices[vertexIndex].set(0f,0f,0f);

				// and normals
				int[] vertexNormalIndices = groupVerticesNormalIndices[vertexIndex];
				Vector3D[] normals = new Vector3D[vertexNormalIndices.length];
				Vector3D[] _normals = new Vector3D[vertexNormalIndices.length];
				for (int k = 0; k < _normals.length; k++) {
					normals[k] = groupNormals[vertexNormalIndices[k]];
					_normals[k] = transformedNormals[vertexNormalIndices[k]].set(0f,0f,0f);
				}

				// compute every influence on vertex and vertex normals
				float totalWeights = 0f;
				Joint[] joints = skinning.getJoints();
				float[] weights = skinning.getWeights();
				JointWeight[][] jointsWeights = skinning.getVerticesJointsWeights();
				for (JointWeight jointWeight : jointsWeights[vertexIndex]) {
					Joint joint = joints[jointWeight.getJointIndex()];
					float weight = weights[jointWeight.getWeightIndex()];

					Matrix3D transformationsMatrix = joint.getBindMatrix().clone();
					transformationsMatrix.multiply(transformationMatrices.get(joint.getGroupId()));

					// vertex
					_vertex.add(transformationsMatrix.multiply(vertex).scale(weight));

					// normals
					for (int k = 0; k < _normals.length; k++) {
						_normals[k].add(transformationsMatrix.multiplyNoTranslation(normals[k]).scale(weight));
					}

					//
					totalWeights += weight;
				}

				// scale to full weight
				if (totalWeights != 1f) {
					float weightNormalized = 1f / totalWeights;

					// vertex
					_vertex.scale(weightNormalized);

					// normals
					for (int k = 0; k < _normals.length; k++) {
						_normals[k].scale(weightNormalized);
					}
				}

				// normalize normals
				for (int k = 0; k < _normals.length; k++) {
					transformedNormals[vertexNormalIndices[k]].normalize();
				}
			}
		} else {
			// transformations for non skinned rendering
			Matrix3D transformationsMatrix = transformationMatrices.get(group.getId());
			for (int vertexIndex = 0; vertexIndex < groupVertices.length; vertexIndex++) {
				transformedVertices[vertexIndex].set(transformationsMatrix.multiply(groupVertices[vertexIndex]));
			}
			for (int normalIndex = 0; normalIndex < groupNormals.length; normalIndex++) {
				transformedNormals[normalIndex].set(transformationsMatrix.multiplyNoTranslation(groupNormals[normalIndex]).normalize());
			}
		}

		// recreate buffers
		recreateBuffers(group);
	}

	/**
	 * Recreates group float buffers
	 * @param group
	 */
	protected void recreateBuffers(Group group) {
		// flip buffers for reusage
		fbVertices.clear();
		fbNormals.clear();

		// (re)create buffers
		for (FacesEntity facesEntity: group.getFacesEntities()) {
			for (Face face: facesEntity.getFaces()) {
				int[] vertexIndices = face.getVertexIndices();
				int[] normalIndices = face.getNormalIndices();
				for (int j = 0; j < vertexIndices.length; j++) {
					fbVertices.put(transformedVertices[vertexIndices[j]].getArray());
					fbNormals.put(transformedNormals[normalIndices[j]].getArray());
				}
			}
		}

		// flip buffers
		fbVertices.flip();
		fbNormals.flip();

		//
		recreatedBuffers = true;
	}

	/**
	 * @return if buffers has been recreated and unsets state
	 */
	protected boolean hasRecreatedBuffers() {
		if (recreatedBuffers == true) {
			recreatedBuffers = false;
			return true;
		} else {
			return false;
		}
	}

	/**
	 * @return transformed vertices
	 */
	public Vector3D[] getTransformedVertices() {
		return transformedVertices;
	}

	/**
	 * @return transformed normals
	 */
	public Vector3D[] getTransformedNormals() {
		return transformedNormals;
	}

}