package net.drewke.tdme.engine;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.math.Vector3D;

/**
 * Shadow map class
 * @author Andreas Drewke
 * @version $Id: 76815d7d1773a4d4002028afa0d09cf19b84249a $
 */
public class ShadowMap {

	protected final static int TEXTUREUNIT = 1;

	private int shadowMappingRatio = 4;

	private ShadowMapping shadowMapping;

	private Camera lightCamera;
	private FrameBuffer frameBuffer;
	private Matrix3D depthBiasMVPMatrix;

	/**
	 * Public constructor
	 * @param width
	 * @param height
	 */
	protected ShadowMap(ShadowMapping shadowMapping, int width, int height) {
		this.shadowMapping = shadowMapping;
		lightCamera = new Camera();
		frameBuffer = new FrameBuffer(
			width * shadowMappingRatio,
			height * shadowMappingRatio
		);
		depthBiasMVPMatrix = new Matrix3D().identity();
	}

	/**
	 * @return width
	 */
	public int getWidth() {
		return frameBuffer.getWidth();
	}

	/**
	 * @return height
	 */
	public int getHeight() {
		return frameBuffer.getHeight();
	}

	/**
	 * Init frame buffer
	 * @param gl
	 */
	protected void init() {
		frameBuffer.init();
	}

	/**
	 * Reshape frame buffer
	 * @param gl
	 * @param width
	 * @param height
	 */
	protected void reshape(int width, int height) {
		frameBuffer.reshape(width * shadowMappingRatio, height * shadowMappingRatio);
	}

	/**
	 * Disposes this shadow mapping
	 * @param gl
	 */
	protected void dispose() {
		frameBuffer.dispose();
	}

	/**
	 * Binds frame buffer depth texture
	 * @param gl
	 */
	protected void bindDepthTexture() {
		frameBuffer.bindDepthTexture();
	}

	/**
	 * @return lightCamera
	 */
	protected Camera getCamera() {
		return lightCamera;
	}

	/**
	 * Renders given objects to shadow map
	 * @param gl
	 * @param objects
	 */
	protected void render(Light light, Hashtable<String,Object3D> objects) {
		// TODO: Spotlights are different!

		// viewers camera
		Camera camera = Engine.getInstance().getCamera();
		float lookLength = camera.getLookAt().clone().sub(camera.getLookFrom()).computeLength() * 4;

		// compute camera from view of light 
		Vector3D lightPosition = new Vector3D(light.getPosition().getArray());
		Vector3D lightLookAt = new Vector3D(0f,0f,0f);
		Vector3D lightDirection = lightLookAt.clone().sub(lightPosition).normalize();
		Vector3D lightTranslationIntoCameraView = camera.getLookAt().clone().sub(lightLookAt);		
		lightLookAt.add(lightTranslationIntoCameraView);
		Vector3D lightLookFrom = lightLookAt.clone().sub(lightDirection.clone().scale(lookLength));

		// set up light camera from view of light
		lightCamera.setZNear(10f);
		lightCamera.setZFar(camera.getZFar());
		lightCamera.getLookFrom().set(lightLookFrom);
		lightCamera.getLookAt().set(lightLookAt);
		lightCamera.getUpVector().set(Camera.computeUpVector(lightCamera.getLookFrom(), lightCamera.getLookAt()));
		lightCamera.updateCamera(frameBuffer.getWidth(), frameBuffer.getHeight());

		// Bind frame buffer to shadow map fbo id
		frameBuffer.enableFrameBuffer();
		
		// disable color rendering, we only want to write to the Z-Buffer
		Engine.renderer.setColorMask(false, false, false, false);

		// clear depth buffer
		Engine.renderer.clear(Engine.renderer.CLEAR_DEPTH_BUFFER_BIT);

		// render backfaces only, avoid self-shadowing
		Engine.renderer.setCullFace(Engine.renderer.CULLFACE_FRONT);

		// render
		Iterator<Map.Entry<String, Object3D>> i = objects.entrySet().iterator();
		while (i.hasNext()) {
			Map.Entry<String, Object3D> entry = i.next();
			Object3D object = entry.getValue();

			// check if disabled or not enabled for dynamic shadowing
			if (object.isEnabled() == false || object.isDynamicShadowingEnabled() == false) {
				continue;
			}

			// draw if visible
			if (lightCamera.getFrustum().isVisible(object.getBoundingBoxTransformed())) {
				object.render();
			}
		}

		// unbind frame buffer
		FrameBuffer.disableFrameBuffer();

		// restore disable color rendering
		Engine.renderer.setColorMask(true, true, true, true);

		// restore render backfaces only
		Engine.renderer.setCullFace(Engine.renderer.CULLFACE_BACK);

		// generate shadow map texture matrix
		computeDepthBiasMVPMatrix();
	}

	/**
	 * Computes shadow texture matrix and stores it
	 * @param gl
	 */
	protected void computeDepthBiasMVPMatrix() {
		// matrices
		Matrix3D modelViewMatrix = Engine.renderer.getModelViewMatrix();
		Matrix3D projectionMatrix = Engine.renderer.getProjectionMatrix();
		Matrix3D biasMatrix = new Matrix3D(	
			0.5f, 0.0f, 0.0f, 0.0f, 
			0.0f, 0.5f, 0.0f, 0.0f,
			0.0f, 0.0f, 0.5f, 0.0f,
			0.5f, 0.5f, 0.5f, 1.0f
		);

		// compute shadow texture matrix
		depthBiasMVPMatrix.set(
			modelViewMatrix.
			clone().
			multiply(projectionMatrix).
			multiply(biasMatrix)
		);
	}

	/**
	 * Set up shadow texture matrix computed and stored before
	 */
	protected void updateDepthBiasMVPMatrix() {
		shadowMapping.updateDepthBiasMVPMatrix(depthBiasMVPMatrix);
	}

}