package net.drewke.tdme.engine;

import java.util.Stack;
import java.util.Vector;

import net.drewke.tdme.math.Matrix3D;

/**
 * Shadow mapping
 * @author Andreas Drewke
 * @version $Id: d528f17565c1a0b287bdcb03dd95a1eccb790709 $
 */
public class ShadowMapping {

	private enum RunState {NONE, PRE, RENDER};

	private Stack<Matrix3D> shadowTransformationsMatrices;
	private Matrix3D depthBiasMVPMatrix;

	private ShadowMap[] shadowMaps;

	private int width;
	private int height;

	private RunState runState;

	private int shadowMappingPreVertexShaderGlId;
	private int shadowMappingPreFragmentShaderGlId;
	private int shadowMappingPreUniformObjectToScreenMatrix;
	private int shadowMappingPreProgramGlId;
	private int shadowMappingRenderVertexShaderGlId;
	private int shadowMappingRenderFragmentShaderGlId;
	private int shadowMappingRenderProgramGlId;
	private int shadowMappingRenderUniformTextureUnit;
	private int shadowMappingRenderUniformTexturePixelWidth;
	private int shadowMappingRenderUniformTexturePixelHeight;
	private int shadowMappingRenderUniformDepthBiasMVPMatrix;
	private int shadowMappingRenderUniformObjectToScreenMatrix;

	/**
	 * Constructor
	 */
	protected ShadowMapping() {
		width = 0;
		height = 0;
		Light[] lights = Engine.getInstance().getLights();
		shadowMaps = new ShadowMap[8];
		for (int i = 0; i < lights.length; i++) {
			shadowMaps[i] = null;
		}
		shadowTransformationsMatrices = new Stack<Matrix3D>();
		depthBiasMVPMatrix = new Matrix3D().identity();
		runState = RunState.NONE;
	}

	/**
	 * Init shadow mapping
	 */
	protected void init() {
		GLRenderer renderer = Engine.renderer;
		String rendererVersion = renderer instanceof GL3Renderer?"gl3":"gl2";
		// load shadow mapping shaders
		//	pre render
		shadowMappingPreVertexShaderGlId = renderer.loadShader(
			renderer.SHADER_VERTEX_SHADER,
			"shader/" + rendererVersion + "/shadowmapping",
			"pre_vertexshader.c"
		);
		shadowMappingPreFragmentShaderGlId = renderer.loadShader(
			renderer.SHADER_FRAGMENT_SHADER,
			"shader/" + rendererVersion + "/shadowmapping",
			"pre_fragmentshader.c"
		);
		//	render
		shadowMappingRenderVertexShaderGlId = renderer.loadShader(
			renderer.SHADER_VERTEX_SHADER,
			"shader/" + rendererVersion + "/shadowmapping",
			"render_vertexshader.c"
		);
		shadowMappingRenderFragmentShaderGlId = renderer.loadShader(
			renderer.SHADER_FRAGMENT_SHADER,
			"shader/" + rendererVersion + "/shadowmapping",
			"render_fragmentshader.c"
		);

		// create shadow mapping render program
		//	pre
		shadowMappingPreProgramGlId = renderer.createProgram();
		renderer.attachShaderToProgram(shadowMappingPreProgramGlId, shadowMappingPreVertexShaderGlId);
		renderer.attachShaderToProgram(shadowMappingPreProgramGlId, shadowMappingPreFragmentShaderGlId);
		renderer.linkProgram(shadowMappingPreProgramGlId);

		//	uniforms
		shadowMappingPreUniformObjectToScreenMatrix = renderer.getProgramUniformLocation(shadowMappingPreProgramGlId, "objectToScreenMatrix");

		//	render
		shadowMappingRenderProgramGlId = renderer.createProgram();

		renderer.attachShaderToProgram(shadowMappingRenderProgramGlId, shadowMappingRenderVertexShaderGlId);
		renderer.attachShaderToProgram(shadowMappingRenderProgramGlId, shadowMappingRenderFragmentShaderGlId);
		renderer.linkProgram(shadowMappingRenderProgramGlId);

		// uniforms
		shadowMappingRenderUniformTextureUnit = renderer.getProgramUniformLocation(shadowMappingRenderProgramGlId, "textureUnit");
		shadowMappingRenderUniformTexturePixelWidth = renderer.getProgramUniformLocation(shadowMappingRenderProgramGlId, "texturePixelWidth");
		shadowMappingRenderUniformTexturePixelHeight = renderer.getProgramUniformLocation(shadowMappingRenderProgramGlId, "texturePixelHeight");
		shadowMappingRenderUniformDepthBiasMVPMatrix = renderer.getProgramUniformLocation(shadowMappingRenderProgramGlId, "depthBiasMVPMatrix");
		shadowMappingRenderUniformObjectToScreenMatrix = renderer.getProgramUniformLocation(shadowMappingRenderProgramGlId, "objectToScreenMatrix");
	}

	/**
	 * Reshape shadow maps
	 * @param width
	 * @param height
	 */
	protected void reshape(int width, int height) {
		this.width = width;
		this.height = height;
		//
		for (int i = 0; i < shadowMaps.length; i++) {
			if (shadowMaps[i] != null) shadowMaps[i].reshape(width, height);
		}
	}

	/**
	 * Create shadow maps
	 */
	protected void createShadowMaps() {
		runState = RunState.PRE;

		GLRenderer renderer = Engine.renderer;

		// Use shadow mapping "pre programm"
		renderer.useProgram(shadowMappingPreProgramGlId);

		// render to shadow maps
		Light[] lights = Engine.getInstance().getLights();
		for (int i = 0; i < lights.length; i++) {
			Light light = lights[i];
			if (light.isEnabled()) {
				// create shadow map for light, if required
				if (shadowMaps[i] == null) {
					ShadowMap shadowMap = new ShadowMap(this, width, height);
					shadowMap.init();
					shadowMaps[i] = shadowMap;
				}
				// render
				//	TODO: do not create shadow map for all objects
				shadowMaps[i].render(lights[i], Engine.getInstance().getObjects());
			} else {
				// dispose shadow map
				if (shadowMaps[i] != null) {
					shadowMaps[i].dispose();
					shadowMaps[i] = null;
				}
			}
		}

		// do not use program
		renderer.useProgram(renderer.ID_NONE);

		// update camera
		Engine.getInstance().getCamera().updateCamera(width, height);

		//
		runState = RunState.NONE;
	}

	/**
	 * Render shadow maps
	 * @param visible objects
	 */
	protected void renderShadowMaps(Vector<Object3D> visibleObjects) {
		runState = RunState.RENDER;

		//
		GLRenderer renderer = Engine.renderer;

		// render using shadow mapping program
		renderer.useProgram(shadowMappingRenderProgramGlId);
		renderer.setProgramUniformInteger(shadowMappingRenderUniformTextureUnit, ShadowMap.TEXTUREUNIT);

		//	do not do blending
		renderer.enableBlending();
		//	do not allow writing to depth buffer
		renderer.disableDepthBuffer();
		//	only process nearest fragments
		renderer.setDepthFunction(renderer.DEPTHFUNCTION_EQUAL);

		// render each shadow map
		for (int i = 0; i < shadowMaps.length; i++) {
			// skip on unused shadow mapping
			if (shadowMaps[i] == null) continue;

			ShadowMap shadowMap = shadowMaps[i];

			renderer.setProgramUniformFloat(shadowMappingRenderUniformTexturePixelWidth, 1.0f / (float)shadowMap.getWidth());
			renderer.setProgramUniformFloat(shadowMappingRenderUniformTexturePixelHeight, 1.0f / (float)shadowMap.getHeight());

			// setup shadow texture matrix
			shadowMap.updateDepthBiasMVPMatrix();

			// bind shadow map texture on shadow map texture unit
			renderer.setTextureUnit(ShadowMap.TEXTUREUNIT);
			shadowMap.bindDepthTexture();
			// switch back to texture0 unit
			renderer.setTextureUnit(0);

			// render objects
			for (Object3D object: visibleObjects) {	
				// draw if visible
				object.render();
			}
		}

		// restore gl texture matrix on texture 7
		renderer.setTextureUnit(ShadowMap.TEXTUREUNIT);
		renderer.bindTexture(renderer.ID_NONE);
		renderer.setTextureUnit(renderer.ID_NONE);

		// restore gl defaults
		renderer.disableBlending();
		renderer.enableDepthBuffer();
		renderer.setDepthFunction(renderer.DEPTHFUNCTION_LESSEQUAL);
		renderer.useProgram(renderer.ID_NONE);

		//
		runState = RunState.NONE;
	}

	/**
	 * Dispose shadow maps
	 */
	protected void dispose() {
		// dispose shadow mappings
		for (int i = 0; i < shadowMaps.length; i++) {
			if (shadowMaps[i] != null) {
				shadowMaps[i].dispose();
				shadowMaps[i] = null;
			}
		}
	}

	/**
	 * Start object transformations
	 * @param gl
	 * @param transformations matrix
	 */
	protected void startObjectTransformations(Matrix3D transformationsMatrix) {
		if (runState == RunState.NONE) return;

		// retrieve current model view matrix and put it on stack
		shadowTransformationsMatrices.push(
			depthBiasMVPMatrix.clone()
		);

		// set up new model view matrix
		depthBiasMVPMatrix.set(
			transformationsMatrix.clone().
			multiply(
				depthBiasMVPMatrix
			)
		);

		//
		updateDepthBiasMVPMatrix();
	}

	/**
	 * End object transformations
	 * @param gl
	 */
	protected void endObjectTransformations() {
		if (runState == RunState.NONE) return;

		// set up new model view matrix
		depthBiasMVPMatrix.set(shadowTransformationsMatrices.pop());

		//
		updateDepthBiasMVPMatrix();
	}

	/**
	 * Update model view and projection matrix
	 */
	protected void updateMVPMatrices(GLRenderer renderer) {
		if (runState == RunState.NONE ||
			shadowMappingPreUniformObjectToScreenMatrix == -1 ||
			shadowMappingRenderUniformObjectToScreenMatrix == -1) return;

		// object to screen matrix
		Matrix3D objectToScreenMatrix = 
			renderer.getModelViewMatrix().clone().
			multiply(
				renderer.getProjectionMatrix()
			);

		// upload
		switch (runState) {
			case PRE: {
				renderer.setProgramUniformFloatMatrix4x4(shadowMappingPreUniformObjectToScreenMatrix, objectToScreenMatrix.getArray());
				break;
			}
			case RENDER: {
				renderer.setProgramUniformFloatMatrix4x4(shadowMappingRenderUniformObjectToScreenMatrix, objectToScreenMatrix.getArray());
				break;
			}
			default: {
				System.out.println("ShadowMapping::updateMVPMatrices(): unsupported run state '" + runState + "'");
				break;
			}
		}

		// upload
		
	}

	/**
	 * Update depth bias mvp matrix with given matrix
	 */
	protected void updateDepthBiasMVPMatrix(Matrix3D depthBiasMVPMatrix) {
		if (runState != RunState.RENDER ||
			shadowMappingRenderUniformDepthBiasMVPMatrix == -1) return;

		// copy matrix
		this.depthBiasMVPMatrix.set(depthBiasMVPMatrix);

		// upload
		Engine.renderer.setProgramUniformFloatMatrix4x4(shadowMappingRenderUniformDepthBiasMVPMatrix, depthBiasMVPMatrix.getArray());
	}

	/**
	 * Update depth bias mvp matrix / upload only
	 */
	protected void updateDepthBiasMVPMatrix() {
		if (runState != RunState.RENDER ||
			shadowMappingRenderUniformDepthBiasMVPMatrix == -1) return;

		// upload
		Engine.renderer.setProgramUniformFloatMatrix4x4(shadowMappingRenderUniformDepthBiasMVPMatrix, depthBiasMVPMatrix.getArray());
	}

}
