package net.drewke.tdme.environment;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import net.drewke.tdme.engine.ModelUtilities;
import net.drewke.tdme.engine.Object3DModel;
import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.model.Model;
import net.drewke.tdme.physics.CollisionDetection;
import net.drewke.tdme.primitives.BoundingBox;
import net.drewke.tdme.primitives.Triangle;

/**
 * Environment collision information
 * @author Andreas Drewke
 * @version $Id: 6b7d8bc67abce341903eabfde79a2c3efc2bba26 $
 */
public class CollisionInformation {

	private float partitionXSize;
	private float partitionZSize;
	private Object3DModel object3DModel;
	private BoundingBox boundingBox;
	private Hashtable<String, Triangle[]> heightTriangles;

	/**
	 * Collision information computing thread
	 * @author Andreas Drewke
	 * @version $Id: 6b7d8bc67abce341903eabfde79a2c3efc2bba26 $
	 */
	private class CollisionInformationComputingThread extends Thread {

		private int threadIdx;
		private Triangle[] triangles;
		private int triangleIdxMin;
		private int triangleIdxMax;
		private Hashtable<String, BoundingBox> partitionBoundingBoxes;
		private Hashtable<String, Vector<Triangle>> heightTriangles;

		/**
		 * Constructor
		 * @param thread ids
		 * @param group
		 * @param triangle index min
		 * @param triangle index max
		 */
		public CollisionInformationComputingThread(int threadIdx, Hashtable<String, BoundingBox> partitionBoundingBoxes, Triangle[] triangles, int triangleIdxMin, int triangleIdxMax) {
			this.partitionBoundingBoxes = partitionBoundingBoxes;
			this.heightTriangles = new Hashtable<String, Vector<Triangle>>();
			this.threadIdx = threadIdx;
			this.triangles = triangles;
			this.triangleIdxMin = triangleIdxMin;
			this.triangleIdxMax = triangleIdxMax;
		}

		/**
		 * @return height triangles by partition
		 */
		public Hashtable<String, Vector<Triangle>> getHeightTriangles() {
			return heightTriangles;
		}

		/*
		 * (non-Javadoc)
		 * @see java.lang.Thread#run()
		 */
		public void run() {
			for (int triangleIdx = triangleIdxMin; triangleIdx < triangleIdxMax; triangleIdx++) {
				Triangle triangle = triangles[triangleIdx];

				// map triangle to partitions
				int trianglePartitionCount = 0;
				Iterator<Map.Entry<String, BoundingBox>> ibb = partitionBoundingBoxes.entrySet().iterator();
				while (ibb.hasNext()) {
					Map.Entry<String, BoundingBox> ibbEntry = ibb.next();
					String partitionId = ibbEntry.getKey();
					BoundingBox partitionBoundingBox = ibbEntry.getValue();
					if (CollisionDetection.doCollide(triangle, partitionBoundingBox) != null) {
						// get height triangle vector
						Vector<Triangle> _heightTriangles = heightTriangles.get(partitionId);
						// does not exist yet?
						if (_heightTriangles == null) {
							// nope, create it
							_heightTriangles = new Vector<Triangle>();
							heightTriangles.put(partitionId, _heightTriangles);
						}
						// add triangle
						_heightTriangles.add(triangle);
						//
						trianglePartitionCount++;
					}
				}
				if (trianglePartitionCount == 0) {
					System.out.println("[" + threadIdx + "] Warning: No collision: " + triangle);
				}
			}
		}
	}

	/**
	 * Public constructor
	 * @param model
	 * @param partition x size
	 * @param partition z size
	 */
	public CollisionInformation(Model model, float partitionXSize, float partitionZSize) {
		this.partitionXSize = partitionXSize;
		this.partitionZSize = partitionZSize;
		this.object3DModel = new Object3DModel(model);
		this.boundingBox = ModelUtilities.createBoundingBox(object3DModel);
		this.heightTriangles = new Hashtable<String, Triangle[]>();

		//
		computeCollisionInformation();
	}

	/**
	 * Public constructor
	 * @param object 3D model
	 * @param partition x size
	 * @param partition z size
	 */
	public CollisionInformation(Object3DModel object3DModel, float partitionXSize, float partitionZSize) {
		this.partitionXSize = partitionXSize;
		this.partitionZSize = partitionZSize;
		this.object3DModel = object3DModel;
		this.boundingBox = ModelUtilities.createBoundingBox(object3DModel);
		this.heightTriangles = new Hashtable<String, Triangle[]>();

		//
		computeCollisionInformation();
	}

	/**
	 * Computes collision information
	 */
	private void computeCollisionInformation() {
		// compute bounding box for each partition
		Hashtable<String, BoundingBox> partitionBoundingBoxes = new Hashtable<String, BoundingBox>();
		float[] minXYZ = boundingBox.getMin().getArray();
		float[] maxXYZ = boundingBox.getMax().getArray();
		for (float x = minXYZ[0]; x < maxXYZ[0]; x+= partitionXSize)
		for (float z = minXYZ[2]; z < maxXYZ[2]; z+= partitionZSize) {
			BoundingBox partitionBoundingBox = new BoundingBox();
			partitionBoundingBox.set(
				x,
				minXYZ[1],
				z,
				x + partitionXSize >= maxXYZ[0]?maxXYZ[0]:x + partitionXSize,
				maxXYZ[1],
				z + partitionZSize >= maxXYZ[2]?maxXYZ[2]:z + partitionZSize
			);

			// partition id
			String partitionId =
				(int)Math.floor(x / partitionXSize) +
				"," +
				(int)Math.floor(z / partitionZSize);

			// add partition bounding box
			partitionBoundingBoxes.put(
				partitionId,
				partitionBoundingBox
			);
		}

		// retrieve all triangles
		Triangle[] triangles = object3DModel.getFaceTriangles();
		int triangleCount = triangles.length;

		//
		CollisionInformationComputingThread[] workers = new CollisionInformationComputingThread[Runtime.getRuntime().availableProcessors()];
		int trianglesPerWorkers = (int)Math.ceil((float)triangleCount / (float)workers.length);
		int trianglesAssigned = 0;
		int workerStarted = 0;
		for (int j = 0; j < workers.length; j++) {
			int trianglesToProcess = trianglesPerWorkers;
			if (trianglesAssigned + trianglesToProcess > triangleCount) trianglesToProcess = triangleCount - trianglesAssigned;

			// skip on worker which has no triangles to process
			if (trianglesToProcess == 0) continue;

			// start worker
			workers[j] = new CollisionInformationComputingThread(
				j,
				partitionBoundingBoxes,
				triangles,
				trianglesAssigned,
				trianglesAssigned + trianglesToProcess
			);
			workers[j].start();
			trianglesAssigned+= trianglesToProcess;

			//
			workerStarted++;
		}

		// block until all workers finished
		for (int j = 0; j < workerStarted; j++) {
			while (true) {
				boolean interrupted = false;
				try { workers[j].join(); } catch (InterruptedException ie) { interrupted = true; }
				if (interrupted == false) break;
			}
		}

		// put together the collision information
		Iterator<Map.Entry<String, BoundingBox>> pbbi = partitionBoundingBoxes.entrySet().iterator();
		while (pbbi.hasNext()) {
			Map.Entry<String, BoundingBox> pbbiEntry = pbbi.next();
			String partitionId = pbbiEntry.getKey();

			// determine triangles for partition
			int partitionTriangleCount = 0;
			for (int j = 0; j < workerStarted; j++) {
				Vector<Triangle> workerPartitionTriangles = workers[j].getHeightTriangles().get(partitionId);
				if (workerPartitionTriangles == null) continue;
				partitionTriangleCount+= workerPartitionTriangles.size();
			}

			// skip on empty partitions
			if (partitionTriangleCount == 0) continue;

			// create height information for partition
			Triangle[] partitionTriangles = new Triangle[partitionTriangleCount];
			heightTriangles.put(partitionId, partitionTriangles);

			// add triangles
			int partitionTriangleIdx = 0;
			for (int j = 0; j < workerStarted; j++) {
				Vector<Triangle> workerPartitionTriangles = workers[j].getHeightTriangles().get(partitionId);

				// skip on empty partition of current worker
				if (workerPartitionTriangles == null) continue;

				// add triangles
				for (Triangle triangle: workerPartitionTriangles) {
					partitionTriangles[partitionTriangleIdx++] = triangle;
				}
			}
		}
	}

	/**
	 * Computes height of environment for given point 
	 * @param point
	 * @return height
	 */
	public float computeHeight(Vector3D point, float defaultYValue) {
		float[] pointXYZ = point.getArray();
		float x = pointXYZ[0];
		float z = pointXYZ[2];

		// determine partition
		String partitionId =
			(int)Math.floor(x / partitionXSize) + 
			"," + 
			(int)Math.floor(z / partitionZSize);

		// get triangles partition
		Triangle[] triangles = heightTriangles.get(partitionId);
		if (triangles == null) return defaultYValue;

		// determine best fitting triangle
		Vector3D bestFit = null;
		float minDistance = Float.MAX_VALUE;

		// iterate triangles
		Vector3D _pointY = point.clone();
		for (Triangle triangle: triangles) {
			Vector3D[] triangleVertices = triangle.getVertices();
			_pointY.setY(
				(
					triangleVertices[0].getY() +
					triangleVertices[1].getY() + 
					triangleVertices[2].getY()
				) / 3.0f
			);

			// compute closest point on triangle for point
			Vector3D closestPointOnTriangle = triangle.computeClosestPointOnBoundingVolume(_pointY);

			// skip on triangles that are on top of given point
			// if (closestPointOnTriangle.getY() > y) continue;

			// compute X,Y component distance of closest point on triangle to point
			float[] distanceVectorXYZ = closestPointOnTriangle.clone().sub(point).getArray();
			float distance = Math.abs(distanceVectorXYZ[0]) + Math.abs(distanceVectorXYZ[2]);

			// determine if better fit
			if (bestFit == null ||
				distance < minDistance) {
				bestFit = closestPointOnTriangle;
				minDistance = distance;
			}
		}

		//
		return bestFit == null?defaultYValue:bestFit.getY();
	}

}