package net.drewke.tdme.environment;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import net.drewke.tdme.engine.ModelUtilities;
import net.drewke.tdme.engine.Object3DGroupMesh;
import net.drewke.tdme.engine.Object3DModel;
import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.model.Face;
import net.drewke.tdme.model.FacesEntity;
import net.drewke.tdme.model.Group;
import net.drewke.tdme.model.Material;
import net.drewke.tdme.model.Model;
import net.drewke.tdme.model.TextureCoordinate;
import net.drewke.tdme.physics.CollisionDetection;
import net.drewke.tdme.primitives.BoundingBox;
import net.drewke.tdme.primitives.Triangle;

/**
 * Model partition logic
 * @author Andreas Drewke
 * @version $Id: 2aaf402c093246353a5d339d1a94cc828761b381 $
 */
public class Partition {

	/**
	 * Partition a model into smaller ones
	 * @param model 
	 * @param max size X
	 * @param max size Y
	 * @param max size Z
	 * @return partitions
	 */
	public static Vector<Model> computePartitions(Object3DModel environment, float partitionSizeX, float partitionSizeY, float partitionSizeZ) {
		Partition p = new Partition(
			environment,
			partitionSizeX,
			partitionSizeY,
			partitionSizeZ
		);
		if (p.computeDimensions()) {
			p.computePartitions();
		}
		return p.partitions;
	}

	/**
	 * Partition a model into smaller ones
	 * @param model 
	 * @param max size X
	 * @param max size Y
	 * @param max size Z
	 * @return partitions
	 */
	public static Vector<Model> computePartitions(Model environment, float partitionSizeX, float partitionSizeY, float partitionSizeZ) {
		Partition p = new Partition(
			new Object3DModel(environment),
			partitionSizeX,
			partitionSizeY,
			partitionSizeZ
		);
		if (p.computeDimensions()) {
			p.computePartitions();
		}
		return p.partitions;
	}

	private Object3DModel environment;
	private float partitionSizeX;
	private float partitionSizeY;
	private float partitionSizeZ;
	private Vector<Model> partitions;
	private float minX;
	private float minY;
	private float minZ;
	private float maxX;
	private float maxY;
	private float maxZ;
	private int partitionFaceCount;

	/**
	 * Constructor
	 * @param environment
	 * @param partitionSizeX
	 * @param partitionSizeY
	 * @param partitionSizeZ
	 */
	public Partition(
		Object3DModel environment, float partitionSizeX,
		float partitionSizeY, float partitionSizeZ) {
		//
		this.environment = environment;
		this.partitionSizeX = partitionSizeX;
		this.partitionSizeY = partitionSizeY;
		this.partitionSizeZ = partitionSizeZ;
		partitions = new Vector<Model>();
	}

	/**
	 * Compute model dimensions recursively
	 * @param groups
	 */
	private boolean computeDimensions() {
		BoundingBox boundingBox = ModelUtilities.createBoundingBox(environment);
		if (boundingBox == null) return false;
		float[] minXYZ = boundingBox.getMin().getArray();
		float[] maxXYZ = boundingBox.getMax().getArray();
		minX = minXYZ[0];
		minY = minXYZ[1];
		minZ = minXYZ[2];
		maxX = maxXYZ[0];
		maxY = maxXYZ[1];
		maxZ = maxXYZ[2];
		return true;
	}

	/**
	 * Computes a environment partition
	 * 	we do not support animations yet
	 * @param partition
	 * @param partition bounding box
	 * @return bool if partition contains data
	 */
	private boolean computePartition(Model partition, BoundingBox partitionBoundingBox) {
		// compute partition
		partitionFaceCount = 0;
		computePartitionGroup(partition, partitionBoundingBox, null, environment.getModel().getSubGroups());
		return partitionFaceCount > 0;
	}

	/**
	 * Compute partition groups recursively
	 * @param environment part
	 * @param partition bounding box
	 * @param envSubGroups
	 */
	private void computePartitionGroup(
		Model partition,
		BoundingBox partitionBoundingBox,
		Group envParentGroup,
		Hashtable<String, Group> envSubGroups) {
		//
		Iterator<Map.Entry<String, Group>> i = envSubGroups.entrySet().iterator();
		while (i.hasNext()) {
			Map.Entry<String, Group> entry = i.next();
			Group envGroup = entry.getValue();

			// skip on joints
			if (envGroup.isJoint() == true) {
				continue;
			}

			// create partition group
			Group partitionGroup = new Group(
				partition,
				envGroup.getId(),
				envGroup.getName()
			);

			// register group groups
			partition.getGroups().put(envGroup.getId(), partitionGroup);

			// register group in parent sub group list
			if (envParentGroup == null) {
				partition.getSubGroups().put(envGroup.getId(), partitionGroup);
			} else {
				envParentGroup.getSubGroups().put(envGroup.getId(), partitionGroup);
			}

			// skip on groups without faces
			if (envGroup.getFaceCount() == 0) {
				continue;
			}

			Object3DGroupMesh mesh = environment.getMesh(envGroup.getId());
			// just put out a warning if no mesh attached
			if (mesh == null) {
				System.out.println("Warning: Partition::computePartitionGroup(): Group has no mesh attached! Program will crash.");
			}

			//
			Vector3D envGroupVertices[] = mesh.getTransformedVertices();
			Vector3D envGroupNormals[] = mesh.getTransformedNormals();
			TextureCoordinate envGroupTextureCoordinates[] = envGroup.getTextureCoordinates();

			Vector<FacesEntity> partitionFacesEntities = new Vector<FacesEntity>(); 
			Hashtable<Integer, Integer> partitionedVertexMapping = new Hashtable<Integer, Integer>();
			Hashtable<Integer, Integer> partitionedNormalsMapping = new Hashtable<Integer, Integer>();
			Hashtable<Integer, Integer> partitionedTextureCoordinatesMapping = new Hashtable<Integer, Integer>();
			Vector<Vector3D> partitionedVertices = new Vector<Vector3D>();
			Vector<Vector3D> partitionedNormals = new Vector<Vector3D>();
			Vector<TextureCoordinate> partitionedTextureCoordinates = envGroupTextureCoordinates == null?null:new Vector<TextureCoordinate>();

			// determine faces in partition
			for (FacesEntity envFacesEntity: envGroup.getFacesEntities()) {
				Vector<Face> partitionedEntityFaces = new Vector<Face>();
				Face[] envEntityFaces = envFacesEntity.getFaces();
				for (int fIdx = 0; fIdx < envEntityFaces.length; fIdx++) {
					Face face = envEntityFaces[fIdx];
					int faceVertexIndices[] = face.getVertexIndices();
					int faceNormalIndices[] = face.getNormalIndices();
					int faceTextureCoordinateIndices[] = face.getTextureCoordinateIndices();
					Triangle triangle = new Triangle(
						envGroupVertices[faceVertexIndices[0]].clone(),
						envGroupVertices[faceVertexIndices[1]].clone(),
						envGroupVertices[faceVertexIndices[2]].clone()
					);
					boolean faceInPartition = CollisionDetection.doCollide(triangle, partitionBoundingBox) != null;
					if (faceInPartition) {
						// process vertices
						int[] partitionedVertexIndexArray = new int[3];
						for (int faceVertexIdx = 0; faceVertexIdx < faceVertexIndices.length; faceVertexIdx++) {
							int faceVertexIndex = faceVertexIndices[faceVertexIdx];
							int partitionedFaceVertexIndex = -1;
							// check if already exists in partition
							Integer tmp = partitionedVertexMapping.get(new Integer(faceVertexIndex));
							if (tmp == null) {
								// nope
								partitionedFaceVertexIndex = partitionedVertices.size();
								partitionedVertices.add(envGroupVertices[faceVertexIndex]);
								// add to mapping
								partitionedVertexMapping.put(
									new Integer(faceVertexIndex),
									new Integer(partitionedFaceVertexIndex)
								);
							} else {
								partitionedFaceVertexIndex = tmp.intValue();
							}
							partitionedVertexIndexArray[faceVertexIdx] = partitionedFaceVertexIndex;
						}
						// process normals
						int[] partitionedNormalIndexArray = new int[3];
						for (int faceNormalIdx = 0; faceNormalIdx < faceNormalIndices.length; faceNormalIdx++) {
							int faceNormalIndex = faceNormalIndices[faceNormalIdx];
							int partitionedFaceNormalIndex = -1;
							// check if already exists in partition
							Integer tmp = partitionedNormalsMapping.get(new Integer(faceNormalIndex));
							if (tmp == null) {
								// nope
								partitionedFaceNormalIndex = partitionedNormals.size();
								partitionedNormals.add(envGroupNormals[faceNormalIndex]);
								// add to mapping
								partitionedNormalsMapping.put(
									new Integer(faceNormalIndex),
									new Integer(partitionedFaceNormalIndex)
								);
							} else {
								partitionedFaceNormalIndex = tmp.intValue();
							}
							partitionedNormalIndexArray[faceNormalIdx] = partitionedFaceNormalIndex;
						}
	
						// process texture coordinates if we have any
						int[] partitionedTextureCoordinateIndexArray = new int[3];
						if (envGroupTextureCoordinates != null) {
							for (int faceTextureCoordinateIdx = 0; faceTextureCoordinateIdx < faceTextureCoordinateIndices.length; faceTextureCoordinateIdx++) {
								int faceTextureCoordinateIndex = faceTextureCoordinateIndices[faceTextureCoordinateIdx];
								int partitionedFaceTextureCoordinateIndex = -1;
								// check if already exists in partition
								Integer tmp = partitionedTextureCoordinatesMapping.get(new Integer(faceTextureCoordinateIndex));
								if (tmp == null) {
									// nope
									partitionedFaceTextureCoordinateIndex = partitionedTextureCoordinates.size();
									partitionedTextureCoordinates.add(envGroupTextureCoordinates[faceTextureCoordinateIndex]);
									// add to mapping
									partitionedTextureCoordinatesMapping.put(
										new Integer(faceTextureCoordinateIndex),
										new Integer(partitionedFaceTextureCoordinateIndex)
									);
								} else {
									partitionedFaceTextureCoordinateIndex = tmp.intValue();
								}
								partitionedTextureCoordinateIndexArray[faceTextureCoordinateIdx] = partitionedFaceTextureCoordinateIndex;
							}
						}
						Face partitionedFace = new Face(
							partitionGroup,
							partitionedVertexIndexArray[0],
							partitionedVertexIndexArray[1],
							partitionedVertexIndexArray[2],
							partitionedNormalIndexArray[0],
							partitionedNormalIndexArray[1],
							partitionedNormalIndexArray[2]
						);
						if (partitionedTextureCoordinates != null) {
							partitionedFace.setTextureCoordinateIndices(
								partitionedTextureCoordinateIndexArray[0],
								partitionedTextureCoordinateIndexArray[1],
								partitionedTextureCoordinateIndexArray[2]
							);
						}
						partitionedEntityFaces.add(partitionedFace);
						partitionFaceCount++;
					}
				}

				// create entity if we have any faces
				if (partitionedEntityFaces.size() > 0) {
					FacesEntity facesEntity = new FacesEntity(partitionGroup, envFacesEntity.getId());
					facesEntity.setupFaces(partitionedEntityFaces);
					partitionFacesEntities.add(facesEntity);
					// handle material if we have any
					if (envFacesEntity.getMaterial() != null) {
						Material material = partition.getMaterials().get(envFacesEntity.getMaterial().getId());
						// did we clone this material already?
						if (material == null) {
							// otherwise clone
							//	TODO: clone complete material
							material = new Material(envFacesEntity.getMaterial().getId());
							material.getAmbientColor().set(envFacesEntity.getMaterial().getAmbientColor());
							material.getDiffuseColor().set(envFacesEntity.getMaterial().getDiffuseColor());
							material.getSpecularColor().set(envFacesEntity.getMaterial().getSpecularColor());
							material.getEmissionColor().set(envFacesEntity.getMaterial().getEmissionColor());
							material.setShininess(envFacesEntity.getMaterial().getShininess());
							material.setDiffuseTexture(envFacesEntity.getMaterial().getDiffuseTexture());
							partition.getMaterials().put(material.getId(), material);
						}
						// yep, use this clone
						facesEntity.setMaterial(material);
					}
				}
			}

			// set up group
			partitionGroup.setupVertices(partitionedVertices);
			partitionGroup.setupNormals(partitionedNormals);
			if (partitionedTextureCoordinates != null) {
				partitionGroup.setupTextureCoordinates(partitionedTextureCoordinates);
			}
			partitionGroup.setupFacesEntities(partitionFacesEntities);

			// check sub groups
			computePartitionGroup(
				partition,
				partitionBoundingBox,
				envGroup,
				envGroup.getSubGroups()
			);
		}
	}

	/**
	 * Partitioning main logic 
	 */
	private void computePartitions() {
		int partitionId = 0;
		for (float xPartition = minX; xPartition < maxX; xPartition+= partitionSizeX)
		for (float yPartition = minY; yPartition < maxY; yPartition+= partitionSizeY)
		for (float zPartition = minZ; zPartition < maxZ; zPartition+= partitionSizeZ) {
			// create partition model
			Model partition = new Model(
				environment.getModel().getId() + "-" + partitionId,
				environment.getModel().getName()
			);

			// create partition bounding box
			BoundingBox partitionBoundingBox = new BoundingBox(
				xPartition,
				yPartition,
				zPartition,
				xPartition + partitionSizeX,
				yPartition + partitionSizeY,
				zPartition + partitionSizeZ
			);

			// create partition
			if (computePartition(
				partition,
				partitionBoundingBox
			) == true) {
				// we have a model for this partition, add it
				partitions.add(partition);
				partitionId++;
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "Partition [minX=" + minX + ", minY=" + minY + ", minZ=" + minZ
				+ ", maxX=" + maxX + ", maxY=" + maxY + ", maxZ=" + maxZ + "]";
	}

}
