package net.drewke.tdme.math;

import java.util.Arrays;

/**
 * 4x4 Matrix class
 * @author Andreas Drewke, Song Ho Ahn <song.ahn@gmail.com>
 * @version $Id: cebe207dabcf16d944cca6ebac0d4020b235321b $
 */
public class Matrix3D {

	final static private float DEG2RAD = 3.141593f / 180;

	protected float data[];

	/**
	 * Public constructor
	 */
	public Matrix3D() {
		data = new float[16];
		for (int i = 0; i < data.length; i++) {
			data[i] = 0f;
		}
	}

	/**
	 * Public constructor
	 * @param float array m
	 */
	public Matrix3D(float[] m) {
		data = new float[16];
		for (int i = 0; i < 16; i++) {
			data[i] = m[i];
		}
	}

	/**
	 * Public constructor
	 * @param matrix
	 */
	public Matrix3D(Matrix3D matrix) {
		data = new float[16];
		for (int i = 0; i < 16; i++) {
			data[i] = matrix.data[i];
		}
	}

	/**
	 * Public constructor
	 * @param xx
	 * @param xy
	 * @param xz
	 * @param xw
	 * @param yx
	 * @param yy
	 * @param yz
	 * @param yw
	 * @param zx
	 * @param zy
	 * @param zz
	 * @param zw
	 * @param wx
	 * @param wy
	 * @param wz
	 * @param ww
	 */
	public Matrix3D(
		float xx, float xy, float xz, float xw,
		float yx, float yy, float yz, float yw,
		float zx, float zy, float zz, float zw,
		float wx, float wy, float wz, float ww) {
		//
		data = new float[16];
		set(
			xx, xy, xz, xw,
			yx, yy, yz, yw,
			zx, zy, zz, zw,
			wx, wy, wz, ww
		);
	}

	/**
	 * Set up matrix by values
	 * @param xx
	 * @param xy
	 * @param xz
	 * @param xw
	 * @param yx
	 * @param yy
	 * @param yz
	 * @param yw
	 * @param zx
	 * @param zy
	 * @param zz
	 * @param zw
	 * @param wx
	 * @param wy
	 * @param wz
	 * @param ww
	 * @return this matrix
	 */
	public Matrix3D set(
		float xx, float xy, float xz, float xw,
		float yx, float yy, float yz, float yw,
		float zx, float zy, float zz, float zw,
		float wx, float wy, float wz, float ww) {
		data[0] = xx;
		data[1] = xy;
		data[2] = xz;
		data[3] = xw;
		data[4] = yx; 
		data[5] = yy;
		data[6] = yz;
		data[7] = yw;
		data[8] = zx;
		data[9] = zy;
		data[10] = zz;
		data[11] = zw;
		data[12] = wx;
		data[13] = wy;
		data[14] = wz;
		data[15] = ww;
		return this;
	}

	/**
	 * Sets up this matrix by matrix m
	 * @param m
	 * @return
	 */
	public Matrix3D set(float[] m) {
		for (int i = 0; i < m.length; i++) {
			data[i] = m[i];
		}
		return this;
	}

	/**
	 * Sets up this matrix by matrix m
	 * @param m
	 * @return
	 */
	public Matrix3D set(Matrix3D m) {
		for (int i = 0; i < m.data.length; i++) {
			data[i] = m.data[i];
		}
		return this;
	}

	/**
	 * Setup identity matrix
	 * @return this matrix
	 */
	public Matrix3D identity() {
	    data[0] = 1.0f;
	    data[1] = 0.0f;
	    data[2] = 0.0f;
	    data[3] = 0.0f;
	    data[4] = 0.0f;
	    data[5] = 1.0f;
	    data[6] = 0.0f;
	    data[7] = 0.0f;
	    data[8] = 0.0f;
	    data[9] = 0.0f;
	    data[10] = 1.0f;
	    data[11] = 0.0f;
	    data[12] = 0.0f;
	    data[13] = 0.0f;
	    data[14] = 0.0f;
	    data[15] = 1.0f;
	    return this;
	}

	/**
	 * Multiplies a vector with this matrix
	 * @param v
	 * @return vector 3d
	 */
	public Vector3D multiply(Vector3D v) {
		return new Vector3D(
			v.data[0] * data[0] + v.data[1] * data[4] + v.data[2] * data[8] + data[12],
			v.data[0] * data[1] + v.data[1] * data[5] + v.data[2] * data[9] + data[13],
			v.data[0] * data[2] + v.data[1] * data[6] + v.data[2] * data[10] + data[14]
	    );
	}

	/**
	 * Multiplies a vector with this matrix ignoring translation
	 * @param v
	 * @return vector 3d
	 */
	public Vector3D multiplyNoTranslation(Vector3D v) {
		return new Vector3D(
			v.data[0] * data[0] + v.data[1] * data[4] + v.data[2] * data[8],
			v.data[0] * data[1] + v.data[1] * data[5] + v.data[2] * data[9],
			v.data[0] * data[2] + v.data[1] * data[6] + v.data[2] * data[10]
	    );
	}

	/**
	 * Multiplies this matrix with another matrix
	 * @param m
	 * @return this matrix
	 */
	public Matrix3D multiply(Matrix3D m) {
		float _data[] = {
			data[0] * m.data[0]  + data[1] * m.data[4]  + data[2] * m.data[8]  + data[3] * m.data[12],
			data[0] * m.data[1]  + data[1] * m.data[5]  + data[2] * m.data[9]  + data[3] * m.data[13],
			data[0] * m.data[2]  + data[1] * m.data[6]  + data[2] * m.data[10] + data[3] * m.data[14],
			data[0] * m.data[3]  + data[1] * m.data[7]  + data[2] * m.data[11] + data[3] * m.data[15],
            data[4] * m.data[0]  + data[5] * m.data[4]  + data[6] * m.data[8]  + data[7] * m.data[12],
            data[4] * m.data[1]  + data[5] * m.data[5]  + data[6] * m.data[9]  + data[7] * m.data[13],
            data[4] * m.data[2]  + data[5] * m.data[6]  + data[6] * m.data[10]  + data[7] * m.data[14],
            data[4] * m.data[3]  + data[5] * m.data[7]  + data[6] * m.data[11]  + data[7] * m.data[15],
            data[8] * m.data[0]  + data[9] * m.data[4]  + data[10] * m.data[8] + data[11] * m.data[12],
            data[8] * m.data[1]  + data[9] * m.data[5]  + data[10] * m.data[9] + data[11] * m.data[13],
            data[8] * m.data[2]  + data[9] * m.data[6]  + data[10] * m.data[10] + data[11] * m.data[14],
            data[8] * m.data[3]  + data[9] * m.data[7]  + data[10] * m.data[11] + data[11] * m.data[15],
            data[12] * m.data[0] + data[13] * m.data[4] + data[14] * m.data[8] + data[15] * m.data[12],
            data[12] * m.data[1] + data[13] * m.data[5] + data[14] * m.data[9] + data[15] * m.data[13],
            data[12] * m.data[2] + data[13] * m.data[6] + data[14] * m.data[10] + data[15] * m.data[14],
            data[12] * m.data[3] + data[13] * m.data[7] + data[14] * m.data[11] + data[15] * m.data[15]
		};

		// set up new matrix data
		data = _data;
		return this;
	}

	/**
	 * Scales this matrix
	 * @param s
	 * @returns this matrix
	 */
	public Matrix3D scale(float s) {
		data[0]*= s;
		data[1]*= s;
		data[2]*= s;
		data[3]*= s;
		data[4]*= s;
		data[5]*= s;
		data[6]*= s;
		data[7]*= s;
		data[8]*= s;
		data[9]*= s;
		data[10]*= s;
		data[11]*= s;
		data[12]*= s;
		data[13]*= s;
		data[14]*= s;
		data[15]*= s;
		return this;
	}

	/**
	 * Scales this matrix by given vector
	 * @param v
	 * @return this matrix
	 */
	public Matrix3D scale(Vector3D v) {
		data[0] *= v.data[0];
		data[1] *= v.data[0];
		data[2] *= v.data[0];
		data[3] *= v.data[0];
		data[4] *= v.data[1];
		data[5] *= v.data[1];
		data[6] *= v.data[1];
		data[7] *= v.data[1];
		data[8] *= v.data[2];
		data[9] *= v.data[2];
		data[10] *= v.data[2];
		data[11] *= v.data[2];
		return this;
	}

	/**
	 * Sets up a translation matrix
	 * @param x
	 * @param y
	 * @param z
	 * @return this matrix
	 */
	public Matrix3D translate(Vector3D v) {
		data[12] += v.data[0] * data[0] + v.data[1] * data[4] + v.data[2] * data[8];
		data[13] += v.data[0] * data[1] + v.data[1] * data[5] + v.data[2] * data[9];
		data[14] += v.data[0] * data[2] + v.data[1] * data[6] + v.data[2] * data[10];
		return this;
	}

	/**
	 * Creates a rotation matrix
	 * @param angle
	 * @param vector v
	 * @return this matrix
	 */
	public Matrix3D rotate(float angle, Vector3D v) {
		float c = (float)Math.cos(angle * DEG2RAD);
		float s = (float)Math.sin(angle * DEG2RAD);
		float xx = v.data[0] * v.data[0];
		float xy = v.data[0] * v.data[1];
		float xz = v.data[0] * v.data[2];
		float yy = v.data[1] * v.data[1];
		float yz = v.data[1] * v.data[2];
		float zz = v.data[2] * v.data[2];
		data[0] = xx * (1 - c) + c;
		data[1] = xy * (1 - c) - v.data[2] * s;
		data[2] = xz * (1 - c) + v.data[1] * s;
		data[3] = 0;
		data[4] = xy * (1 - c) + v.data[2] * s;
		data[5] = yy * (1 - c) + c;
		data[6] = yz * (1 - c) - v.data[0] * s;
		data[7] = 0;
		data[8] = xz * (1 - c) - v.data[1] * s;
		data[9] = yz * (1 - c) + v.data[0] * s;
		data[10]= zz * (1 - c) + c;
		data[11]= 0;
		data[12]= 0;
		data[13]= 0;
		data[14]= 0;
		data[15]= 1;
		return this;
	}

	/**
	 * Transposes this matrix
	 * @return this matrix
	 */
	public Matrix3D transpose() {
		float _data[] = new float[16];
		_data[0] = data[0];
		_data[1] = data[4];
		_data[2] = data[8];
		_data[3] = data[12];
		_data[4] = data[1];
		_data[5] = data[5];
		_data[6] = data[9];
		_data[7] = data[13];
		_data[8] = data[2];
		_data[9] = data[6];
		_data[10] = data[10];
		_data[11] = data[14];
		_data[12] = data[3];
		_data[13] = data[7];
		_data[14] = data[11];
		_data[15] = data[15];
		data = _data;
		return this;
	}

	/**
	 * Returns the matrix determinant
	 * @return
	 */
	public float getDeterminant() {
		return
			data[0] * getCofactor(data[5],data[6],data[7], data[9],data[10],data[11], data[13],data[14],data[15]) -
			data[1] * getCofactor(data[4],data[6],data[7], data[8],data[10],data[11], data[12],data[14],data[15]) +
			data[2] * getCofactor(data[4],data[5],data[7], data[8],data[9], data[11], data[12],data[13],data[15]) -
			data[3] * getCofactor(data[4],data[5],data[6], data[8],data[9], data[10], data[12],data[13],data[14]);
	}


	/**
	 * Returns co factor
	 * @param m0
	 * @param m1
	 * @param m2
	 * @param m3
	 * @param m4
	 * @param m5
	 * @param m6
	 * @param m7
	 * @param m8
	 * @return
	 */
	public float getCofactor(
		float m0, float m1, float m2,
        float m3, float m4, float m5,
        float m6, float m7, float m8) {
			//
			return
				m0 * (m4 * m8 - m5 * m7) -
				m1 * (m3 * m8 - m5 * m6) +
				m2 * (m3 * m7 - m4 * m6);
	}

	/**
	 * Inverts the matrix
	 * @return this matrix
	 */
	public Matrix3D invert() {
		// get cofactors of minor matrices
		float cofactor0 = getCofactor(data[5], data[6], data[7], data[9], data[10], data[11], data[13], data[14], data[15]);
		float cofactor1 = getCofactor(data[4], data[6], data[7], data[8], data[10], data[11], data[12], data[14], data[15]);
		float cofactor2 = getCofactor(data[4], data[5], data[7], data[8], data[9], data[11], data[12], data[13], data[15]);
		float cofactor3 = getCofactor(data[4], data[5], data[6], data[8], data[9], data[10], data[12], data[13], data[14]);
		
		// get determinant
		float determinant = data[0] * cofactor0 - data[1] * cofactor1 + data[2] * cofactor2 - data[3] * cofactor3;
		if((float)Math.abs(determinant) <= 0.00001f) {
		    identity();
		    return this;
		}
		
		// get rest of cofactors for adj(M)
		float cofactor4 = getCofactor(data[1],data[2],data[3], data[9],data[10],data[11], data[13],data[14],data[15]);
		float cofactor5 = getCofactor(data[0],data[2],data[3], data[8],data[10],data[11], data[12],data[14],data[15]);
		float cofactor6 = getCofactor(data[0],data[1],data[3], data[8],data[9], data[11], data[12],data[13],data[15]);
		float cofactor7 = getCofactor(data[0],data[1],data[2], data[8],data[9], data[10], data[12],data[13],data[14]);
		
		float cofactor8 = getCofactor(data[1], data[2], data[3], data[5], data[6], data[7], data[13], data[14], data[15]);
		float cofactor9 = getCofactor(data[0], data[2], data[3], data[4], data[6], data[7], data[12], data[14], data[15]);
		float cofactor10 = getCofactor(data[0], data[1], data[3], data[4], data[5], data[7], data[12], data[13], data[15]);
		float cofactor11 = getCofactor(data[0], data[1], data[2], data[4], data[5], data[6], data[12], data[13], data[14]);
		
		float cofactor12 = getCofactor(data[1], data[2], data[3], data[5], data[6], data[7], data[9], data[10], data[11]);
		float cofactor13 = getCofactor(data[0], data[2], data[3], data[4], data[6], data[7], data[8], data[10], data[11]);
		float cofactor14 = getCofactor(data[0], data[1], data[3], data[4], data[5], data[7], data[8], data[9], data[11]);
		float cofactor15 = getCofactor(data[0], data[1], data[2], data[4], data[5], data[6], data[8], data[9], data[10]);
		
		// build inverse matrix = adj(M) / det(M)
		// adjugate of M is the transpose of the cofactor matrix of M
		float invDeterminant = 1.0f / determinant;
		data[0] =  invDeterminant * cofactor0;
		data[1] = -invDeterminant * cofactor4;
		data[2] =  invDeterminant * cofactor8;
		data[3] = -invDeterminant * cofactor12;
		
		data[4] = -invDeterminant * cofactor1;
		data[5] =  invDeterminant * cofactor5;
		data[6] = -invDeterminant * cofactor9;
		data[7] =  invDeterminant * cofactor13;
		
		data[8] =  invDeterminant * cofactor2;
		data[9] = -invDeterminant * cofactor6;
		data[10]=  invDeterminant * cofactor10;
		data[11]= -invDeterminant * cofactor14;
		
		data[12]= -invDeterminant * cofactor3;
		data[13]=  invDeterminant * cofactor7;
		data[14]= -invDeterminant * cofactor11;
		data[15]=  invDeterminant * cofactor15;

		//
		return this;
	}

	/**
	 * Returns if this matrix equals m
	 * @param m
	 * @return equals
	 */
	public boolean equals(Matrix3D m) {
		return
			data[0] == m.data[0] &&
			data[1] == m.data[1] &&
			data[2] == m.data[2] &&
			data[3] == m.data[3] &&
			data[4] == m.data[4] &&
			data[5] == m.data[5] &&
			data[6] == m.data[6] &&
			data[7] == m.data[7] &&
			data[8] == m.data[8] &&
			data[9] == m.data[9] &&
			data[10] == m.data[10] &&
			data[11] == m.data[11] &&
			data[12] == m.data[12] &&
			data[13] == m.data[13] &&
			data[14] == m.data[14] &&
			data[15] == m.data[15];
	}

	/**
	 * Returns array data
	 * @return array data
	 */
	public float[] getArray() {
		return data;
	}

	/**
	 * Clones this matrix
	 * @return new cloned matrix
	 */
	public Matrix3D clone() {
		return new Matrix3D(
			data[0], data[1], data[2], data[3],
			data[4], data[5], data[6], data[7],
			data[8], data[9], data[10], data[11],
			data[12], data[13], data[14], data[15]
		);
	}

	/**
	 * Clones this matrix
	 * @return new cloned matrix
	 */
	public Matrix3D cloneNoTranslation() {
		return new Matrix3D(
			data[0], data[1], data[2], data[3],
			data[4], data[5], data[6], data[7],
			data[8], data[9], data[10], data[11],
			0.0f, 0.0f, 0.0f, data[15]
		);
	}

	/**
	 * Interpolates between matrix 1 and matrix 2 by 0f<=T<=1f linearly 
	 * @param m1
	 * @param m2
	 * @param t
	 * @return interpolated matrix
	 */
	public static Matrix3D interpolateLinear(Matrix3D m1, Matrix3D m2, float t) {
		return new Matrix3D(
			(m2.data[0] * t) + ((1f - t) * m1.data[0]),
			(m2.data[1] * t) + ((1f - t) * m1.data[1]),
			(m2.data[2] * t) + ((1f - t) * m1.data[2]),
			(m2.data[3] * t) + ((1f - t) * m1.data[3]),
			(m2.data[4] * t) + ((1f - t) * m1.data[4]),
			(m2.data[5] * t) + ((1f - t) * m1.data[5]),
			(m2.data[6] * t) + ((1f - t) * m1.data[6]),
			(m2.data[7] * t) + ((1f - t) * m1.data[7]),
			(m2.data[8] * t) + ((1f - t) * m1.data[8]),
			(m2.data[9] * t) + ((1f - t) * m1.data[9]),
			(m2.data[10] * t) + ((1f - t) * m1.data[10]),
			(m2.data[11] * t) + ((1f - t) * m1.data[11]),
			(m2.data[12] * t) + ((1f - t) * m1.data[12]),
			(m2.data[13] * t) + ((1f - t) * m1.data[13]),
			(m2.data[14] * t) + ((1f - t) * m1.data[14]),
			(m2.data[15] * t) + ((1f - t) * m1.data[15])
		);		
	}

	/**
	 * @return string representation
	 */
	public String toString() {
		return Arrays.toString(data);
	}

}