package net.drewke.tdme.math;

import java.util.Arrays;

/**
 * Vector3D class
 * @author andreas.drewke
 * @version $Id: 6b7169e85b6f17d2595dc64c4ed5055e6a124d19 $
 */
public class Vector3D {

	protected float data[];

	/**
	 * Public constructor
	 */
	public Vector3D() {
		data = new float[3];
		data[0] = 0f;
		data[1] = 0f;
		data[2] = 0f;		
	}

	/**
	 * Public constructor
	 * 
	 * @param x
	 * @param y
	 * @param z
	 */
	public Vector3D(float x, float y, float z) {
		data = new float[3];
		data[0] = x;
		data[1] = y;
		data[2] = z;
	}

	/**
	 * Public constructor
	 * @param float array containing x,y,z values
	 */
	public Vector3D(float[] v) {
		data = new float[3];
		data[0] = v[0];
		data[1] = v[1];
		data[2] = v[2];		
	}

	/**
	 * Set up vector
	 * @param x
	 * @return this vector
	 */
	public Vector3D set(float x, float y, float z) {
		data[0] = x;
		data[1] = y;
		data[2] = z;
		return this;
	}

	/**
	 * Set up vector
	 * @param float array containing x,y,z values
	 * @return this vector
	 */
	public Vector3D set(float[] v) {
		data[0] = v[0];
		data[1] = v[1];
		data[2] = v[2];
		return this;
	}

	/**
	 * Set up vector
	 * @param v
	 * @return this vector
	 */
	public Vector3D set(Vector3D v) {
		data[0] = v.data[0];
		data[1] = v.data[1];
		data[2] = v.data[2];
		return this;
	}

	/**
	 * @return x
	 */
	public float getX() {
		return data[0];
	}

	/**
	 * Set X
	 * @param x
	 */
	public void setX(float x) {
		data[0] = x;
	}

	/**
	 * @return y
	 */
	public float getY() {
		return data[1];
	}

	/**
	 * Set Y
	 * @param y
	 */
	public void setY(float y) {
		data[1] = y;
	}

	/**
	 * @return z
	 */
	public float getZ() {
		return data[2];
	}

	/**
	 * Set Z
	 * @param z
	 */
	public void setZ(float z) {
		data[2] = z;
	}

	/**
	 * @return vector as array
	 */
	public float[] getArray() {
		return data;
	}

	/**
	 * Compute the cross product of vector v1 and v2
	 * @param v1
	 * @param v2
	 * @return cross product vector of v1 and v2
	 */
	public static Vector3D computeCrossProduct(Vector3D v1, Vector3D v2) {
		return new Vector3D(
			(v1.data[1] * v2.data[2]) - (v1.data[2] * v2.data[1]),
			(v1.data[2] * v2.data[0]) - (v1.data[0] * v2.data[2]),
			(v1.data[0] * v2.data[1]) - (v1.data[1] * v2.data[0])
		);
	}

	/**
	 * Compute the dot product of vector v1 and v2
	 * @param v1
	 * @param v2
	 * @return Vector3D
	 */
	public static float computeDotProduct(Vector3D v1, Vector3D v2) {
		return 
			(v1.data[0] * v2.data[0]) +
			(v1.data[1] * v2.data[1]) +
			(v1.data[2] * v2.data[2]);
	}

	/**
	 * @return the vectors length
	 */
	public float computeLength() {
		return (float)Math.sqrt(
			(data[0] * data[0])	+ (data[1] * data[1]) + (data[2] * data[2])
		);
	}

	/**
	 * Normalize the vector
	 * @return this vector
	 */
	public Vector3D normalize() {
		float length = computeLength();
		data[0] /= length;
		data[1] /= length;
		data[2] /= length;
		return this;
	}

	/**
	 * Adds a vector
	 * @param v
	 * @return this vector
	 */
	public Vector3D add(Vector3D v) {
		data[0]+= v.data[0];
		data[1]+= v.data[1];
		data[2]+= v.data[2];
		return this;
	}

	/**
	 * Adds a vector
	 * @param v
	 * @return this vector 
	 */
	public Vector3D sub(Vector3D v) {
		data[0]-= v.data[0];
		data[1]-= v.data[1];
		data[2]-= v.data[2];
		return this;
	}

	/**
	 * Scale this vector
	 * @param scale
	 * @return this vector 
	 */
	public Vector3D scale(float scale) {
		data[0]*= scale;
		data[1]*= scale;
		data[2]*= scale;
		return this;
	}

	/**
	 * Scale this vector
	 * @param scale
	 * @return this vector 
	 */
	public Vector3D scale(Vector3D scale) {
		data[0]*= scale.data[0];
		data[1]*= scale.data[1];
		data[2]*= scale.data[2];
		return this;
	}

	/**
	 * Clones the vector
	 * @return new cloned vector
	 */
	public Vector3D clone() {
		return new Vector3D(
			data[0],
			data[1],
			data[2]
		);
	}

	/**
	 * @return string representation
	 */
	public String toString() {
		return Arrays.toString(data);
	}

}