package net.drewke.tdme.model;

import java.util.Arrays;
import java.util.Hashtable;
import java.util.Vector;

import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.math.Vector3D;

/**
 * 3d object group
 * @author andreas.drewke
 * @version $Id: f1d2c3186b6e6b87851cad0c2472399a417caf6a $
 */
public class Group {

	private Model model;
	private String id;
	private String name;
	private boolean isJoint;
	private Matrix3D transformationsMatrix;
	private Vector3D[] vertices;
	private int[][] verticesNormalIndices;
	private Vector3D[] normals;
	private TextureCoordinate[] textureCoordinates;
	private Animation animation;
	private Skinning skinning;
	private FacesEntity[] facesEntities;
	private Hashtable<String, Group> subGroups;

	/**
	 * Public constructor
	 * @param id
	 * @param name
	 */
	public Group(Model model, String id, String name) {
		this.model = model;
		this.id = id;
		this.name = name;
		this.transformationsMatrix = new Matrix3D().identity();
		this.vertices = new Vector3D[0];
		this.verticesNormalIndices = new int[0][0];
		this.normals = new Vector3D[0];
		this.textureCoordinates = null;
		this.animation = null;
		this.skinning = null;
		this.facesEntities = new FacesEntity[0];
		this.subGroups = new Hashtable<String, Group>();
	}

	/**
	 * @return model
	 */
	public Model getModel() {
		return model;
	}

	/**
	 * Returns id
	 * @return id
	 */
	public String getId() {
		return id;
	}

	/**
	 * @return group's name
	 */
	public String getName() {
		return name;
	}

	/**
	 * @return if this group is a joint/bone
	 */
	public boolean isJoint() {
		return isJoint;
	}

	/**
	 * Sets up if this group is a joint or not 
	 * @param isbone
	 */
	public void setJoint(boolean isJoint) {
		this.isJoint = isJoint;
	}

	/**
	 * @return transformations matrix related to parent group
	 */
	public Matrix3D getTransformationsMatrix() {
		return transformationsMatrix;
	}

	/**
	 * Set up vertices
	 * @param vertices
	 */
	public void setupVertices(Vector<Vector3D> vertices) {
		this.vertices = vertices.toArray(new Vector3D[vertices.size()]);
	}

	/**
	 * @return vertices
	 */
	public Vector3D[] getVertices() {
		return vertices;
	}

	/**
	 * We have assigned several normals to a single vertex
	 * @return vertices normal indices
	 */
	public int[][] getVerticesNormalIndices() {
		return verticesNormalIndices;
	}

	/**
	 * Set up vertices normal indices
	 * @param verticesNormalIndices
	 */
	public void setupVerticesNormalIndices(Vector<Vector<Integer>> verticesNormalIndices) {
		this.verticesNormalIndices = new int[verticesNormalIndices.size()][];
		for (int i = 0; i < this.verticesNormalIndices.length; i++) {
			this.verticesNormalIndices[i] = new int[verticesNormalIndices.get(i).size()];
			for (int j = 0; j < this.verticesNormalIndices[i].length; j++) {
				this.verticesNormalIndices[i][j] = verticesNormalIndices.get(i).get(j);
			}
		}
	}

	/**
	 * @return normals
	 */
	public Vector3D[] getNormals() {
		return normals;
	}

	/**
	 * Set up normals
	 * @param normals
	 */
	public void setupNormals(Vector<Vector3D> normals) {
		this.normals = normals.toArray(new Vector3D[normals.size()]);
	}

	/**
	 * @return texture coordinates or null (optional)
	 */
	public TextureCoordinate[] getTextureCoordinates() {
		return textureCoordinates;
	}

	/**
	 * Setup texture coordinates
	 * @param textureCoordinates
	 */
	public void setupTextureCoordinates(Vector<TextureCoordinate> textureCoordinates) {
		this.textureCoordinates = textureCoordinates.toArray(new TextureCoordinate[textureCoordinates.size()]);
	}

	/**
	 * @return animation
	 */
	public Animation getAnimation() {
		return animation;
	}

	/**
	 * Creates an empty animation object
	 * @param frames per seconds
	 * @param frames
	 * @return animation
	 */
	public Animation createAnimation(float fps, int frames) {
		animation = new Animation(fps, frames);
		return animation;
	}

	/**
	 * @return skinning or null
	 */
	public Skinning getSkinning() {
		return skinning;
	}

	/**
	 * Creates an empty skinning object
	 * @return skinning
	 */
	public Skinning createSkinning() {
		skinning = new Skinning();
		return skinning;
	}

	/**
	 * @return number of faces in group
	 */
	public int getFaceCount() {
		// determine face count
		int faceCount = 0;
		for (FacesEntity facesEntity: facesEntities) {
			faceCount+= facesEntity.getFaces().length;
		}
		return faceCount;
	}

	/**
	 * @return faces entities
	 */
	public FacesEntity[] getFacesEntities() {
		return facesEntities;
	}

	/**
	 * Set up faces entities
	 * @param faces
	 */
	public void setupFacesEntities(Vector<FacesEntity> facesEntities) {
		this.facesEntities = facesEntities.toArray(new FacesEntity[facesEntities.size()]);
	}

	/**
	 * @return sub sub groups of this group
	 */
	public Hashtable<String, Group> getSubGroups() {
		return subGroups;
	}

	/**
	 * Returns a sub group by id
	 * @param groupId
	 * @return sub group or null
	 */
	public Group getSubGroupById(String groupId) {
		return subGroups.get(groupId);
	}

	/**
	 * @return string representation
	 */
	public String toString() {
		return
			"ObjectGroup [id=" + id +
			", name=" + name +
			", faces entities=" + Arrays.toString(facesEntities) +
			", vertices =" + Arrays.toString(vertices) +
			", normals =" + Arrays.toString(normals) +
			", texture coordinates =" + Arrays.toString(textureCoordinates) +
			", subgroups =" + subGroups +
			"]";
	}	

}