package net.drewke.tdme.model;

import java.io.IOException;

import net.drewke.tdme.assets.textures.Texture;

/**
 * Represents a object material
 * @author andreas.drewke
 * @version $Id: f9548499df69bcbc41dad85bdb2ccbe57e69515f $
 */
public class Material {

	private final static String defaultMaterialId = "tdme.default_material";
	private final static Material defaultMaterial = new Material(defaultMaterialId); 

	/**
	 * @return default material
	 */
	public static Material getDefaultMaterial() {
		return defaultMaterial;
	}

	private String id;
	private Color4 ambientColor;
	private Color4 diffuseColor;
	private Color4 specularColor;
	private Color4 emissionColor;
	private float shininess;
	private Texture diffuseTexture;

	/**
	 * Public constructor
	 * @param name
	 */
	public Material(String id) {
		this.id = id;
		ambientColor = new Color4(0.2f, 0.2f, 0.2f, 1.0f);
		diffuseColor = new Color4(0.8f, 0.8f, 0.8f, 1.0f);
		specularColor = new Color4(0.0f, 0.0f, 0.0f, 1.0f);
		emissionColor = new Color4(0.0f, 0.0f, 0.0f, 1.0f);
		shininess = 0.0f;
		diffuseTexture = null;
	}

	/**
	 * @return material id
	 */
	public String getId() {
		return id;
	}

	/**
	 * @return ambient color
	 */
	public Color4 getAmbientColor() {
		return ambientColor;
	}

	/**
	 * @return diffuse color
	 */
	public Color4 getDiffuseColor() {
		return diffuseColor;
	}

	/**
	 * @return specular color
	 */
	public Color4 getSpecularColor() {
		return specularColor;
	}

	/**
	 * @return emission color
	 */
	public Color4 getEmissionColor() {
		return emissionColor;
	}

	/**
	 * @return shininess
	 */
	public float getShininess() {
		return shininess;
	}

	/**
	 * Set up material shininess
	 * @param shininess
	 */
	public void setShininess(float shininess) {
		this.shininess = shininess;
	}

	/**
	 * Set ups a texture
	 * @param pathName
	 * @param fileName
	 * @throws IOException
	 */
	public void setupDiffuseTexture(String pathName, String fileName) {
		diffuseTexture = net.drewke.tdme.assets.TextureLoader.loadTexture(pathName, fileName);
	}

	/**
	 * @return if material has a diffuse texture
	 */
	public boolean hasDiffuseTexture() {
		return diffuseTexture != null;
	}

	/**
	 * @return the material's diffuse texture
	 */
	public Texture getDiffuseTexture() {
		return diffuseTexture;
	}

	/**
	 * Set up diffuse texture
	 * @param diffuseTexture
	 */
	public void setDiffuseTexture(Texture diffuseTexture) {
		this.diffuseTexture = diffuseTexture;
	}

	/**
	 * @return string representation
	 */
	public String toString() {
		return "Material [id=" + id + ", ambientColor=" + ambientColor
				+ ", diffuseColor=" + diffuseColor + ", specularColor="
				+ specularColor + ", emissionColor=" + emissionColor
				+ ", shininess=" + shininess + ", diffuseTexture="
				+ diffuseTexture + "]";
	}	

}