package net.drewke.tdme.parser;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.model.Animation;
import net.drewke.tdme.model.AnimationSetup;
import net.drewke.tdme.model.Face;
import net.drewke.tdme.model.FacesEntity;
import net.drewke.tdme.model.Group;
import net.drewke.tdme.model.Joint;
import net.drewke.tdme.model.JointWeight;
import net.drewke.tdme.model.Material;
import net.drewke.tdme.model.Model;
import net.drewke.tdme.model.Skinning;
import net.drewke.tdme.model.TextureCoordinate;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Collada DAE parser
 * @author Andreas Drewke
 * @version $Id: 3c0cb06d7f1919f282615b50b2d4b5a1661a0538 $
 */
public class DAEParser {

	/**
	 * Parses a Collada DAE file
	 * @param pathName
	 * @param fileName
	 * @return Model instance
	 * @throws IOException
	 */
	public static Model parse(String pathName, String fileName) throws Exception {
		// load dae xml document
		DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
		Document document = builder.parse(new File(pathName + "/" + fileName));
		Element xmlRoot = document.getDocumentElement();

		// 	create object
		Model model = new Model(fileName, fileName);

		// determine up axis
		for(Element xmlAsset: getChildrenByTagName(xmlRoot, "asset"))
		for(Element xmlAssetUpAxis: getChildrenByTagName(xmlAsset, "up_axis")) {
			String upAxis = xmlAssetUpAxis.getTextContent();
			if (upAxis.equalsIgnoreCase("Y_UP")) {
				// default: do nothing
			} else 
			if (upAxis.equalsIgnoreCase("Z_UP")) {
				// set up import transformation matrix
				model.getImportTransformationsMatrix().rotate(90f, new Vector3D(1f,0f,0f));
			} else
			if (upAxis.equalsIgnoreCase("X_UP")) {
				// set up import transformation matrix
				model.getImportTransformationsMatrix().rotate(90f, new Vector3D(0f,1f,0f));
			} else {
				System.out.println("Warning: Unknown up axis: " + upAxis);
			}
		}

		// parse scene from xml
		String xmlSceneId = null;
		Element xmlScene = getChildrenByTagName(xmlRoot, "scene").get(0);
		for(Element xmlInstanceVisualscene: getChildrenByTagName(xmlScene, "instance_visual_scene")) {
			xmlSceneId = xmlInstanceVisualscene.getAttribute("url").substring(1);
		}

		// check for xml scene id
		if (xmlSceneId == null) {
			throw new ParserException("No scene id found");
		}

		// parse visual scenes
		Element xmlLibraryVisualScenes = getChildrenByTagName(xmlRoot, "library_visual_scenes").get(0);
		for(Element xmlLibraryVisualScene: getChildrenByTagName(xmlLibraryVisualScenes, "visual_scene")) {
			String xmlVisualSceneId = xmlLibraryVisualScene.getAttribute("id");
			if (xmlVisualSceneId.equals(xmlSceneId)) {
				// default FPS
				float fps = 30f;

				// parse frames per second
				List<Element> xmlExtraNodes = getChildrenByTagName(xmlLibraryVisualScene, "extra");
				if (xmlExtraNodes.isEmpty() == false) {
					Element xmlExtraNode = xmlExtraNodes.get(0);
					for (Element xmlTechnique: getChildrenByTagName(xmlExtraNode, "technique")) {
						List<Element> xmlFrameRateNodes = getChildrenByTagName(xmlTechnique, "frame_rate");
						if (xmlFrameRateNodes.isEmpty() == false) {
							fps = Float.parseFloat(xmlFrameRateNodes.get(0).getTextContent());
							break;
						}
					}
				}

				// set up frames per seconds
				model.setFPS(fps);

				// visual scene root nodes
				for(Element xmlNode: getChildrenByTagName(xmlLibraryVisualScene, "node")) {
					Group group = parseVisualSceneNode(pathName, model, xmlRoot, xmlNode, fps);
					if (group != null) {
						model.getSubGroups().put(group.getId(), group);
						model.getGroups().put(group.getId(), group);
					}
				}
			}
		}

		// determine joints and mark them as joints
		Iterator<Map.Entry<String, Group>> i = model.getSubGroups().entrySet().iterator();
		Hashtable<String,Group> groups = model.getGroups();
		while (i.hasNext()) {
			Map.Entry<String, Group> entry = i.next();
			Group group = entry.getValue();
			Skinning skinning = group.getSkinning();
			// do we have a skinning
			if (skinning != null) {
				// yep
				for(Joint joint: skinning.getJoints()) {
					setJoint(groups.get(joint.getGroupId()));
				}
			}
		}

		//
		return model;
	}

	/**
	 * Sets up a group as joint taking all subgroups into account
	 * @param group
	 */
	private static void setJoint(Group root) {
		root.setJoint(true);
		Iterator<Map.Entry<String, Group>> i = root.getSubGroups().entrySet().iterator();
		while (i.hasNext()) {
			Map.Entry<String, Group> entry = i.next();
			Group group = entry.getValue();
			setJoint(group);
		}
	}

	/**
	 * Parses a DAE visual scene node
	 * @param pathName
	 * @param model
	 * @param xmlNode
	 * @param xmlRoot
	 * @param frames per second
	 */
	private static Group parseVisualSceneNode(String pathName, Model model, Element xmlRoot, Element xmlNode, float fps) throws Exception {
		List<Element> xmlInstanceControllers = getChildrenByTagName(xmlNode, "instance_controller");
		if (xmlInstanceControllers.isEmpty() == false) {
			return parseVisualSceneInstanceController(pathName, model, xmlRoot, xmlNode);
		} else {
			return parseNode(pathName, model, xmlRoot, xmlNode, fps);
		}
	}

	/**
	 * Parses a DAE visual scene group node
	 * @param pathName
	 * @param model
	 * @param xmlNode
	 * @param xmlRoot
	 * @param frames per seconds
	 */
	private static Group parseNode(String pathName, Model model, Element xmlRoot, Element xmlNode, float fps) throws Exception {
		String xmlNodeId = xmlNode.getAttribute("id");
		String xmlNodeName = xmlNode.getAttribute("name");
		if (xmlNodeId.length() == 0) xmlNodeId = xmlNodeName;

		StringTokenizer t = null;

		// default node matrix
		Matrix3D transformationsMatrix = null;

		// set up local transformations matrix
		List<Element> xmlMatrixElements = getChildrenByTagName(xmlNode, "matrix");
		if (xmlMatrixElements.size() == 1) {
			String xmlMatrix = getChildrenByTagName(xmlNode, "matrix").get(0).getTextContent();
			t = new StringTokenizer(xmlMatrix, " \n\r");

			// 
			transformationsMatrix = new Matrix3D(
				Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
				Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
				Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()), 
				Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
				Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
				Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
				Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()), 
				Float.parseFloat(t.nextToken()),  Float.parseFloat(t.nextToken())
			).transpose();
		}

		// tdme model definitions
		Group group = new Group(
			model,
			xmlNodeId,
			xmlNodeName
		);

		//
		if (transformationsMatrix != null) {
			group.getTransformationsMatrix().multiply(transformationsMatrix);
		}

		// parse animations
		List<Element> xmlAnimationsLibrary = getChildrenByTagName(xmlRoot, "library_animations");
		if (xmlAnimationsLibrary.isEmpty() == false) {
			List<Element> xmlAnimations = getChildrenByTagName(xmlAnimationsLibrary.get(0), "animation");
			for (Element xmlAnimation: xmlAnimations) {
				// older DAE has animation/animation xml nodes
				List<Element> _xmlAnimation = getChildrenByTagName(xmlAnimation, "animation");
				if (_xmlAnimation.isEmpty() == false) {
					xmlAnimation = _xmlAnimation.get(0);
				}

				// find sampler source
				String xmlSamplerSource = null;
				Element xmlChannel = getChildrenByTagName(xmlAnimation, "channel").get(0);
				if (xmlChannel.getAttribute("target").equals(xmlNodeId + "/matrix")) {
					xmlSamplerSource = xmlChannel.getAttribute("source").substring(1);
				}
	
				// check for sampler source
				if (xmlSamplerSource == null) {
					continue;
				}
	
				// parse animation output matrices
				String xmlSamplerOutputSource = null;
				Element xmlSampler = getChildrenByTagName(xmlAnimation, "sampler").get(0);
				for (Element xmlSamplerInput: getChildrenByTagName(xmlSampler, "input")) {
					if (xmlSamplerInput.getAttribute("semantic").equals("OUTPUT")) {
						xmlSamplerOutputSource = xmlSamplerInput.getAttribute("source").substring(1);
					}
				}
	
				// check for sampler source
				if (xmlSamplerOutputSource == null) {
					throw new ParserException("Could not fid xml sampler output source for animation for " + xmlNodeId);
				}
	
				// load animation output matrices
				for(Element xmlAnimationSource: getChildrenByTagName(xmlAnimation, "source")) {
					if (xmlAnimationSource.getAttribute("id").equals(xmlSamplerOutputSource)) {
						Element xmlFloatArray = getChildrenByTagName(xmlAnimationSource, "float_array").get(0);
						int frames = Integer.parseInt(xmlFloatArray.getAttribute("count")) / 16 - 1;

						// some models have animations without frames
						if (frames > 0) {

							// add default model animation setup
							AnimationSetup defaultAnimation = model.getAnimationSetup(Model.ANIMATIONSETUP_DEFAULT);
							if (defaultAnimation == null) {
								model.addAnimationSetup(Model.ANIMATIONSETUP_DEFAULT, 0, frames - 1, true);
							} else {
								// check default animation setup
								if (defaultAnimation.getStartFrame() != 0 || defaultAnimation.getEndFrame() != frames - 1) {
									System.out.println("Warning: default animation mismatch");
								}
							}
	
							//
							Animation animation = group.createAnimation(fps, frames);
							String valueString = xmlFloatArray.getTextContent();
							t = new StringTokenizer(valueString, " \n\r");
		
							// first frame is not a animation matrix
							Matrix3D frame0Matrix = new Matrix3D(
								Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()), 
								Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
								Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
								Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
								Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()), 
								Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
								Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()), 
								Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken())
							).transpose().invert();
		
							int frame = 0;
							Matrix3D[] transformationsMatrices = animation.getTransformationsMatrices();
							while (t.hasMoreTokens()) {
								// set animation transformation matrix at frame
								transformationsMatrices[frame].set(
									Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()), 
									Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()), 
									Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()), 
									Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken())
								).transpose().multiply(frame0Matrix);
								frame++;
							}
						}
					}
				}
			}
		}

		// parse sub groups
		for(Element _xmlNode: getChildrenByTagName(xmlNode, "node")) {
			Group _group = parseVisualSceneNode(pathName, model, xmlRoot, _xmlNode, fps);
			if (_group != null) {
				group.getSubGroups().put(_group.getId(), _group);
				model.getGroups().put(_group.getId(), _group);
			}
		}

		// check for geometry data
		String xmlInstanceGeometryId = null;
		List<Element> xmlInstanceGeometryElements = getChildrenByTagName(xmlNode, "instance_geometry");
		if (xmlInstanceGeometryElements.isEmpty() == false) {
			Element xmlInstanceGeometryElement = xmlInstanceGeometryElements.get(0);

			// fetch instance geometry url
			xmlInstanceGeometryId = xmlInstanceGeometryElement.getAttribute("url").substring(1); 

			// determine bound materials
			Hashtable<String, String> materialSymbols = new Hashtable<String, String>(); 
			for (Element xmlBindMaterial: getChildrenByTagName(xmlInstanceGeometryElement, "bind_material"))
			for (Element xmlTechniqueCommon: getChildrenByTagName(xmlBindMaterial, "technique_common"))
			for (Element xmlInstanceMaterial: getChildrenByTagName(xmlTechniqueCommon, "instance_material")) {
				materialSymbols.put(
					xmlInstanceMaterial.getAttribute("symbol"),
					xmlInstanceMaterial.getAttribute("target")
				);
			}

			// parse geometry
			parseGeometry(pathName, model, group, xmlRoot, xmlInstanceGeometryId, materialSymbols);

			//
			return group;
		}

		// otherwise check for "instance_node"
		String xmlInstanceNodeId = null;
		for (Element xmlInstanceNodeElement: getChildrenByTagName(xmlNode, "instance_node")) {
			xmlInstanceNodeId = xmlInstanceNodeElement.getAttribute("url").substring(1); 
		}

		// do we have a instance node id?
		if (xmlInstanceNodeId != null) {
			for (Element xmlLibraryNodes: getChildrenByTagName(xmlRoot, "library_nodes"))
			for (Element xmlLibraryNode: getChildrenByTagName(xmlLibraryNodes, "node"))
			if (xmlLibraryNode.getAttribute("id").equals(xmlInstanceNodeId)) {
				// parse sub groups
				for(Element _xmlNode: getChildrenByTagName(xmlLibraryNode, "node")) {
					Group _group = parseVisualSceneNode(pathName, model, xmlRoot, _xmlNode, fps);
					if (_group != null) {
						group.getSubGroups().put(_group.getId(), _group);
						model.getGroups().put(_group.getId(), _group);
					}
				}

				// parse geometry 
				for (Element xmlInstanceGeometry: getChildrenByTagName(xmlLibraryNode, "instance_geometry")) {
					String xmlGeometryId = xmlInstanceGeometry.getAttribute("url").substring(1);
					// parse material symbols
					Hashtable<String, String> materialSymbols = new Hashtable<String, String>(); 
					for (Element xmlBindMaterial: getChildrenByTagName(xmlInstanceGeometry, "bind_material"))
					for (Element xmlTechniqueCommon: getChildrenByTagName(xmlBindMaterial, "technique_common"))
					for (Element xmlInstanceMaterial: getChildrenByTagName(xmlTechniqueCommon, "instance_material")) {
						materialSymbols.put(
							xmlInstanceMaterial.getAttribute("symbol"),
							xmlInstanceMaterial.getAttribute("target")
						);
					}

					// parse geometry
					parseGeometry(pathName, model, group, xmlRoot, xmlGeometryId, materialSymbols);
				}
			}
		}

		//
		return group;
	}

	/**
	 * Parses a instance controller
	 * @param pathName
	 * @param model
	 * @param xmlRoot
	 * @param xmlNode
	 * @return
	 * @throws Exception
	 */
	private static Group parseVisualSceneInstanceController(String pathName, Model model, Element xmlRoot, Element xmlNode) throws Exception {
		StringTokenizer t;

		String xmlNodeId = xmlNode.getAttribute("id");
		String xmlNodeName = xmlNode.getAttribute("name");

		Hashtable<String, String> materialSymbols = new Hashtable<String, String>(); 

		// geometry id
		String xmlGeometryId = null;

		// parse library controllers, find our controller
		List<Element> xmlInstanceControllers = getChildrenByTagName(xmlNode, "instance_controller");
		Element xmlSkin = null;
		Element xmlInstanceController = xmlInstanceControllers.get(0);

		// parse material symbols
		for (Element xmlBindMaterial: getChildrenByTagName(xmlInstanceController, "bind_material"))
		for (Element xmlTechniqueCommon: getChildrenByTagName(xmlBindMaterial, "technique_common"))
		for (Element xmlInstanceMaterial: getChildrenByTagName(xmlTechniqueCommon, "instance_material")) {
			materialSymbols.put(
				xmlInstanceMaterial.getAttribute("symbol"),
				xmlInstanceMaterial.getAttribute("target")
			);
		}

		String xmlInstanceControllerId = xmlInstanceController.getAttribute("url").substring(1);
		Element xmlLibraryControllers = getChildrenByTagName(xmlRoot, "library_controllers").get(0);
		for (Element xmlLibraryController: getChildrenByTagName(xmlLibraryControllers, "controller")) {
			// our controller ?
			if (xmlLibraryController.getAttribute("id").equals(xmlInstanceControllerId)) {
				// parse skin
				List<Element> xmlSkins = getChildrenByTagName(xmlLibraryController, "skin");
				if (xmlSkins.isEmpty() == false) {
					xmlSkin = xmlSkins.get(0);
				}
			}
		}

		// check for xml skin
		if (xmlSkin == null) {
			throw new ParserException("skin not found for instance controller " + xmlNodeId);
		}

		// get geometry id
		xmlGeometryId = xmlSkin.getAttribute("source").substring(1);

		// parse bind shape matrix
		String xmlMatrix = getChildrenByTagName(xmlSkin, "bind_shape_matrix").get(0).getTextContent();
		t = new StringTokenizer(xmlMatrix, " \n\r");

		// 
		Matrix3D bindShapeMatrix = new Matrix3D(
			Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
			Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
			Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()), 
			Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
			Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
			Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()),
			Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken()), 
			Float.parseFloat(t.nextToken()), Float.parseFloat(t.nextToken())
		).transpose();

		// tdme model definitions
		Group group = new Group(
			model,
			xmlNodeId,
			xmlNodeName
		);

		// create skinning
		Skinning skinning = group.createSkinning();

		// parse geometry
		parseGeometry(pathName, model, group, xmlRoot, xmlGeometryId, materialSymbols);

		// parse joints
		String xmlJointsSource = null;
		String xmlJointsInverseBindMatricesSource = null;
		Element xmlJoints = getChildrenByTagName(xmlSkin, "joints").get(0); 
		for(Element xmlJointsInput: getChildrenByTagName(xmlJoints, "input")) {
			if (xmlJointsInput.getAttribute("semantic").equals("JOINT")) {
				xmlJointsSource = xmlJointsInput.getAttribute("source").substring(1);
			} else
			if (xmlJointsInput.getAttribute("semantic").equals("INV_BIND_MATRIX")) {
				xmlJointsInverseBindMatricesSource = xmlJointsInput.getAttribute("source").substring(1);
			}
		}

		// check for joints sources
		if (xmlJointsSource == null) {
			throw new ParserException("joint source not found for instance controller " + xmlNodeId);
		}

		// parse joint ids
		Vector<Joint> joints = new Vector<Joint>();
		for (Element xmlSkinSource: getChildrenByTagName(xmlSkin, "source")) {
			if (xmlSkinSource.getAttribute("id").equals(xmlJointsSource)) {
				t = new StringTokenizer(getChildrenByTagName(xmlSkinSource, "Name_array").get(0).getTextContent(), " \n\r");
				while (t.hasMoreTokens()) {
					joints.add(new Joint(t.nextToken()));
				}
			}
		}
		skinning.setupJoints(joints);

		// check for inverse bind matrices source
		if (xmlJointsInverseBindMatricesSource == null) {
			throw new ParserException("inverse bind matrices source not found for instance controller " + xmlNodeId);
		}

		// parse joints inverse bind matrix
		for (Element xmlSkinSource: getChildrenByTagName(xmlSkin, "source")) {
			if (xmlSkinSource.getAttribute("id").equals(xmlJointsInverseBindMatricesSource)) {
				t = new StringTokenizer(getChildrenByTagName(xmlSkinSource, "float_array").get(0).getTextContent(), " \n\r");
				Joint[] _joints = skinning.getJoints();
				for (int i = 0; i < _joints.length; i++) {
					_joints[i].getBindMatrix().set(
						bindShapeMatrix.clone().multiply(
							// inverse bind matrix
							new Matrix3D(
								Float.parseFloat(t.nextToken()),Float.parseFloat(t.nextToken()),
								Float.parseFloat(t.nextToken()),Float.parseFloat(t.nextToken()),
								Float.parseFloat(t.nextToken()),Float.parseFloat(t.nextToken()),
								Float.parseFloat(t.nextToken()),Float.parseFloat(t.nextToken()),
								Float.parseFloat(t.nextToken()),Float.parseFloat(t.nextToken()),
								Float.parseFloat(t.nextToken()),Float.parseFloat(t.nextToken()),
								Float.parseFloat(t.nextToken()),Float.parseFloat(t.nextToken()),
								Float.parseFloat(t.nextToken()),Float.parseFloat(t.nextToken())
							).transpose()
						)
					);
				}
			}
		}

		// read vertex influences
		Vector<Float> weights = new Vector<Float>();
		int xmlJointOffset = -1;
		int xmlWeightOffset = -1;
		String xmlWeightsSource = null;
		Element xmlVertexWeights = getChildrenByTagName(xmlSkin, "vertex_weights").get(0);
		List<Element> xmlVertexWeightInputs = getChildrenByTagName(xmlVertexWeights, "input");
		for(Element xmlVertexWeightInput: xmlVertexWeightInputs) {
			if (xmlVertexWeightInput.getAttribute("semantic").equals("JOINT")) {
				if (xmlVertexWeightInput.getAttribute("source").substring(1).equals(xmlJointsSource) == false) {
					throw new ParserException("joint inverse bind matrices source do not match");
				}
				xmlJointOffset = Integer.parseInt(xmlVertexWeightInput.getAttribute("offset")); 
			} else
			if (xmlVertexWeightInput.getAttribute("semantic").equals("WEIGHT")) {
				xmlWeightOffset = Integer.parseInt(xmlVertexWeightInput.getAttribute("offset"));
				xmlWeightsSource = xmlVertexWeightInput.getAttribute("source").substring(1); 
			}
		}

		// check for vertex weight parameter 
		if (xmlJointOffset == -1) {
			throw new ParserException("xml vertext weight joint offset missing for node " + xmlNodeId);
		}
		if (xmlWeightOffset == -1) {
			throw new ParserException("xml vertext weight weight offset missing for node " + xmlNodeId);
		}
		if (xmlWeightsSource == null) {
			throw new ParserException("xml vertext weight weight source missing for node " + xmlNodeId);
		}

		// parse weights
		for (Element xmlSkinSource: getChildrenByTagName(xmlSkin, "source")) {
			if (xmlSkinSource.getAttribute("id").equals(xmlWeightsSource)) {
				t = new StringTokenizer(getChildrenByTagName(xmlSkinSource, "float_array").get(0).getTextContent(), " \n\r");
				while (t.hasMoreTokens()) {
					weights.add(new Float(Float.parseFloat(t.nextToken())));
				}
			}
		}
		skinning.setupWeights(weights);

		// actually do parse joint influences of each vertex
		int xmlVertexWeightInputCount = xmlVertexWeightInputs.size();
		String vertexJointsInfluenceCountString = getChildrenByTagName(xmlVertexWeights, "vcount").get(0).getTextContent();
		String vertexJointsInfluencesString = getChildrenByTagName(xmlVertexWeights, "v").get(0).getTextContent();

		t = new StringTokenizer(vertexJointsInfluenceCountString, " \n\r");
		StringTokenizer t2 = new StringTokenizer(vertexJointsInfluencesString, " \n\r");
		int offset = 0;
		Vector<Vector<JointWeight>> verticesJointsWeights = new Vector<Vector<JointWeight>>();
		while (t.hasMoreTokens()) {
			// read joint influences for current vertex
			int vertexJointsInfluencesCount = Integer.parseInt(t.nextToken());
			Vector<JointWeight> vertexJointsWeights = new Vector<JointWeight>();
			for (int i = 0; i < vertexJointsInfluencesCount; i++) {
				int vertexJoint = -1;
				int vertexWeight = -1;
				while (vertexJoint == -1 || vertexWeight == -1) {
					int value = Integer.parseInt(t2.nextToken());
					if (offset % xmlVertexWeightInputCount == xmlJointOffset) {
						vertexJoint = value;
					} else
					if (offset % xmlVertexWeightInputCount == xmlWeightOffset) {
						vertexWeight = value;
					}
					offset++;
				}
				vertexJointsWeights.add(new JointWeight(vertexJoint, vertexWeight));
			}
			verticesJointsWeights.add(vertexJointsWeights);
		}
		skinning.setupVerticesJointsWeights(verticesJointsWeights);

		return group;
	}

	/**
	 * Parses a geometry
	 * @param pathName
	 * @param model
	 * @param group
	 * @param xmlRoot
	 * @param xmlNodeId
	 */
	public static void parseGeometry(String pathName, Model model, Group group, Element xmlRoot, String xmlNodeId, Hashtable<String, String> materialSymbols) throws Exception {
		StringTokenizer t;

		//
		FacesEntity facesEntity = null;
		Vector<FacesEntity> facesEntities = new Vector<FacesEntity>(Arrays.asList(group.getFacesEntities()));

		int verticesOffset = group.getVertices().length;
		Vector<Vector3D> vertices = new Vector<Vector3D>(Arrays.asList(group.getVertices()));

		int normalsOffset = group.getNormals().length;
		Vector<Vector3D> normals = new Vector<Vector3D>(Arrays.asList(group.getNormals()));

		int textureCoordinatesOffset =
			group.getTextureCoordinates() != null?
			group.getTextureCoordinates().length:
			0;
		Vector<TextureCoordinate> textureCoordinates =
			group.getTextureCoordinates() != null?
			new Vector<TextureCoordinate>(Arrays.asList(group.getTextureCoordinates())):
			new Vector<TextureCoordinate>();

		Vector<Vector<Integer>> verticesNormalIndices = new Vector<Vector<Integer>>();
		for (int[] vertexNormalsIndices: group.getVerticesNormalIndices()) {
			Vector<Integer> vertexNormalIndicesVector = new Vector<Integer>();
			for (int vertexNormalIndex: vertexNormalsIndices) {
				vertexNormalIndicesVector.add(new Integer(vertexNormalIndex));
			}
			verticesNormalIndices.add(vertexNormalIndicesVector);
		}

		Element xmlLibraryGeometries = getChildrenByTagName(xmlRoot, "library_geometries").get(0);
		for (Element xmlGeometry: getChildrenByTagName(xmlLibraryGeometries, "geometry")) {
			if (xmlGeometry.getAttribute("id").equals(xmlNodeId)) {
				Element xmlMesh = getChildrenByTagName(xmlGeometry, "mesh").get(0);

				Vector<Element> xmlPolygonsList = new Vector<Element>(); 

				// try to read from triangles
				for (Element xmlTriangesElement: getChildrenByTagName(xmlMesh, "triangles")) {
					xmlPolygonsList.add(xmlTriangesElement);
				}

				// try to read from polylist
				for (Element xmlPolyListElement: getChildrenByTagName(xmlMesh, "polylist")) {
					xmlPolygonsList.add(xmlPolyListElement);
				}

				// parse from xml polygons elements
				for(Element xmlPolygons: xmlPolygonsList) {
					Vector<Face>faces = new Vector<Face>();
					facesEntity = new FacesEntity(group, xmlNodeId);
					if (xmlPolygons.getNodeName().toLowerCase().equals("polylist")) {
						t = new StringTokenizer(getChildrenByTagName(xmlPolygons, "vcount").get(0).getTextContent());
						while (t.hasMoreTokens()) {
							int vertexCount = Integer.parseInt(t.nextToken());
							if (vertexCount != 3) {
								throw new ParserException("we only support triangles in " + xmlNodeId);
							}
						}
					}
					// parse triangles
					int xmlInputs = -1;
					int xmlVerticesOffset = -1;
					String xmlVerticesSource = null;
	
					int xmlNormalsOffset = -1;
					String xmlNormalsSource = null;
	
					int xmlTexCoordOffset = -1;
					String xmlTexCoordSource = null;

					// material
					String xmlMaterialId = xmlPolygons.getAttribute("material");
					String materialSymbol = materialSymbols.get(xmlMaterialId);
					if (materialSymbol != null) xmlMaterialId = materialSymbol.substring(1);
					if (xmlMaterialId != null && xmlMaterialId.length() > 0) {
						Material material = model.getMaterials().get(xmlMaterialId);
						if (material == null) {
							// parse material as we do not have it yet
							material = parseMaterial(pathName, model, xmlRoot, xmlMaterialId);
						}
						// set it up
						facesEntity.setMaterial(material);
					}

					List<Element> xmlTrianglesInputs = getChildrenByTagName(xmlPolygons, "input");
					xmlInputs = xmlTrianglesInputs.size();
					for (Element xmlTrianglesInput: xmlTrianglesInputs) {
						// check for vertices sources
						if (xmlTrianglesInput.getAttribute("semantic").equals("VERTEX")) {
							xmlVerticesOffset = Integer.parseInt(xmlTrianglesInput.getAttribute("offset"));
							xmlVerticesSource = xmlTrianglesInput.getAttribute("source").substring(1);
						} else
						// check for normals sources
						if (xmlTrianglesInput.getAttribute("semantic").equals("NORMAL")) {
							xmlNormalsOffset = Integer.parseInt(xmlTrianglesInput.getAttribute("offset"));
							xmlNormalsSource = xmlTrianglesInput.getAttribute("source").substring(1);
						}
						// check for texture coordinate sources
						if (xmlTrianglesInput.getAttribute("semantic").equals("TEXCOORD")) {
							xmlTexCoordOffset = Integer.parseInt(xmlTrianglesInput.getAttribute("offset"));
							xmlTexCoordSource = xmlTrianglesInput.getAttribute("source").substring(1);
						}
					}

					// get vertices source
					for (Element xmlVertices: getChildrenByTagName(xmlMesh, "vertices")) {
						if (xmlVertices.getAttribute("id").equals(xmlVerticesSource)) {
							for (Element xmlVerticesInput: getChildrenByTagName(xmlVertices, "input")) {
								if (xmlVerticesInput.getAttribute("semantic").equalsIgnoreCase("position")) {
									xmlVerticesSource = xmlVerticesInput.getAttribute("source").substring(1);
								} else
								if (xmlVerticesInput.getAttribute("semantic").equalsIgnoreCase("normal")) {
									xmlNormalsSource = xmlVerticesInput.getAttribute("source").substring(1);
								}									
							}
						}
					}

					// check for triangles vertices sources
					if (xmlVerticesSource == null) {
						throw new ParserException("Could not determine triangles vertices source for '" + xmlNodeId + "'");
					}
	
					// check for triangles normals sources
					if (xmlNormalsSource == null) {
						throw new ParserException("Could not determine triangles normal source for '" + xmlNodeId + "'");
					}
	
					// load vertices, normals, texture coordinates
					for (Element xmlMeshSource: getChildrenByTagName(xmlMesh, "source")) {
						// vertices
						if (xmlMeshSource.getAttribute("id").equals(xmlVerticesSource)) {
							Element xmlFloatArray = getChildrenByTagName(xmlMeshSource, "float_array").get(0);
							String valueString = xmlFloatArray.getTextContent();
							t = new StringTokenizer(valueString, " \n\r");
							while (t.hasMoreTokens()) {
								Vector3D v = new Vector3D(
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken())
								);
								vertices.add(v);
							}
						} else
						// normals
						if (xmlMeshSource.getAttribute("id").equals(xmlNormalsSource)) {
							Element xmlFloatArray = getChildrenByTagName(xmlMeshSource, "float_array").get(0);
							String valueString = xmlFloatArray.getTextContent();
							t = new StringTokenizer(valueString, " \n\r");
							while (t.hasMoreTokens()) {
								Vector3D v = new Vector3D(
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken())
								);
								normals.add(v);
							}
						}
						// texture coordinates
						if (xmlTexCoordSource != null) {
							if (xmlMeshSource.getAttribute("id").equals(xmlTexCoordSource)) {
								Element xmlFloatArray = getChildrenByTagName(xmlMeshSource, "float_array").get(0);
								String valueString = xmlFloatArray.getTextContent();
								t = new StringTokenizer(valueString, " \n\r");
								while (t.hasMoreTokens()) {
									TextureCoordinate tc = new TextureCoordinate(
										Float.parseFloat(t.nextToken()),
										Float.parseFloat(t.nextToken())
									);
									textureCoordinates.add(tc);
								}
							}
						}
					}

					// setup vertex normal mapping
					for(int i = 0; i < vertices.size(); i++) {					
						verticesNormalIndices.add(new Vector<Integer>());
					}

					// load faces
					String valueString = getChildrenByTagName(xmlPolygons, "p").get(0).getTextContent();
					t = new StringTokenizer(valueString, " \n\r");
					int vi[] = new int[3];
					int viIdx = 0;
					int ni[] = new int[3];
					int niIdx = 0;
					int ti[] = xmlTexCoordSource == null?null:new int[3];
					int tiIdx = 0;
					int valueIdx = 0;
					while (t.hasMoreTokens()) {
						boolean valid = true;
						int value = Integer.parseInt(t.nextToken());
						if (valueIdx % xmlInputs == xmlVerticesOffset) {
							vi[viIdx++] = value;
							// validate
							if (value < 0 || value >= vertices.size() - verticesOffset) {
								valid = false;
							}
							// fix for some strange models
							if (xmlNormalsSource != null && xmlNormalsOffset == -1) {
								ni[niIdx++] = value;
								// validate
								if (value < 0 || value >= normals.size() - normalsOffset) {
									valid = false;
								}
							}
						}
						if (xmlNormalsOffset != -1 && valueIdx % xmlInputs == xmlNormalsOffset) {
							ni[niIdx++] = value;
							// validate
							if (value < 0 || value >= normals.size() - normalsOffset) {
								valid = false;
							}
						}
						if (xmlTexCoordOffset != -1 && valueIdx % xmlInputs == xmlTexCoordOffset) {
							ti[tiIdx++] = value;
							// validate
							if (value < 0 || value >= textureCoordinates.size() - textureCoordinatesOffset) {
								valid = false;
							}
						}
						if (viIdx == 3 && niIdx == 3 && (ti == null || tiIdx == 3)) {
							// only add valid faces
							if (valid) {
								// assign normals to vertices
								for (int i = 0; i < 3; i++) {
									int vertexIndex = vi[i] + verticesOffset;
									Integer normalIndex = ni[i] + normalsOffset;
									Vector<Integer> vertexNormalIndices = verticesNormalIndices.get(vertexIndex);
									if (vertexNormalIndices.contains(normalIndex) == false) {
										vertexNormalIndices.add(normalIndex);
									}
								}
								// add face
								Face f = new Face(
									group,
									vi[0] + verticesOffset,
									vi[1] + verticesOffset,
									vi[2] + verticesOffset,
									ni[0] + normalsOffset,
									ni[1] + normalsOffset,
									ni[2] + normalsOffset
								);
								if (ti != null) {
									f.setTextureCoordinateIndices(
										ti[0] + textureCoordinatesOffset,
										ti[1] + textureCoordinatesOffset,
										ti[2] + textureCoordinatesOffset
									);
								}
								faces.add(f);
							}
							viIdx = 0;
							niIdx = 0;
							tiIdx = 0;
						}
						valueIdx++;
					}
					facesEntity.setupFaces(faces);
					facesEntities.add(facesEntity);
				}
			}
		}

		// set up group
		group.setupVertices(vertices);
		group.setupNormals(normals);
		group.setupVerticesNormalIndices(verticesNormalIndices);
		group.setupTextureCoordinates(textureCoordinates);
		group.setupFacesEntities(facesEntities);
	}

	/**
	 * Parses a material
	 * @param pathName
	 * @param model
	 * @param xmlRoot
	 * @param xmlNodeId
	 * @throws Exception
	 */
	public static Material parseMaterial(String pathName, Model model, Element xmlRoot, String xmlNodeId) throws Exception {
		// determine effect id
		String xmlEffectId = null;
		Element xmlLibraryMaterials = getChildrenByTagName(xmlRoot, "library_materials").get(0);
		for (Element xmlMaterial: getChildrenByTagName(xmlLibraryMaterials, "material")) {
			if (xmlMaterial.getAttribute("id").equals(xmlNodeId)) {
				Element xmlInstanceEffect = getChildrenByTagName(xmlMaterial, "instance_effect").get(0);
				xmlEffectId = xmlInstanceEffect.getAttribute("url").substring(1); 
			}
		}

		if (xmlEffectId == null) {
			System.out.println("Could not determine effect id for '" + xmlNodeId + "'");
			return null;
		}

		// parse effect
		Material material = new Material(xmlNodeId);
		String xmlDiffuseTextureId = null;
		Element xmlLibraryEffects = getChildrenByTagName(xmlRoot, "library_effects").get(0);
		for (Element xmlEffect: getChildrenByTagName(xmlLibraryEffects, "effect")) {
			if (xmlEffect.getAttribute("id").equals(xmlEffectId)) {
				// diffuse texture
				Element xmlProfile = getChildrenByTagName(xmlEffect, "profile_COMMON").get(0);
				Hashtable<String, String> samplerSurfaceMapping = new Hashtable<String, String>();
				Hashtable<String, String> surfaceImageMapping = new Hashtable<String, String>();
				for(Element xmlNewParam: getChildrenByTagName(xmlProfile, "newparam")) {
					String xmlNewParamSID = xmlNewParam.getAttribute("sid");
					for(Element xmlSurface: getChildrenByTagName(xmlNewParam, "surface"))
					for(Element xmlSurfaceInitFrom: getChildrenByTagName(xmlSurface, "init_from")) {
						surfaceImageMapping.put(xmlNewParamSID, xmlSurfaceInitFrom.getTextContent());
					}
					for(Element xmlSampler2D: getChildrenByTagName(xmlNewParam, "sampler2D"))
					for(Element xmlSampler2DSource: getChildrenByTagName(xmlSampler2D, "source")) {
						samplerSurfaceMapping.put(xmlNewParamSID, xmlSampler2DSource.getTextContent());
					}					
				}

				// 
				for (Element xmlTechnique: getChildrenByTagName(xmlProfile, "technique")) {
					NodeList xmlTechniqueNodes = xmlTechnique.getChildNodes();
					for (int i = 0; i < xmlTechniqueNodes.getLength(); i++) {
						Node xmlTechniqueNode = xmlTechniqueNodes.item(i);

						// skip if not an element
						if (xmlTechniqueNode.getNodeType() != Node.ELEMENT_NODE) continue;

						// diffuse
						for (Element xmlDiffuse: getChildrenByTagName((Element)xmlTechniqueNode, "diffuse")) {
							// texture
							for (Element xmlTexture: getChildrenByTagName(xmlDiffuse, "texture")) {
								xmlDiffuseTextureId = xmlTexture.getAttribute("texture");
								String sample2Surface = samplerSurfaceMapping.get(xmlDiffuseTextureId);
								String surface2Image = null;
								if (sample2Surface != null) surface2Image = surfaceImageMapping.get(sample2Surface);
								if (surface2Image != null) xmlDiffuseTextureId = surface2Image; 
							}
							// color
							for (Element xmlColor: getChildrenByTagName(xmlDiffuse, "color")) {
								StringTokenizer t = new StringTokenizer(xmlColor.getTextContent(), " ");
								material.getDiffuseColor().set(
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken())
								);
							}
						}
						// ambient
						for (Element xmlAmbient: getChildrenByTagName((Element)xmlTechniqueNode, "ambient")) {
							// color
							for (Element xmlColor: getChildrenByTagName(xmlAmbient, "color")) {
								StringTokenizer t = new StringTokenizer(xmlColor.getTextContent(), " ");
								material.getAmbientColor().set(
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken())
								);
							}
						}
						// emission
						for (Element xmlEmission: getChildrenByTagName((Element)xmlTechniqueNode, "emission")) {
							// color
							for (Element xmlColor: getChildrenByTagName(xmlEmission, "color")) {
								StringTokenizer t = new StringTokenizer(xmlColor.getTextContent(), " ");
								material.getEmissionColor().set(
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken())
								);
							}
						}
						// specular
						for (Element xmlSpecular: getChildrenByTagName((Element)xmlTechniqueNode, "specular")) {
							// color
							for (Element xmlColor: getChildrenByTagName(xmlSpecular, "color")) {
								StringTokenizer t = new StringTokenizer(xmlColor.getTextContent(), " ");
								material.getSpecularColor().set(
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken()),
									Float.parseFloat(t.nextToken())
								);
							}
						}
						// shininess
						for (Element xmlShininess: getChildrenByTagName((Element)xmlTechniqueNode, "shininess")) {
							// color
							for (Element xmlFloat: getChildrenByTagName(xmlShininess, "float")) {
								material.setShininess(Float.parseFloat(xmlFloat.getTextContent()));
							}
						}
					}
				}
			}
		}

		// do we have a diffuse texture?
		String xmlDiffuseTextureFilename = null;
		if (xmlDiffuseTextureId != null) {
			// yep
			Element xmlLibraryImages = getChildrenByTagName(xmlRoot, "library_images").get(0);
			for (Element xmlImage: getChildrenByTagName(xmlLibraryImages, "image")) {
				if (xmlImage.getAttribute("id").equals(xmlDiffuseTextureId)) {
					xmlDiffuseTextureFilename = getChildrenByTagName(xmlImage, "init_from").get(0).getTextContent();
					break;
				}
			}
		}

		// do we have a file name
		if (xmlDiffuseTextureFilename != null) {
			// yep, texture must be in the same path as DAE, so cut path off
			int indexSlash = xmlDiffuseTextureFilename.lastIndexOf("/");
			int indexBackslash = xmlDiffuseTextureFilename.lastIndexOf("\\");
			if (indexSlash != -1 && indexBackslash != -1) {
				if (indexSlash > indexBackslash) {
					xmlDiffuseTextureFilename = xmlDiffuseTextureFilename.substring(indexSlash + 1);
				} else {
					xmlDiffuseTextureFilename = xmlDiffuseTextureFilename.substring(indexBackslash + 1);
				}
			}

			// add texture
			material.setupDiffuseTexture(pathName, xmlDiffuseTextureFilename);
		}

		// add material to library
		model.getMaterials().put(material.getId(), material);

		//
		return material;
	}

	/**
	 * Returns immediate children by tagnames of parent
	 * @param parent
	 * @param name
	 * @return children with given name
	 */
	public static List<Element> getChildrenByTagName(Element parent, String name) {
		List<Element> nodeList = new ArrayList<Element>();
		for (Node child = parent.getFirstChild(); child != null; child = child.getNextSibling()) {
			if (child.getNodeType() == Node.ELEMENT_NODE && name.equals(child.getNodeName())) {
				nodeList.add((Element) child);
			}
		}
		return nodeList;
	}

	/**
	 * Converts an element to string
	 * @param node
	 * @return string representation
	 */
	private static String nodeToString(Node node) {
		NamedNodeMap attributes = node.getAttributes();
		String attributesString = "";
		for(int i = 0; i < attributes.getLength(); i++) {
			if (i > 0) {
				attributesString+=", ";
			}
			attributesString+= attributes.item(i).getNodeName();
			attributesString+= "=";
			attributesString+= attributes.item(i).getNodeValue();
		}
		return
			"[" +
			"name=" + node.getNodeName() +
			", attributes=[" + attributesString + "]" +
			"]";
	}

}
