package net.drewke.tdme.parser;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.model.Face;
import net.drewke.tdme.model.FacesEntity;
import net.drewke.tdme.model.Group;
import net.drewke.tdme.model.Material;
import net.drewke.tdme.model.Model;
import net.drewke.tdme.model.TextureCoordinate;


/**
 * Wavefront object Parser
 * @author andreas.drewke
 * @version $Id: 4c8c0d1dfd1d3c9309c434b4e0bc3f8fdff9ac0d $
 */
public class WFObjParser {

	/**
	 * Parses a wave front object file
	 * 	Does no validation yet
	 * @param pathName
	 * @param fileName
	 * @return TDObject instance
	 * @throws IOException
	 */
	public static Model parse(String pathName, String fileName) throws IOException, ParserException {
		// create object
		Model model = new Model(fileName, fileName);	

		Vector<Vector3D> vertices = new Vector<Vector3D>();
		Vector<TextureCoordinate> textureCoordinates = new Vector<TextureCoordinate>();

		Hashtable<String, Material> materials = model.getMaterials();
		Hashtable<String, Group> subGroups = model.getSubGroups();
		Hashtable<String, Group> groups = model.getGroups();

		// current group
		Group group = null;

		// model vertices -> group vertices mapping
		Hashtable<Integer, Integer> modelGroupVerticesMapping = null;
		// model texture coordinates -> group texture coordinates mapping
		Hashtable<Integer, Integer> modelGroupTextureCoordinatesMapping = null;

		// current group data
		Vector<Face> groupFacesEntityFaces = null;
		Vector<Vector3D> groupVertices = null;
		Vector<Vector3D> groupNormals = null;
		Vector<TextureCoordinate> groupTextureCoordinates = null;

		// current group's faces entity
		Vector<FacesEntity> groupFacesEntities = null;
		FacesEntity groupFacesEntity = null;

		//
		DataInputStream inputStream = new DataInputStream(new FileInputStream(pathName + "/" + fileName));
		BufferedReader reader = new BufferedReader(new InputStreamReader(inputStream));

		//
		try {
			//
			String line;
			while ((line = reader.readLine()) != null) {
				line = line.trim();

				// skip on comments
				if (line.startsWith("#")) {
					continue;
				}

				// determine index of first ' ' which will separate command from arguments
				int commandEndIdx = line.indexOf(' ');
				if (commandEndIdx == -1) commandEndIdx = line.length();

				// determine command
				String command = line.substring(0, commandEndIdx).trim().toLowerCase();

				// determine arguments if any exist
				String arguments = command.length() + 1 > line.length()?"":line.substring(command.length() + 1);

				// parse
				if (command.equals("mtllib")) {
					String materialFileName = arguments;
					materials = WFObjParser.parseMaterials(pathName, materialFileName);
				} else
				if (command.equals("v")) {
					StringTokenizer t = new StringTokenizer(arguments, " ");
					float x = Float.parseFloat(t.nextToken());
					float y = Float.parseFloat(t.nextToken());
					float z = Float.parseFloat(t.nextToken());
					// add vertex
					vertices.add(new Vector3D(x, y, z));
				} else
				if (command.equals("vt")) {
					StringTokenizer t = new StringTokenizer(arguments, " ");
					float u = Float.parseFloat(t.nextToken());
					float v = Float.parseFloat(t.nextToken());
					textureCoordinates.add(new TextureCoordinate(u, v));
				} else
				if (command.equals("f")) {
					StringTokenizer t2;
					StringTokenizer t = new StringTokenizer(arguments, " ");
					int v0 = -1;
					int v1 = -1;
					int v2 = -1;
					int vt0 = -1;
					int vt1 = -1;
					int vt2 = -1;

					// parse vertex index 0, vertex texture index 0
					t2 = new StringTokenizer(t.nextToken(), "/");
					v0 = Integer.parseInt(t2.nextToken()) - 1;
					if (t2.hasMoreTokens()) {
						vt0 = Integer.parseInt(t2.nextToken()) - 1;
					}
					// parse vertex index 1, vertex texture index 1
					t2 = new StringTokenizer(t.nextToken(), "/");
					v1 = Integer.parseInt(t2.nextToken()) - 1;
					if (t2.hasMoreTokens()) {
						vt1 = Integer.parseInt(t2.nextToken()) - 1;
					}
					// parse vertex index 2, vertex texture index 2
					t2 = new StringTokenizer(t.nextToken(), "/");
					v2 = Integer.parseInt(t2.nextToken()) - 1;
					if (t2.hasMoreTokens()) {
						vt2 = Integer.parseInt(t2.nextToken()) - 1;
					}
	
					// check if triangulated
					if (t.hasMoreTokens()) {
						throw new ParserException("We only support triangulated meshes");
					}

					Integer mappedVertex = null;

					// map v0 to group
					mappedVertex = modelGroupVerticesMapping.get(v0);
					if (mappedVertex == null) {
						groupVertices.add(vertices.get(v0).clone());
						v0 = groupVertices.size() - 1;
					} else {
						v0 = mappedVertex.intValue();
					}

					// map v1 to group
					mappedVertex = modelGroupVerticesMapping.get(v1);
					if (mappedVertex == null) {
						groupVertices.add(vertices.get(v1).clone());
						v1 = groupVertices.size() - 1;
					} else {
						v1 = mappedVertex.intValue();
					}

					// map v2 to group
					mappedVertex = modelGroupVerticesMapping.get(v2);
					if (mappedVertex == null) {
						groupVertices.add(vertices.get(v2).clone());
						v2 = groupVertices.size() - 1;
					} else {
						v2 = mappedVertex.intValue();
					}

					//
					Integer mappedTextureCoordinate = null;

					// map vt0 to group
					mappedTextureCoordinate = modelGroupTextureCoordinatesMapping.get(vt0);
					if (mappedTextureCoordinate == null) {
						groupTextureCoordinates.add(textureCoordinates.get(vt0).clone());
						vt0 = groupTextureCoordinates.size() - 1;
					} else {
						vt0 = mappedTextureCoordinate.intValue();
					}

					// map vt1 to group
					mappedTextureCoordinate = modelGroupTextureCoordinatesMapping.get(vt1);
					if (mappedTextureCoordinate == null) {
						groupTextureCoordinates.add(textureCoordinates.get(vt1).clone());
						vt1 = groupTextureCoordinates.size() - 1;
					} else {
						vt1 = mappedTextureCoordinate.intValue();
					}

					// map vt2 to group
					mappedTextureCoordinate = modelGroupTextureCoordinatesMapping.get(vt2);
					if (mappedTextureCoordinate == null) {
						groupTextureCoordinates.add(textureCoordinates.get(vt2).clone());
						vt2 = groupTextureCoordinates.size() - 1;
					} else {
						vt2 = mappedTextureCoordinate.intValue();
					}

					// compute face normal,
					//	TODO: move face vertex computation into a own class
					Vector3D faceVertices[] = {groupVertices.get(v0), groupVertices.get(v1), groupVertices.get(v2)};
					Vector3D faceNormal = Vector3D.computeCrossProduct(
						faceVertices[2].clone().sub(faceVertices[0]),
						faceVertices[1].clone().sub(faceVertices[0])
					).normalize();

					// compute vertex normal
					Vector3D[] faceVertexNormals = new Vector3D[3];
					for (int i = 0; i < faceVertices.length; i++) {
						faceVertexNormals[i] = faceVertices[i].clone().add(faceNormal).normalize();
					}

					// store group normals
					int n0 = groupNormals.size();
					groupNormals.add(faceVertexNormals[0]);
					int n1 = groupNormals.size();
					groupNormals.add(faceVertexNormals[1]);
					int n2 = groupNormals.size();
					groupNormals.add(faceVertexNormals[2]);

					// create face with vertex indices
					//	we only support triangulated faces
					Face face = new Face(group, v0, v1, v2, n0, n1, n2);
					if (vt0 != -1 && vt1 != -1 && vt2 != -1) {
						// set optional texture coordinate index
						face.setTextureCoordinateIndices(vt0, vt1, vt2);
					}
					groupFacesEntityFaces.add(face);
				} else
				if (command.equals("g")) {
					if (group != null) {
						// current faces entity
						groupFacesEntity.setupFaces(groupFacesEntityFaces);
						groupFacesEntities.add(groupFacesEntity);

						// group
						group.setupVertices(groupVertices);
						group.setupNormals(groupNormals);
						group.setupTextureCoordinates(groupTextureCoordinates);
						group.setupFacesEntities(groupFacesEntities);					
					}
					StringTokenizer t = new StringTokenizer(arguments, " ");
					String name = t.nextToken();
					groupVertices = new Vector<Vector3D>();
					groupNormals = new Vector<Vector3D>();
					groupTextureCoordinates = new Vector<TextureCoordinate>();
					groupFacesEntityFaces = new Vector<Face>();
					group = new Group(
						model,
						name,
						name
					);
					groupFacesEntity = new FacesEntity(group, name);
					groupFacesEntities = new Vector<FacesEntity>();
					modelGroupVerticesMapping = new Hashtable<Integer, Integer>();
					modelGroupTextureCoordinatesMapping = new Hashtable<Integer, Integer>();
					subGroups.put(name, group);
					groups.put(name, group);
				} else
				if (command.equals("usemtl")) {
					if (group != null) {
						// current faces entity
						groupFacesEntity.setupFaces(groupFacesEntityFaces);
						groupFacesEntities.add(groupFacesEntity);
						groupFacesEntity = new FacesEntity(group, "#" + groupFacesEntities.size());
						groupFacesEntityFaces = new Vector<Face>();
					}
					groupFacesEntity.setMaterial(materials.get(arguments));
				} else {
					// not supported
				}
	
			}
	
			// finish last group
			if (group != null) {
				// current faces entity
				groupFacesEntity.setupFaces(groupFacesEntityFaces);
				groupFacesEntities.add(groupFacesEntity);

				// group
				group.setupVertices(groupVertices);
				group.setupNormals(groupNormals);
				group.setupTextureCoordinates(groupTextureCoordinates);
				group.setupFacesEntities(groupFacesEntities);					
			}
		} finally {
			// close resouces
			reader.close();
			inputStream.close();
		}

		//
		return model;
	}

	/**
	 * Parse a wavefront object material library
	 * @param pathName
	 * @param fileName
	 * @return
	 */
	private static Hashtable<String, Material> parseMaterials(String pathName, String fileName) throws IOException {
		Hashtable<String, Material> materials = new Hashtable<String, Material>();
		Material current = null;
		//
		DataInputStream inputStream = new DataInputStream(new FileInputStream(pathName + "/" + fileName));
		BufferedReader reader = new BufferedReader(new InputStreamReader(inputStream));
		String line;
		float alpha = 1.0f;
		while ((line = reader.readLine()) != null) {
			line = line.trim();

			// skip on comments
			if (line.startsWith("#")) {
				continue;
			}

			// determine index of first ' ' which will separate command from arguments
			int commandEndIdx = line.indexOf(' ');
			if (commandEndIdx == -1) commandEndIdx = line.length();

			// determine command
			String command = line.substring(0, commandEndIdx).trim().toLowerCase();

			// determine arguments if any exist
			String arguments = command.length() + 1 > line.length()?"":line.substring(command.length() + 1);

			// parse
			if (command.equals("newmtl")) {
				String name = arguments;
				current = new Material(name);
				materials.put(name, current);
			} else
			if (command.equals("map_ka")) {
				current.setupDiffuseTexture(pathName, arguments);
			} else
			if (command.equals("map_kd")) {
				current.setupDiffuseTexture(pathName, arguments);
			} else
			if (command.equals("ka")) {
				StringTokenizer t = new StringTokenizer(arguments, " ");
				current.getAmbientColor().set(
					Float.parseFloat(t.nextToken()),
					Float.parseFloat(t.nextToken()),
					Float.parseFloat(t.nextToken()),
					alpha
				);
			} else
			if (command.equals("kd")) {
				StringTokenizer t = new StringTokenizer(arguments, " ");
				current.getDiffuseColor().set(
					Float.parseFloat(t.nextToken()),
					Float.parseFloat(t.nextToken()),
					Float.parseFloat(t.nextToken()),
					alpha
				);
			} else
			if (command.equals("ks")) {
				StringTokenizer t = new StringTokenizer(arguments, " ");
				current.getSpecularColor().set(
					Float.parseFloat(t.nextToken()),
					Float.parseFloat(t.nextToken()),
					Float.parseFloat(t.nextToken()),
					alpha
				);
			} else
			if (command.equals("tr")) {
				alpha = Float.parseFloat(arguments);
				current.getAmbientColor().setAlpha(alpha);
				current.getDiffuseColor().setAlpha(alpha);
				current.getSpecularColor().setAlpha(alpha);
				current.getEmissionColor().setAlpha(alpha);
			} else
			if (command.equals("d")) {
				alpha = Float.parseFloat(arguments);
				current.getAmbientColor().setAlpha(alpha);
				current.getDiffuseColor().setAlpha(alpha);
				current.getSpecularColor().setAlpha(alpha);
				current.getEmissionColor().setAlpha(alpha);
			}
		}

		// Close the input stream
		inputStream.close();

		//
		return materials;
	}

}
