package net.drewke.tdme.physics;

import net.drewke.tdme.math.SeparatingAxisTheorem;
import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.primitives.BoundingBox;
import net.drewke.tdme.primitives.Capsule;
import net.drewke.tdme.primitives.LineSegment;
import net.drewke.tdme.primitives.OrientedBoundingBox;
import net.drewke.tdme.primitives.Sphere;
import net.drewke.tdme.primitives.Triangle;

/**
 * Collision detection
 * @author Andreas Drewke
 * @version $Id: 776024107efbc38917ba1e3f1fdaaf39b24199a9 $
 */
public class CollisionDetection {

	/**
	 * Returns if axis aligned bounding boxes do collide
	 * @param axis aligned bounding box 1
	 * @param axis aligned bounding box 2
	 * @return collision response of axis aligned bounding box 1 and 2 or null 
	 */
	public static CollisionResponse doCollide(BoundingBox b1, BoundingBox b2) {
		// see
		//	http://www.gamedev.net/topic/567310-platform-game-collision-detection/
		float b1MinXYZ[] = b1.getMin().getArray();
		float b1MaxXYZ[] = b1.getMax().getArray();
		float b2MinXYZ[] = b2.getMin().getArray();
		float b2MaxXYZ[] = b2.getMax().getArray();

		// face distances
		float faceDistances[] = {
			b2MaxXYZ[0] - b1MinXYZ[0],	// left face normal / -x
			b1MaxXYZ[0] - b2MinXYZ[0],  // right face normal / +x
			b2MaxXYZ[1] - b1MinXYZ[1],	// bottom face normal / -y
			b1MaxXYZ[1] - b2MinXYZ[1],	// top face normal / +y
			b2MaxXYZ[2] - b1MinXYZ[2],	// far face normal / -z
			b1MaxXYZ[2] - b2MinXYZ[2],	// near face normal / +z
		};

		// face normals
		Vector3D faceNormals[] = {
			new Vector3D(-1f, -0f, 0f),
			new Vector3D(+1f, -0f, 0f),
			new Vector3D(0f, -1f, 0f),
			new Vector3D(0f, +1f, 0f),
			new Vector3D(0f, 0f, -1f),
			new Vector3D(0f, 0f, +1f)
		};

		// find minimum distance
		float distance = faceDistances[0];
		Vector3D normal = faceNormals[0];
		for (int i = 1; i < 6; i++) {
			// check for better distance hit
			if (i == 0 || faceDistances[i] < distance) {
				distance = faceDistances[i];
				normal = faceNormals[i];
			}
		}

		//
		return -distance < 0f?new CollisionResponse(-distance, normal):null;
	}

	/**
	 * Checks if axis aligned bounding box collides with sphere
	 * @param axis aligned bounding box
	 * @param sphere
	 * @return collision response from axis aligned bounding box -> sphere or null
	 */
	public static CollisionResponse doCollide(BoundingBox aabb, Sphere sphere) {
		Vector3D sphereCenter = sphere.getCenter();
		Vector3D sphereClosestPointOnAABB = aabb.computeClosestPointOnBoundingVolume(sphere.getCenter());
		Vector3D axis = sphereCenter.clone().sub(sphereClosestPointOnAABB);
		float distance = axis.computeLength() - sphere.getRadius();
		return
			distance < 0f?
			new CollisionResponse(distance, axis.normalize()):
			null;
	}

	/**
	 * Checks if sphere is colliding with axis aligned bounding box
	 * @param axis aligned bounding box
	 * @param sphere
	 * @return collision response from sphere -> axis aligned bounding box or null
	 */
	public static CollisionResponse doCollide(Sphere sphere, BoundingBox aabb) {
		Vector3D sphereCenter = sphere.getCenter();
		Vector3D sphereClosestPointOnOBB = aabb.computeClosestPointOnBoundingVolume(sphere.getCenter());
		Vector3D axis = sphereClosestPointOnOBB.clone().sub(sphereCenter);
		float distance = axis.computeLength() - sphere.getRadius();
		return
			distance < 0f?
			new CollisionResponse(distance, axis.normalize()):
			null;
	}

	/**
	 * Checks if axis aligned bounding box collides with capsule
	 * @param bounding box
	 * @param capsule
	 * @return collision response from bounding box to capsule or null
	 */
	public static CollisionResponse doCollide(BoundingBox aabb, Capsule capsule) {
		float r = capsule.getRadius();

		// clone aabb into aabb extended
		BoundingBox aabbExtended = (BoundingBox)aabb.clone();

		// extend min and mix with capsule
		aabbExtended.getMin().add(
			new Vector3D(
				-r,
				-r,
				-r
			)
		);
		aabbExtended.getMax().add(
			new Vector3D(
				+r,
				+r,
				+r
			)
		);

		// test collision with capsule extended aabb and capsule line segment 
		Vector3D contact = new Vector3D();
		if (LineSegment.doBoundingBoxCollideWithSegment(
			aabbExtended,
			capsule.getA(),
			capsule.getB(),
			contact
		) == true) {
			// TODO: handle edge cases
			// yep, just do the normal aabb->sphere collision test
			CollisionResponse collision = doCollide(
				aabb,
				new Sphere(
					contact,
					r
				)
			);
			return collision;
		}

		// no collision
		return null;
	}

	/**
	 * Checks if capsule collides with axis aligned bounding box
	 * @param axis aligned bounding box
	 * @param capsule
	 * @return collision response of capsule -> axis aligned bounding box or null
	 */
	public static CollisionResponse doCollide(Capsule c, BoundingBox aabb) {
		CollisionResponse collision = doCollide(aabb, c);
		if (collision != null) {
			collision.getNormal().scale(-1f);
		}
		return collision;
	}

	/**
	 * Checks if 2 spheres are colliding 
	 * @param sphere 1
	 * @param sphere 2
	 * @return collision response of sphere -> sphere or null
	 */
	public static CollisionResponse doCollide(Sphere s1, Sphere s2) {
		Vector3D axis = s2.getCenter().clone().sub(s1.getCenter());
		float distance = axis.computeLength();
		float _distance = distance - (s1.getRadius() + s2.getRadius());
		return
			_distance < 0f?
			new CollisionResponse(
				_distance,
				axis.clone().normalize()
			):
			null;
	}

	/**
	 * Checks if 2 capsules do collide
	 * @param capsule 1
	 * @param capsule 2
	 * @return collision response of capsule -> capsule or null
	 */
	public static CollisionResponse doCollide(Capsule c1, Capsule c2) {
		// compute closest points on capsules to each other
		Vector3D closestPointOnCapsule1 = new Vector3D();
		Vector3D closestPointOnCapsule2 = new Vector3D();
		LineSegment.computeClosestPointsOnLineSegments(
			c1.getA(),
			c1.getB(),
			c2.getA(),
			c2.getB(),
			closestPointOnCapsule1,
			closestPointOnCapsule2
		);

		// do the sphere collision tests
		return
			doCollide(
				new Sphere(closestPointOnCapsule1, c1.getRadius()),
				new Sphere(closestPointOnCapsule2, c2.getRadius())
			);
	}

	/**
	 * Checks if capsule is colliding with sphere
	 * @param capsule
	 * @param sphere
	 * @return collision response from capsule to sphere or null
	 */
	public static CollisionResponse doCollide(Capsule c, Sphere s) {
		// compute closest points from C1 to C2.A and C2.B
		Vector3D cpCvsS = c.computeClosestPointOnBoundingVolume(s.getCenter());
		Vector3D axis = s.getCenter().clone().sub(cpCvsS);
		float distance = axis.computeLength();
		float _distance = distance - (c.getRadius() + s.getRadius());
		return
			_distance < 0f?
			new CollisionResponse(
				_distance,
				axis.clone().normalize()
			):
			null;
	}

	/**
	 * Checks if sphere is colliding with capsule
	 * @param sphere
	 * @param capsule
	 * @return collision response from capsule -> sphere or null
	 */
	public static CollisionResponse doCollide(Sphere s, Capsule c) {
		// compute closest points from C1 to C2.A and C2.B
		Vector3D cpCvsS = c.computeClosestPointOnBoundingVolume(s.getCenter());
		Vector3D axis = cpCvsS.clone().sub(s.getCenter());
		float distance = axis.computeLength();
		float _distance = distance - (c.getRadius() + s.getRadius());
		return
			_distance < 0f?
			new CollisionResponse(
				_distance,
				axis.clone().normalize()
			):
			null;
	}

	/**
	 * Checks if oriented bounding box 1 collides with oriented bounding box 2
	 * 	ported from "game physics - a practical introduction/ben kenwright"
	 * @param oriented bounding box 1
	 * @param oriented bounding box 2
	 * @return collision response oriented bounding box 1 -> oriented bounding box 2 or null
	 */
	public static CollisionResponse doCollide(OrientedBoundingBox obb1, OrientedBoundingBox obb2) {
		Vector3D[] axes1 = obb1.getAxes();
		Vector3D[] axes2 = obb2.getAxes();
		Vector3D[] axes = new Vector3D[15];
		float[] penetrations = new float[15];
		for (int i = 0; i < axes.length; i++) axes[i] = new Vector3D();

		// compute obb 1,2 vertices
		Vector3D[] obb1Vertices = obb1.computeVertices();
		Vector3D[] obb2Vertices = obb2.computeVertices();

		penetrations[0] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, axes1[0], axes[0]);
		if (penetrations[0] < 0f) return null;

		penetrations[1] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, axes1[1], axes[1]);
		if (penetrations[1] < 0f) return null;

		penetrations[2] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, axes1[2], axes[2]);
		if (penetrations[2] < 0f) return null;

		penetrations[3] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, axes2[0], axes[3]);
		if (penetrations[3] < 0f) return null;

		penetrations[4] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, axes2[1], axes[4]);
		if (penetrations[4] < 0f) return null;

		penetrations[5] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, axes2[2], axes[5]);
		if (penetrations[5] < 0f) return null;

		penetrations[6] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, Vector3D.computeCrossProduct(axes1[0], axes2[0]), axes[6]);
		if (penetrations[6] < 0f) return null;

		penetrations[7] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, Vector3D.computeCrossProduct(axes1[0], axes2[1]), axes[7]);
		if (penetrations[7] < 0f) return null;

		penetrations[8] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, Vector3D.computeCrossProduct(axes1[0], axes2[2]), axes[8]);
		if (penetrations[8] < 0f) return null;

		penetrations[9] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, Vector3D.computeCrossProduct(axes1[1], axes2[0]), axes[9]);
		if (penetrations[9] < 0f) return null;

		penetrations[10] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, Vector3D.computeCrossProduct(axes1[1], axes2[1]), axes[10]);
		if (penetrations[10] < 0f) return null;

		penetrations[11] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, Vector3D.computeCrossProduct(axes1[1], axes2[2]), axes[11]);
		if (penetrations[11] < 0f) return null;

		penetrations[12] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, Vector3D.computeCrossProduct(axes1[2], axes2[0]), axes[12]);
		if (penetrations[12] < 0f) return null;

		penetrations[13] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, Vector3D.computeCrossProduct(axes1[2], axes2[1]), axes[13]);
		if (penetrations[13] < 0f) return null;

		penetrations[14] = SeparatingAxisTheorem.doSpanIntersect(obb1Vertices, obb2Vertices, Vector3D.computeCrossProduct(axes1[2], axes2[2]), axes[14]);
		if (penetrations[14] < 0f) return null;

		// determine axis with least penetration
		Vector3D normal = axes[0];
		float penetration = penetrations[0];
		for (int i = 1; i < penetrations.length; i++) {
			if (penetrations[i] < penetration) {
				normal = axes[i];
				penetration = penetrations[i];
			}
		}

		// we have a collision, return collision response
		return new CollisionResponse(-penetration, normal);
	}

	/**
	 * Checks if oriented bounding box collides with axis aligned bounding box
	 * @param oriented bounding box
	 * @param axis aligned bounding box
	 * @return collision response oriented bounding box -> axis aligned bounding box or null
	 */
	public static CollisionResponse doCollide(OrientedBoundingBox obb, BoundingBox aabb) {
		// transform aabb to obb2
		OrientedBoundingBox obb2 = new OrientedBoundingBox();
		obb2.fromBoundingBox(aabb);

		// use obb->obb collision detection
		return doCollide(obb, obb2);
	}

	/**
	 * Checks if axis aligned bounding box collides with oriented bounding box
	 * @param axis aligned bounding box
	 * @param oriented bounding box
	 * @return collision response axis aligned bounding box -> oriented bounding box or null
	 */
	public static CollisionResponse doCollide(BoundingBox aabb, OrientedBoundingBox obb) {
		// transform aabb to obb1
		OrientedBoundingBox obb1 = new OrientedBoundingBox();
		obb1.fromBoundingBox(aabb);

		// use obb->obb collision detection
		return doCollide(obb1, obb);
	}

	/**
	 * Checks if oriented bounding box collides with sphere
	 * @param oriented bounding box
	 * @param sphere
	 * @return collision response oriented bounding box -> sphere or null
	 */
	public static CollisionResponse doCollide(OrientedBoundingBox obb, Sphere sphere) {
		Vector3D sphereCenter = sphere.getCenter();
		Vector3D sphereClosestPointOnOBB = obb.computeClosestPointOnBoundingVolume(sphere.getCenter());
		Vector3D axis = sphereCenter.clone().sub(sphereClosestPointOnOBB);
		float distance = axis.computeLength() - sphere.getRadius();
		return
			distance < 0f?
			new CollisionResponse(distance, axis.normalize()):
			null;
	}

	/**
	 * Checks if sphere collides with oriented bounding box
	 * @param sphere
	 * @param oriented bounding box
	 * @return collision response sphere -> oriented bounding box or null
	 */
	public static CollisionResponse doCollide(Sphere sphere, OrientedBoundingBox obb) {
		Vector3D sphereCenter = sphere.getCenter();
		Vector3D sphereClosestPointOnOBB = obb.computeClosestPointOnBoundingVolume(sphere.getCenter());
		Vector3D axis = sphereClosestPointOnOBB.clone().sub(sphereCenter);
		float distance = axis.computeLength() - sphere.getRadius();
		return
			distance < 0f?
			new CollisionResponse(distance, axis.normalize()):
			null;
	}

	/**
	 * Checks if oriented bounding box collides with capsule
	 * @param oriented bounding box
	 * @param capsule
	 * @return collision response oriented bounding box -> capsule or null
	 */
	public static CollisionResponse doCollide(OrientedBoundingBox obb, Capsule capsule) {
		float r = capsule.getRadius();

		// clone obb into obb extended
		OrientedBoundingBox obbExtended = (OrientedBoundingBox)obb.clone();

		// extend with capsule
		obbExtended.getHalfExtension().add(new Vector3D(r,r,r));

		// test collision with capsule extended obb and capsule line segment 
		Vector3D contact = new Vector3D();
		if (LineSegment.doOrientedBoundingBoxCollideWithSegment(
			obbExtended,
			capsule.getA(),
			capsule.getB(),
			contact
		) == true) {
			// TODO: handle edge cases
			// yep, just do the normal obb->sphere collision test
			CollisionResponse collision = doCollide(
				obb,
				new Sphere(
					contact,
					capsule.getRadius()
				)
			);
			return collision;
		}

		// no collision
		return null;
	}

	/**
	 * Checks if capsule collides with oriented bounding box
	 * @param capsule
	 * @param oriented bounding box
	 * @return collision response capsule -> oriented bounding box or null 
	 */
	public static CollisionResponse doCollide(Capsule capsule, OrientedBoundingBox obb) {
		CollisionResponse collision = doCollide(obb, capsule);
		if (collision != null) {
			collision.getNormal().scale(-1f);
		}
		return collision;
	}

	/**
	 * Checks if triangle collides with sphere
	 * @param triangle
	 * @param sphere
	 * @return collision response triangle -> sphere or null
	 */
	public static CollisionResponse doCollide(Triangle triangle, Sphere sphere) {
		Vector3D sphereCenter = sphere.getCenter();
		Vector3D sphereClosestPointOnTriangle = triangle.computeClosestPointOnBoundingVolume(sphere.getCenter());
		Vector3D axis = sphereCenter.clone().sub(sphereClosestPointOnTriangle);
		float distance = axis.computeLength() - sphere.getRadius();
		return
			distance < 0f?
			new CollisionResponse(distance, axis.normalize()):
			null;
	}

	/**
	 * Checks if sphere collides with triangle
	 * @param sphere
	 * @param triangle
	 * @return collision response sphere -> triangle or null
	 */
	public static CollisionResponse doCollide(Sphere sphere, Triangle triangle) {
		Vector3D sphereCenter = sphere.getCenter();
		Vector3D sphereClosestPointOnTriangle = triangle.computeClosestPointOnBoundingVolume(sphere.getCenter());
		Vector3D axis = sphereClosestPointOnTriangle.clone().sub(sphereCenter);
		float distance = axis.computeLength() - sphere.getRadius();
		return
			distance < 0f?
			new CollisionResponse(distance, axis.normalize()):
			null;
	}

	/**
	 * Check if triangle collides with bounding box
	 * @param triangle
	 * @param bounding box
	 * @return collision response or null
	 */
	public static CollisionResponse doCollide(Triangle triangle, BoundingBox bb) {
		OrientedBoundingBox obb = new OrientedBoundingBox();
		obb.fromBoundingBox(bb);
		return doCollide(triangle, obb);
	}

	/**
	 * Check if bounding box collides with triangle
	 * @param bounding box
	 * @param triangle
	 * @return collision response or null
	 */
	public static CollisionResponse doCollide(BoundingBox bb, Triangle triangle) {
		OrientedBoundingBox obb = new OrientedBoundingBox();
		obb.fromBoundingBox(bb);
		return doCollide(obb, triangle);
	}

	/**
	 * Check if triangle collides with oriented bounding box
	 * @param triangle
	 * @param oriented bounding box
	 * @return collision response or null
	 */
	public static CollisionResponse doCollide(Triangle triangle, OrientedBoundingBox obb) {
		Vector3D[] obbVertices = obb.computeVertices();
		Vector3D[] obbAxes = obb.getAxes();

		Vector3D[] triangleVertices = triangle.getVertices();
		Vector3D triangleEdge1 = triangleVertices[1].clone().sub(triangleVertices[0]);
		Vector3D triangleEdge2 = triangleVertices[2].clone().sub(triangleVertices[1]);
		Vector3D triangleEdge3 = triangleVertices[0].clone().sub(triangleVertices[2]);
		Vector3D triangleNormal = Vector3D.computeCrossProduct(triangleEdge1, triangleEdge2).normalize();
		Vector3D[] axes = new Vector3D[13];
		for (int i = 0; i < axes.length; i++) {
			axes[i] = new Vector3D();
		}

		float[] penetrations = new float[13];
		for (int i = 0; i < penetrations.length; i++) {
			penetrations[i] = 0.0f;
		}

		// compute penetrations
		penetrations[0] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, obbAxes[0], axes[0]);
		if (penetrations[0] < 0f) return null;

		penetrations[1] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, obbAxes[1], axes[1]);
		if (penetrations[1] < 0f) return null;

		penetrations[2] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, obbAxes[2], axes[2]);
		if (penetrations[2] < 0f) return null;

		penetrations[3] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, triangleNormal, axes[3]);
		if (penetrations[3] < 0f) return null;

		penetrations[4] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, Vector3D.computeCrossProduct(obbAxes[0], triangleEdge1), axes[4]);
		if (penetrations[4] < 0f) return null;

		penetrations[5] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, Vector3D.computeCrossProduct(obbAxes[0], triangleEdge2), axes[5]);
		if (penetrations[5] < 0f) return null;

		penetrations[6] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, Vector3D.computeCrossProduct(obbAxes[0], triangleEdge3), axes[6]);
		if (penetrations[6] < 0f) return null;

		penetrations[7] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, Vector3D.computeCrossProduct(obbAxes[1], triangleEdge1), axes[7]);
		if (penetrations[7] < 0f) return null;

		penetrations[8] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, Vector3D.computeCrossProduct(obbAxes[1], triangleEdge2), axes[8]);
		if (penetrations[8] < 0f) return null;

		penetrations[9] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, Vector3D.computeCrossProduct(obbAxes[1], triangleEdge3), axes[9]);
		if (penetrations[9] < 0f) return null;

		penetrations[10] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, Vector3D.computeCrossProduct(obbAxes[2], triangleEdge1), axes[10]);
		if (penetrations[10] < 0f) return null;

		penetrations[11] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, Vector3D.computeCrossProduct(obbAxes[2], triangleEdge2), axes[11]);
		if (penetrations[11] < 0f) return null;

		penetrations[12] = SeparatingAxisTheorem.doSpanIntersect(obbVertices, triangleVertices, Vector3D.computeCrossProduct(obbAxes[2], triangleEdge3), axes[12]);
		if (penetrations[12] < 0f) return null;

		// determine axis with least penetration
		Vector3D normal = axes[0];
		float penetration = penetrations[0];
		for (int i = 1; i < penetrations.length; i++) {
			if (penetrations[i] < penetration) {
				normal = axes[i];
				penetration = penetrations[i];
			}
		}

		// we have a collision, return collision response
		return new CollisionResponse(-penetration, normal);
	}

	/**
	 * Check if oriented bounding collides with triangle box
	 * @param triangle
	 * @param oriented bounding box
	 * @return collision response or null
	 */
	public static CollisionResponse doCollide(OrientedBoundingBox obb, Triangle triangle) {
		CollisionResponse collision = doCollide(triangle, obb);
		if (collision != null) {
			collision.getNormal().scale(-1f);
		}
		return collision;
	}

	/**
	 * Check if triangle collides with another triangle
	 * @param triangle 1
	 * @param triangle 2
	 * @return collision response or null
	 */
	public static CollisionResponse doCollide(Triangle triangle1, Triangle triangle2) {
		Vector3D[] triangle1Vertices = triangle1.getVertices();
		Vector3D triangle1Edge1 = triangle1Vertices[1].clone().sub(triangle1Vertices[0]);
		Vector3D triangle1Edge2 = triangle1Vertices[2].clone().sub(triangle1Vertices[1]);
		Vector3D triangle1Edge3 = triangle1Vertices[0].clone().sub(triangle1Vertices[2]);
		Vector3D triangle1Normal = Vector3D.computeCrossProduct(triangle1Edge1, triangle1Edge2).normalize();

		Vector3D[] triangle2Vertices = triangle2.getVertices();
		Vector3D triangle2Edge1 = triangle2Vertices[1].clone().sub(triangle2Vertices[0]);
		Vector3D triangle2Edge2 = triangle2Vertices[2].clone().sub(triangle2Vertices[1]);
		Vector3D triangle2Edge3 = triangle2Vertices[0].clone().sub(triangle2Vertices[2]);
		Vector3D triangle2Normal = Vector3D.computeCrossProduct(triangle2Edge1, triangle2Edge2).normalize();

		Vector3D[] axes = new Vector3D[11];
		for (int i = 0; i < axes.length; i++) {
			axes[i] = new Vector3D();
		}

		float[] penetrations = new float[11];
		for (int i = 0; i < penetrations.length; i++) {
			penetrations[i] = 0.0f;
		}

		// compute penetrations
		penetrations[0] = SeparatingAxisTheorem.doSpanIntersect(triangle1Vertices, triangle2Vertices, triangle1Normal, axes[0]);
		if (penetrations[0] < 0f) return null;

		penetrations[1] = SeparatingAxisTheorem.doSpanIntersect(triangle1Vertices, triangle2Vertices, triangle2Normal, axes[1]);
		if (penetrations[1] < 0f) return null;

		penetrations[2] = SeparatingAxisTheorem.doSpanIntersect(triangle1Vertices, triangle2Vertices, Vector3D.computeCrossProduct(triangle1Edge1, triangle2Edge1), axes[2]);
		if (penetrations[2] < 0f) return null;

		penetrations[3] = SeparatingAxisTheorem.doSpanIntersect(triangle1Vertices, triangle2Vertices, Vector3D.computeCrossProduct(triangle1Edge1, triangle2Edge2), axes[3]);
		if (penetrations[3] < 0f) return null;

		penetrations[4] = SeparatingAxisTheorem.doSpanIntersect(triangle1Vertices, triangle2Vertices, Vector3D.computeCrossProduct(triangle1Edge1, triangle2Edge3), axes[4]);
		if (penetrations[4] < 0f) return null;

		penetrations[5] = SeparatingAxisTheorem.doSpanIntersect(triangle1Vertices, triangle2Vertices, Vector3D.computeCrossProduct(triangle1Edge2, triangle2Edge1), axes[5]);
		if (penetrations[5] < 0f) return null;

		penetrations[6] = SeparatingAxisTheorem.doSpanIntersect(triangle1Vertices, triangle2Vertices, Vector3D.computeCrossProduct(triangle1Edge2, triangle2Edge2), axes[6]);
		if (penetrations[6] < 0f) return null;

		penetrations[7] = SeparatingAxisTheorem.doSpanIntersect(triangle1Vertices, triangle2Vertices, Vector3D.computeCrossProduct(triangle1Edge2, triangle2Edge3), axes[7]);
		if (penetrations[7] < 0f) return null;

		penetrations[8] = SeparatingAxisTheorem.doSpanIntersect(triangle1Vertices, triangle2Vertices, Vector3D.computeCrossProduct(triangle1Edge3, triangle2Edge1), axes[8]);
		if (penetrations[8] < 0f) return null;

		penetrations[9] = SeparatingAxisTheorem.doSpanIntersect(triangle1Vertices, triangle2Vertices, Vector3D.computeCrossProduct(triangle1Edge3, triangle2Edge2), axes[9]);
		if (penetrations[9] < 0f) return null;

		penetrations[10] = SeparatingAxisTheorem.doSpanIntersect(triangle1Vertices, triangle2Vertices, Vector3D.computeCrossProduct(triangle1Edge3, triangle2Edge3), axes[10]);
		if (penetrations[10] < 0f) return null;

		// determine axis with least penetration
		Vector3D normal = axes[0];
		float penetration = penetrations[0];
		for (int i = 1; i < penetrations.length; i++) {
			if (penetrations[i] < penetration) {
				normal = axes[i];
				penetration = penetrations[i];
			}
		}

		// we have a collision, return collision response
		return new CollisionResponse(-penetration, normal);
	}

	/**
	 * Check if triangle collides with capsule
	 * @param triangle
	 * @param capsule
	 * @return collision response or null
	 */
	public static CollisionResponse doCollide(Triangle triangle, Capsule capsule) {
		Vector3D[] triangleVertices = triangle.getVertices();

		// init closest points
		Vector3D closestPointsOnCapsuleSegment[] = new Vector3D[3];
		Vector3D closestPointsOnTriangleSegments[] = new Vector3D[3];
		for (int i = 0; i < 3; i++) {
			closestPointsOnCapsuleSegment[i] = new Vector3D();
			closestPointsOnTriangleSegments[i] = new Vector3D();
		}

		// check capsule segment against triangle segment 0
		LineSegment.computeClosestPointsOnLineSegments(
			capsule.getA(), capsule.getB(),
			triangleVertices[1], triangleVertices[0],
			closestPointsOnCapsuleSegment[0], closestPointsOnTriangleSegments[0]
		);

		// check capsule segment against triangle segment 1
		LineSegment.computeClosestPointsOnLineSegments(
			capsule.getA(), capsule.getB(),
			triangleVertices[2], triangleVertices[1],
			closestPointsOnCapsuleSegment[1], closestPointsOnTriangleSegments[1]
		);

		// check capsule segment against triangle segment 2
		LineSegment.computeClosestPointsOnLineSegments(
			capsule.getA(), capsule.getB(),
			triangleVertices[0], triangleVertices[2],
			closestPointsOnCapsuleSegment[2], closestPointsOnTriangleSegments[2]
		);

		// determine capsule and triangle segment N with minimum distance
		//	will be the best fit to do triangle <-> sphere collision test
		float bestFitLength =
			closestPointsOnCapsuleSegment[0].clone().sub(
				closestPointsOnTriangleSegments[0]
			).computeLength();
		int bestFitTest = 0;
		for (int i = 1; i < 3; i++) {			
			float testLength =
				closestPointsOnCapsuleSegment[1].clone().sub(
					closestPointsOnTriangleSegments[1]
				).computeLength();
			if (testLength < bestFitLength) {
				bestFitLength = testLength;
				bestFitTest = i;
			}
		}

		// do the triangle sphere collision test
		return CollisionDetection.doCollide(
			triangle,
			new Sphere(
				closestPointsOnCapsuleSegment[bestFitTest],
				capsule.getRadius()
			)
		);
	}

	/**
	 * Check if capsule collides with triangle
	 * @param capsule
	 * @param triangle
	 * @return collision response or null
	 */
	public static CollisionResponse doCollide(Capsule capsule, Triangle triangle) {
		CollisionResponse collision = doCollide(triangle, capsule);
		if (collision != null) {
			collision.getNormal().scale(-1f);
		}
		return collision;
	}

}