package net.drewke.tdme.primitives;

import net.drewke.tdme.engine.Transformations;
import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.physics.CollisionDetection;
import net.drewke.tdme.physics.CollisionResponse;

/**
 * Axis Aligned Bounding Box
 * @author Andreas Drewke
 * @version $Id: f5e1eb2984331c0eed653546b96568b9be790591 $
 */
public class BoundingBox implements BoundingVolume {

	protected Vector3D min;
	protected Vector3D max;

	/**
	 * Public default constructor
	 */
	public BoundingBox() {
		this.min = new Vector3D(0f,0f,0f);
		this.max = new Vector3D(0f,0f,0f);
	}

	/**
	 * Public constructor
	 * @param min vector
	 * @param max vector
	 */
	public BoundingBox(Vector3D min, Vector3D max) {
		this.min = min;
		this.max = max;
	}

	/**
	 * Public constructor
	 * @param minX
	 * @param minY
	 * @param minZ
	 * @param maxX
	 * @param maxY
	 * @param maxZ
	 */
	public BoundingBox(
		float minX, float minY, float minZ,
		float maxX, float maxY, float maxZ) {
		// bounds
		this.min = new Vector3D(minX, minY, minZ);
		this.max = new Vector3D(maxX, maxY, maxZ);
	}

	/**
	 * Set bounds
	 * @param minX
	 * @param minY
	 * @param minZ
	 * @param maxX
	 * @param maxY
	 * @param maxZ
	 */
	public void set(
		float minX, float minY, float minZ,
		float maxX, float maxY, float maxZ) {
		// bounds
		this.min.set(minX, minY, minZ);
		this.max.set(maxX, maxY, maxZ);
	}

	/**
	 * @return min x,y,z vertex
	 */
	public Vector3D getMin() {
		return min;
	}

	/**
	 * @return max x,y,z vertex
	 */
	public Vector3D getMax() {
		return max;
	}

	/**
	 * Extend this bounding box with given bounding box
	 * @param boundingBox
	 */
	public void extend(BoundingBox boundingBox) {
		float[] minThis = min.getArray();
		float[] maxThis = max.getArray();
		float[] minOther = boundingBox.min.getArray();
		float[] maxOther = boundingBox.max.getArray();
		if (minOther[0] < minThis[0]) min.setX(minOther[0]);
		if (minOther[1] < minThis[1]) min.setY(minOther[1]);
		if (minOther[2] < minThis[2]) min.setZ(minOther[2]);
		if (maxOther[0] > maxThis[0]) max.setX(maxOther[0]);
		if (maxOther[1] > maxThis[1]) max.setY(maxOther[1]);
		if (maxOther[2] > maxThis[2]) max.setZ(maxOther[2]);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#fromBoundingVolumeWithTransformations(net.drewke.tdme.primitives.BoundingVolume, net.drewke.tdme.engine.Transformations)
	 */
	public void fromBoundingVolumeWithTransformations(BoundingVolume original, Transformations transformations) {
		// check for same type of original
		if (original instanceof BoundingBox == false) {
			System.out.println("BoundingBox::fromBoundingVolumeWithTransformations(): original is not of same type");
			return;
		}

		//
		BoundingBox boundingBox = (BoundingBox)original;

		//
		Matrix3D transformationsMatrix = transformations.getMatrix();
		float minXYZ[] = boundingBox.getMin().getArray();
		float maxXYZ[] = boundingBox.getMax().getArray();
		Vector3D vertices[] = new Vector3D[8];

		// left, top, near
		vertices[0] = new Vector3D(minXYZ[0], minXYZ[1], minXYZ[2]);
		// left, bottom, near
		vertices[1] = new Vector3D(minXYZ[0], maxXYZ[1], minXYZ[2]);
		// right, top, near
		vertices[2] = new Vector3D(maxXYZ[0], minXYZ[1], minXYZ[2]);
		// right, bottom, near
		vertices[3] = new Vector3D(maxXYZ[0], maxXYZ[1], minXYZ[2]);
		// left, top, far
		vertices[4] = new Vector3D(minXYZ[0], minXYZ[1], maxXYZ[2]);
		// left, bottom, far
		vertices[5] = new Vector3D(minXYZ[0], maxXYZ[1], maxXYZ[2]);
		// right, top, far
		vertices[6] = new Vector3D(maxXYZ[0], minXYZ[1], maxXYZ[2]);
		// right, bottom, far
		vertices[7] = new Vector3D(maxXYZ[0], maxXYZ[1], maxXYZ[2]);

		// apply transformations
		for(Vector3D vertex: vertices) {
			vertex.set(transformationsMatrix.multiply(vertex));
		}

		// determine axis aligned bounding box constraints
		float[] vertexXYZ = vertices[0].getArray(); 
		float minX = vertexXYZ[0], minY = vertexXYZ[1], minZ = vertexXYZ[2];
		float maxX = vertexXYZ[0], maxY = vertexXYZ[1], maxZ = vertexXYZ[2];
		for(int vertexIndex = 1; vertexIndex < vertices.length; vertexIndex++) {
			Vector3D vertex = vertices[vertexIndex];
			vertexXYZ = vertex.getArray(); 
			if (vertexXYZ[0] < minX) minX = vertexXYZ[0];
			if (vertexXYZ[1] < minY) minY = vertexXYZ[1];
			if (vertexXYZ[2] < minZ) minZ = vertexXYZ[2];
			if (vertexXYZ[0] > maxX) maxX = vertexXYZ[0];
			if (vertexXYZ[1] > maxY) maxY = vertexXYZ[1];
			if (vertexXYZ[2] > maxZ) maxZ = vertexXYZ[2];				
		}

		min.set(minX, minY, minZ);
		max.set(maxX, maxY, maxZ);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#computeClosestPointOnBoundingVolume(net.drewke.tdme.math.Vector3D)
	 */
	public Vector3D computeClosestPointOnBoundingVolume(Vector3D point) {
		float pointXYZ[] = point.getArray();
		float minXYZ[] = min.getArray();
		float maxXYZ[] = max.getArray();
		float closestX = pointXYZ[0] < minXYZ[0]?minXYZ[0]:pointXYZ[0] > maxXYZ[0]?maxXYZ[0]:pointXYZ[0];
		float closestY = pointXYZ[1] < minXYZ[1]?minXYZ[1]:pointXYZ[1] > maxXYZ[1]?maxXYZ[1]:pointXYZ[1];
		float closestZ = pointXYZ[2] < minXYZ[2]?minXYZ[2]:pointXYZ[2] > maxXYZ[2]?maxXYZ[2]:pointXYZ[2];
		return
			new Vector3D(
				closestX,
				closestY,
				closestZ
			);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#containsPoint(net.drewke.tdme.math.Vector3D)
	 */
	public boolean containsPoint(Vector3D point) {
		float pointXYZ[] = point.getArray();
		float minXYZ[] = min.getArray();
		float maxXYZ[] = max.getArray();
		for (int i = 0; i < 3; i++) {
			if (pointXYZ[i] < minXYZ[i]) return false;
			if (pointXYZ[i] > maxXYZ[i]) return false;
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#doesCollideWith(net.drewke.tdme.primitives.BoundingVolume)
	 */
	public CollisionResponse doesCollideWith(BoundingVolume bv2) {
		if (bv2 instanceof BoundingBox) {
			return CollisionDetection.doCollide(this, (BoundingBox)bv2);
		} else
		if (bv2 instanceof OrientedBoundingBox) {
			return CollisionDetection.doCollide(this, (OrientedBoundingBox)bv2);
		} else
		if (bv2 instanceof Sphere) {
			return CollisionDetection.doCollide(this, (Sphere)bv2);
		} else
		if (bv2 instanceof Capsule) {
			return CollisionDetection.doCollide(this, (Capsule)bv2);
		} else
		if (bv2 instanceof Triangle) {
			return CollisionDetection.doCollide(this, (Triangle)bv2);
		} else {
			System.out.println("BoundingBox::doesCollideWith(): unsupported bounding volume 2");
			return null;
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public BoundingVolume clone() {
		return new BoundingBox(min.clone(), max.clone());
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "BoundingBox [min=" + min + ", max=" + max + "]";
	}

}