package net.drewke.tdme.primitives;

import net.drewke.tdme.engine.Transformations;
import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.physics.CollisionDetection;
import net.drewke.tdme.physics.CollisionResponse;

/**
 * Capsule primitive
 * @author Andreas Drewke
 * @version $Id: e3833dcaa641753781262f17f26276f23b2d18bc $
 */
public class Capsule implements BoundingVolume {

	private Vector3D a;
	private Vector3D b;
	private float radius;

	/**
	 * Public constructor
	 */
	public Capsule() {
		a = new Vector3D(0f,0f,0f);
		b = new Vector3D(0f,0f,0f);
		radius = 1f; 
	}

	/**
	 * Public constructor
	 * @param a
	 * @param b
	 * @param radius
	 */
	public Capsule(Vector3D a, Vector3D b, float radius) {
		this.a = a;
		this.b = b;
		this.radius = radius;
	}

	/**
	 * @return radius
	 */
	public float getRadius() {
		return radius;
	}

	/**
	 * Set up radius
	 * @param radius
	 */
	public void setRadius(float radius) {
		this.radius = radius;
	}

	/**
	 * @return line segment point a
	 */
	public Vector3D getA() {
		return a;
	}

	/**
	 * @return line segment point b
	 */
	public Vector3D getB() {
		return b;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#fromBoundingVolumeWithTransformations(net.drewke.tdme.primitives.BoundingVolume, net.drewke.tdme.engine.Transformations)
	 */
	public void fromBoundingVolumeWithTransformations(BoundingVolume original, Transformations transformations) {
		// check for same type of original
		if (original instanceof Capsule == false) {
			System.out.println("Capsule::fromBoundingVolumeWithTransformations(): original is not of same type");
			return;
		}

		//
		Capsule capsule = (Capsule)original;

		//
		Matrix3D transformationsMatrix = transformations.getMatrix();
		a.set(transformationsMatrix.multiply(capsule.a));
		b.set(transformationsMatrix.multiply(capsule.b));
		radius = 
			transformationsMatrix.multiply(
				capsule.a.clone().add(
					new Vector3D(capsule.radius, 0f, 0f)
				)
			).
			sub(a).
			computeLength();
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#computeClosestPointOnBoundingVolume(net.drewke.tdme.math.Vector3D)
	 * 
	 * based on an algorithm from "game physics - a practical introduction/ben kenwright"
	 */
	public Vector3D computeClosestPointOnBoundingVolume(Vector3D point) {
		Vector3D baSub = b.clone().sub(a); 
		float baSubLength = baSub.computeLength();
		if (baSubLength > 0f) {
			baSub.normalize();
			float t = 0f;
			t = Vector3D.computeDotProduct(
				point.clone().sub(a),
				baSub
			) / baSubLength;
			if (t < 0f) t = 0f;
			if (t > 1f) t = 1f;
			return a.clone().add(baSub.scale(t * baSubLength));
		} else {
			return a.clone();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#containsPoint(net.drewke.tdme.math.Vector3D)
	 */
	public boolean containsPoint(Vector3D point) {
		Vector3D cpCvsP = computeClosestPointOnBoundingVolume(point);
		float distance = point.clone().sub(cpCvsP).computeLength();
		return distance <= radius;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#doesCollideWith(net.drewke.tdme.primitives.BoundingVolume)
	 */
	public CollisionResponse doesCollideWith(BoundingVolume bv2) {
		if (bv2 instanceof BoundingBox) {
			return CollisionDetection.doCollide(this, (BoundingBox)bv2);
		} else
		if (bv2 instanceof OrientedBoundingBox) {
			return CollisionDetection.doCollide(this, (OrientedBoundingBox)bv2);
		} else
		if (bv2 instanceof Sphere) {
			return CollisionDetection.doCollide(this, (Sphere)bv2);
		} else
		if (bv2 instanceof Capsule) {
			return CollisionDetection.doCollide(this, (Capsule)bv2);
		} else
		if (bv2 instanceof Triangle) {
			return CollisionDetection.doCollide(this, (Triangle)bv2);
		} else {
			System.out.println("Capsule::doesCollideWith(): unsupported bounding volume 2");
			return null;
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public BoundingVolume clone() {
		return new Capsule(a.clone(), b.clone(), radius);
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "Capsule [a=" + a + ", b=" + b + ", radius=" + radius + "]";
	}

}