package net.drewke.tdme.primitives;

import net.drewke.tdme.math.MathTools;
import net.drewke.tdme.math.Vector3D;

/**
 * Line segment
 * @author Andreas Drewke
 * @version $Id: 4296e5874d93340080f0e3d75077aaaca1237d98 $
 */
public class LineSegment {

	/**
	 * Computes closest points c1, c2 on line segment p1->q1, p2->q2
	 * 	based on an algorithm from "Real-Time Collision Detection" / Ericson
	 * 	Credit:
	 * 		"From Real-Time Collision Detection by Christer Ericson
	 * 		published by Morgan Kaufman Publishers, (c) 2005 Elsevier Inc"
	 * @param point p1 on line segment 1
	 * @param point q1 on line segment 1
	 * @param point p2 on line segment 2
	 * @param point q2 on line segment 2
	 * @param closest point on line segment 1 c1
	 * @param closest point on line segment 2 c2
	 */
	public static void computeClosestPointsOnLineSegments(Vector3D p1, Vector3D q1, Vector3D p2, Vector3D q2, Vector3D c1, Vector3D c2) {
		float s;
		float t;
		Vector3D d1 = q1.clone().sub(p1);
		Vector3D d2 = q2.clone().sub(p2);
		Vector3D r = p1.clone().sub(p2);
		float a = Vector3D.computeDotProduct(d1, d1);
		float e = Vector3D.computeDotProduct(d2, d2);
		float f = Vector3D.computeDotProduct(d2, r);
		// both line segments degenerate into points?
		if (a <= MathTools.EPSILON && e <= MathTools.EPSILON) {
			s = 0.0f;
			t = 0.0f;
			c1 = p1;
			c2 = p2;
			return;
		}
		// first line segment degenerates into point?
		if (a <= MathTools.EPSILON) {
			s = 0.0f;
			t = f / e;
			t = MathTools.clamp(t, 0.0f, 1.0f);
		} else {
			float c = Vector3D.computeDotProduct(d1, r);
			// second line segment degenerates into point?
			if (e <= MathTools.EPSILON) {
				t = 0.0f;
				s = MathTools.clamp(-c / a, 0.0f, 1.0f);
			} else {
				// nope, use general case
				float b = Vector3D.computeDotProduct(d1, d2);
				float denom = a * e - b * b;
				if (denom != 0.0f) {
					s = MathTools.clamp((b * f - c * e) / denom, 0.0f, 1.0f);
				} else {
					s = 0.0f;
				}
				t = (b * s + f) / e;
				if (t < 0.0f) {
					t = 0.0f;
					s = MathTools.clamp(-c / a, 0.0f, 1.0f);
				} else
				if (t > 1.0f) {
					t = 1.0f;
					s = MathTools.clamp((b - c) / a, 0.0f, 1.0f);
				}
			}
		}
		c1.set(p1).add(d1.clone().scale(s));
		c2.set(p2).add(d2.clone().scale(t));
	}

	/**
	 * Check if segment collides with bounding box
	 * 	based on an algorithm from "Real-Time Collision Detection" / Ericson
	 * 	Credit:
	 * 		"From Real-Time Collision Detection by Christer Ericson
	 * 		published by Morgan Kaufman Publishers, (c) 2005 Elsevier Inc"
	 * @param bounding box
	 * @param point p on line segment
	 * @param point q on line segment
	 * @param contact point
	 * @return true if collides or false if not
	 */
	public static boolean doBoundingBoxCollideWithSegment(BoundingBox boundingBox, Vector3D p, Vector3D q, Vector3D contact) {
		float tmin = 0.0f;
		float tmax = Float.MAX_VALUE;
		float minXYZ[] = boundingBox.min.getArray();
		float maxXYZ[] = boundingBox.max.getArray();
		Vector3D direction = q.clone().sub(p); 
		float directionXYZ[] = direction.getArray();
		float pointXYZ[] = p.getArray();
		for (int i = 0; i < 3; i++) {
			if (Math.abs(directionXYZ[i]) < MathTools.EPSILON &&
				(pointXYZ[i] < minXYZ[i] ||
				pointXYZ[i] > maxXYZ[i])) {
				return false;
			} else {
				float odd = 1.0f / directionXYZ[i];
				float t1 = (minXYZ[i] - pointXYZ[i]) * odd;
				float t2 = (maxXYZ[i] - pointXYZ[i]) * odd;
				if (t1 > t2) {
					float tmp = t1;
					t1 = t2;
					t2 = tmp;
				}
				tmin = Math.max(tmin,  t1);
				tmax = Math.min(tmax,  t2);
				if (tmin > tmax) return false;
			}
		}

		//
		if (tmin > 1.0) return false;

		// compute contact point
		contact.set(p.clone().add(direction.clone().scale(tmin)));

		// we have a collision
		return true;
	}

	/**
	 * Check if segment collides with oriented bounding box
	 * 	based on an algorithm from "Real-Time Collision Detection" / Ericson
	 * 	Credit:
	 * 		"From Real-Time Collision Detection by Christer Ericson
	 * 		published by Morgan Kaufman Publishers, (c) 2005 Elsevier Inc"
	 * @param oriented bounding box
	 * @param point p on line segment
	 * @param point q on line segment
	 * @param contact point
	 * @return true if collides or false if not
	 */
	public static boolean doOrientedBoundingBoxCollideWithSegment(OrientedBoundingBox orientedBoundingBox, Vector3D p, Vector3D q, Vector3D contact) {
		float tmin = 0.0f;
		float tmax = Float.MAX_VALUE;
		Vector3D[] obbAxes = orientedBoundingBox.axes;
		Vector3D obbCenter = orientedBoundingBox.center;
		Vector3D obbHalfExtension = orientedBoundingBox.halfExtension;
		float obbHalfExtensionXYZ[] = obbHalfExtension.getArray();
		Vector3D direction = q.clone().sub(p);
		for (int i = 0; i < 3; i++) {
			float directionLengthOnAxis = Vector3D.computeDotProduct(direction, obbAxes[i]);
			float obbExtensionLengthOnAxis = obbHalfExtensionXYZ[i];
			float obbCenterLengthOnAxis = Vector3D.computeDotProduct(obbCenter, obbAxes[i]);
			float pointLengthOnAxis = Vector3D.computeDotProduct(p, obbAxes[i]);
			if (Math.abs(directionLengthOnAxis) < MathTools.EPSILON &&
				(pointLengthOnAxis < obbCenterLengthOnAxis - obbExtensionLengthOnAxis ||
				pointLengthOnAxis > obbCenterLengthOnAxis + obbExtensionLengthOnAxis)) {
				return false;
			} else {
				float odd = 1.0f / directionLengthOnAxis;
				float t1 = (obbCenterLengthOnAxis - obbExtensionLengthOnAxis - pointLengthOnAxis) * odd;
				float t2 = (obbCenterLengthOnAxis + obbExtensionLengthOnAxis - pointLengthOnAxis) * odd;
				if (t1 > t2) {
					float tmp = t1;
					t1 = t2;
					t2 = tmp;
				}
				tmin = Math.max(tmin,  t1);
				tmax = Math.min(tmax,  t2);
				if (tmin > tmax) return false;
			}
		}

		// compute contact point
		contact.set(p.clone().add(direction.clone().scale(tmin)));

		// we have a collision
		return true;
	}

}
