package net.drewke.tdme.primitives;

import java.util.Arrays;

import net.drewke.tdme.engine.Transformations;
import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.physics.CollisionDetection;
import net.drewke.tdme.physics.CollisionResponse;

/**
 * Oriented Bounding Box
 * @author Andreas Drewke
 * @version $Id: 66fbf0e6b138e4b4763996c3ac385774fa152926 $
 */
public class OrientedBoundingBox implements BoundingVolume {

	protected Vector3D center;
	protected Vector3D[] axes;
	protected Vector3D halfExtension;

	/**
	 * Public default constructor
	 */
	public OrientedBoundingBox() {
		this.center = new Vector3D(0f,0f,0f);
		this.axes = new Vector3D[3];
		this.axes[0] = new Vector3D(1f,0f,0f);
		this.axes[1] = new Vector3D(0f,1f,0f);
		this.axes[2] = new Vector3D(0f,0f,1f);
		this.halfExtension = new Vector3D(1f,1f,1f);
	}

	/**
	 * Public constructor
	 * @param center
	 * @param axis 0
	 * @param axis 1
	 * @param axis 2
	 * @param half extension
	 */
	public OrientedBoundingBox(Vector3D center, Vector3D axis0, Vector3D axis1, Vector3D axis2, Vector3D halfExtension) {
		this.center = center;
		this.axes = new Vector3D[3];
		this.axes[0] = axis0;
		this.axes[1] = axis1;
		this.axes[2] = axis2;
		this.halfExtension = halfExtension;
	}

	/**
	 * @return center
	 */
	public Vector3D getCenter() {
		return center;
	}

	/**
	 * @return 3 axes
	 */
	public Vector3D[] getAxes() {
		return axes;
	}

	/**
	 * @return half extension
	 */
	public Vector3D getHalfExtension() {
		return halfExtension;
	}

	/**
	 * Create oriented bounding box from bounding box
	 * @param bounding box
	 * @param transformations
	 */
	public void fromBoundingBox(BoundingBox bb) {
		this.halfExtension = bb.getMax().clone().sub(bb.getMin()).scale(0.5f);
		this.center = bb.getMin().clone().add(halfExtension);
		this.axes = new Vector3D[3];
		this.axes[0] = new Vector3D(1f,0f,0f);
		this.axes[1] = new Vector3D(0f,1f,0f);
		this.axes[2] = new Vector3D(0f,0f,1f);
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#fromBoundingVolumeWithTransformations(net.drewke.tdme.primitives.BoundingVolume, net.drewke.tdme.engine.Transformations)
	 */
	public void fromBoundingVolumeWithTransformations(BoundingVolume original, Transformations transformations) {
		// check for same type of original
		if (original instanceof OrientedBoundingBox == false) {
			System.out.println("OrientedBoundingBox::fromBoundingVolumeWithTransformations(): original is not of same type");
			return;
		}

		//
		OrientedBoundingBox obb = (OrientedBoundingBox)original;

		//
		Matrix3D transformationsMatrix = transformations.getMatrix();

		// apply rotation, scale, translation
		center.set(transformationsMatrix.multiply(obb.center));

		// apply transformations rotation + scale to axis
		Vector3D axis0transformed = transformationsMatrix.multiplyNoTranslation(obb.axes[0]);
		Vector3D axis1transformed = transformationsMatrix.multiplyNoTranslation(obb.axes[1]);
		Vector3D axis2transformed = transformationsMatrix.multiplyNoTranslation(obb.axes[2]);

		// set up axes
		axes[0].set(axis0transformed).normalize();
		axes[1].set(axis1transformed).normalize();
		axes[2].set(axis2transformed).normalize();

		// apply scale to half extension
		halfExtension.set(obb.halfExtension);
		halfExtension.scale(
			new Vector3D(
				axis0transformed.computeLength(),
				axis1transformed.computeLength(),
				axis2transformed.computeLength()
			)
		);
	}

	/**
	 * Computes vertices of this oriented bounding box
	 * @return oriented bounding box vertices
	 */
	public Vector3D[] computeVertices() {
		float[] halfExtensionXYZ = halfExtension.getArray();

		// create vertices for oriented bounding box
		Vector3D[] vertices = new Vector3D[8];

		// just for my imagination
		//	near left top
		vertices[0] = center.clone();
		vertices[0].add(axes[0].clone().scale(-halfExtensionXYZ[0]));
		vertices[0].add(axes[1].clone().scale(-halfExtensionXYZ[1]));
		vertices[0].add(axes[2].clone().scale(-halfExtensionXYZ[2]));

		// just for my imagination
		//	near right top
		vertices[1] = center.clone();
		vertices[1].add(axes[0].clone().scale(+halfExtensionXYZ[0]));
		vertices[1].add(axes[1].clone().scale(-halfExtensionXYZ[1]));
		vertices[1].add(axes[2].clone().scale(-halfExtensionXYZ[2]));

		// just for my imagination
		//	near right bottom
		vertices[2] = center.clone();
		vertices[2].add(axes[0].clone().scale(+halfExtensionXYZ[0]));
		vertices[2].add(axes[1].clone().scale(+halfExtensionXYZ[1]));
		vertices[2].add(axes[2].clone().scale(-halfExtensionXYZ[2]));

		// just for my imagination
		//	near left bottom
		vertices[3] = center.clone();
		vertices[3].add(axes[0].clone().scale(-halfExtensionXYZ[0]));
		vertices[3].add(axes[1].clone().scale(+halfExtensionXYZ[1]));
		vertices[3].add(axes[2].clone().scale(-halfExtensionXYZ[2]));

		// just for my imagination
		//	far left top
		vertices[4] = center.clone();
		vertices[4].add(axes[0].clone().scale(-halfExtensionXYZ[0]));
		vertices[4].add(axes[1].clone().scale(-halfExtensionXYZ[1]));
		vertices[4].add(axes[2].clone().scale(+halfExtensionXYZ[2]));

		// just for my imagination
		//	far right top
		vertices[5] = center.clone();
		vertices[5].add(axes[0].clone().scale(+halfExtensionXYZ[0]));
		vertices[5].add(axes[1].clone().scale(-halfExtensionXYZ[1]));
		vertices[5].add(axes[2].clone().scale(+halfExtensionXYZ[2]));

		// just for my imagination
		//	far right bottom
		vertices[6] = center.clone();
		vertices[6].add(axes[0].clone().scale(+halfExtensionXYZ[0]));
		vertices[6].add(axes[1].clone().scale(+halfExtensionXYZ[1]));
		vertices[6].add(axes[2].clone().scale(+halfExtensionXYZ[2]));

		// just for my imagination
		//	far left bottom
		vertices[7] = center.clone();
		vertices[7].add(axes[0].clone().scale(-halfExtensionXYZ[0]));
		vertices[7].add(axes[1].clone().scale(+halfExtensionXYZ[1]));
		vertices[7].add(axes[2].clone().scale(+halfExtensionXYZ[2]));

		//
		return vertices;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#computeClosestPointOnBoundingVolume(net.drewke.tdme.math.Vector3D)
	 *
	 * 	based on an algorithm from "Real-Time Collision Detection" / Ericson
	 * 	Credit:
	 * 		"From Real-Time Collision Detection by Christer Ericson
	 * 		published by Morgan Kaufman Publishers, (c) 2005 Elsevier Inc"
	 */
	public Vector3D computeClosestPointOnBoundingVolume(Vector3D point) {
		Vector3D direction = point.clone().sub(center);
		Vector3D closestPoint = center.clone();
		float[] halfExtensionXYZ = halfExtension.getArray();
		for (int i = 0; i < axes.length; i++) {
			float distance = Vector3D.computeDotProduct(direction, axes[i]);
			if (distance > halfExtensionXYZ[i]) distance = halfExtensionXYZ[i];
			if (distance < -halfExtensionXYZ[i]) distance = -halfExtensionXYZ[i];
			closestPoint.add(axes[i].clone().scale(distance)); 
		}
		return closestPoint;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#containsPoint(net.drewke.tdme.math.Vector3D)
	 * 
	 * 	based on an algorithm from "Real-Time Collision Detection" / Ericson
	 * 	Credit:
	 * 		"From Real-Time Collision Detection by Christer Ericson
	 * 		published by Morgan Kaufman Publishers, (c) 2005 Elsevier Inc"
	 */
	public boolean containsPoint(Vector3D point) {
		Vector3D direction = point.clone().sub(center);
		float[] halfExtensionXYZ = halfExtension.getArray();
		for (int i = 0; i < axes.length; i++) {
			float distance = Vector3D.computeDotProduct(direction, axes[i]);
			if (distance > halfExtensionXYZ[i]) return false;
			if (distance < -halfExtensionXYZ[i]) return false;
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#doesCollideWith(net.drewke.tdme.primitives.BoundingVolume)
	 */
	public CollisionResponse doesCollideWith(BoundingVolume bv2) {
		if (bv2 instanceof BoundingBox) {
			return CollisionDetection.doCollide(this, (BoundingBox)bv2);
		} else
		if (bv2 instanceof OrientedBoundingBox) {
			return CollisionDetection.doCollide(this, (OrientedBoundingBox)bv2);
		} else
		if (bv2 instanceof Sphere) {
			return CollisionDetection.doCollide(this, (Sphere)bv2);
		} else
		if (bv2 instanceof Capsule) {
			return CollisionDetection.doCollide(this, (Capsule)bv2);
		} else
		if (bv2 instanceof Triangle) {
			return CollisionDetection.doCollide(this, (Triangle)bv2);
		} else {
			System.out.println("OrientedBoundingBox::doesCollideWith(): unsupported bounding volume 2");
			return null;
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public BoundingVolume clone() {
		return new OrientedBoundingBox(
			center.clone(),
			axes[0].clone(),
			axes[1].clone(),
			axes[2].clone(),
			halfExtension.clone()
		);
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "OrientedBoundingBox [center=" + center + ", axes="
				+ Arrays.toString(axes) + ", halfExtension=" + halfExtension
				+ "]";
	}

}