package net.drewke.tdme.primitives;

import net.drewke.tdme.engine.Transformations;
import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.physics.CollisionDetection;
import net.drewke.tdme.physics.CollisionResponse;

/**
 * Sphere
 * @author Andreas Drewke
 * @version $Id: be1477c707cd77bdbcbe4f39b5e18e00d9621bda $
 */
public class Sphere implements BoundingVolume {

	protected Vector3D center;
	protected float radius;

	/**
	 * Public default constructor
	 */
	public Sphere() {
		center = new Vector3D();
		radius = 0f;
	}

	/**
	 * Public constructor
	 * @param center
	 * @param radius
	 */
	public Sphere(Vector3D center, float radius) {
		super();
		this.center = center;
		this.radius = radius;
	}

	/**
	 * Creates an sphere from bounding box
	 * @param b
	 */
	public void fromBoundingBox(BoundingBox boundingBox) {
		float min[] = boundingBox.min.getArray();
		float max[] = boundingBox.max.getArray();
		center.set(
			(min[0] + max[0]) / 2,
			(min[1] + max[1]) / 2,
			(min[2] + max[2]) / 2
		);
		float radiusMin = boundingBox.min.clone().sub(center).computeLength();
		float radiusMax = boundingBox.max.clone().sub(center).computeLength();
		radius = radiusMin <= radiusMax?radiusMin:radiusMax;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#fromBoundingVolumeWithTransformations(net.drewke.tdme.primitives.BoundingVolume, net.drewke.tdme.engine.Transformations)
	 */
	public void fromBoundingVolumeWithTransformations(BoundingVolume original, Transformations transformations) {
		// check for same type of original
		if (original instanceof Sphere == false) {
			System.out.println("Sphere::fromBoundingVolumeWithTransformations(): original is not of same type");
			return;
		}

		//
		Sphere sphere = (Sphere)original;

		//
		Matrix3D transformationsMatrix = transformations.getMatrix();

		// apply translations
		// 	translate center
		center.set(transformationsMatrix.multiply(sphere.center));

		// compute new radius as scale could have changed
		radius = 
			transformationsMatrix.multiply(
				sphere.center.clone().add(
					new Vector3D(sphere.radius, 0f, 0f)
				)
			).
			sub(center).
			computeLength();
	}

	/**
	 * @return center
	 */
	public Vector3D getCenter() {
		return center;
	}

	/**
	 * @return float radius
	 */
	public float getRadius() {
		return radius;
	}

	/**
	 * Set up radius
	 * @param radius
	 */
	public void setRadius(float radius) {
		this.radius = radius;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#computeClosestPointOnBoundingVolume(net.drewke.tdme.math.Vector3D)
	 */
	public Vector3D computeClosestPointOnBoundingVolume(Vector3D point) {
		// TODO: implement me
		System.out.println("Sphere::computeClosestPointOnBoundingVolume(): not implemented yet.");
		return null;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#containsPoint(net.drewke.tdme.math.Vector3D)
	 */
	public boolean containsPoint(Vector3D point) {
		float distance = point.clone().sub(center).computeLength();
		return distance <= radius;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#doesCollideWith(net.drewke.tdme.primitives.BoundingVolume)
	 */
	public CollisionResponse doesCollideWith(BoundingVolume bv2) {
		if (bv2 instanceof BoundingBox) {
			return CollisionDetection.doCollide(this, (BoundingBox)bv2);
		} else
		if (bv2 instanceof OrientedBoundingBox) {
			return CollisionDetection.doCollide(this, (OrientedBoundingBox)bv2);
		} else
		if (bv2 instanceof Sphere) {
			return CollisionDetection.doCollide(this, (Sphere)bv2);
		} else
		if (bv2 instanceof Capsule) {
			return CollisionDetection.doCollide(this, (Capsule)bv2);
		} else
		if (bv2 instanceof Triangle) {
			return CollisionDetection.doCollide(this, (Triangle)bv2);
		} else {
			System.out.println("Sphere::doesCollideWith(): unsupported bounding volume 2");
			return null;
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public BoundingVolume clone() {
		return new Sphere(center.clone(), radius);
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "Sphere [center=" + center + ", radius=" + radius + "]";
	}

}
