package net.drewke.tdme.primitives;

import java.util.Arrays;

import net.drewke.tdme.engine.Transformations;
import net.drewke.tdme.math.MathTools;
import net.drewke.tdme.math.Matrix3D;
import net.drewke.tdme.math.Vector3D;
import net.drewke.tdme.model.Face;
import net.drewke.tdme.physics.CollisionDetection;
import net.drewke.tdme.physics.CollisionResponse;

/**
 * Triangle primitive
 * @author Andreas Drewke
 * @version $Id: c4578c78fbfefc74623d5b0b0d85684026f9f3ca $
 */
public class Triangle implements BoundingVolume {

	protected Vector3D[] vertices;

	/**
	 * Public default constructor
	 */
	public Triangle() {
		this.vertices = new Vector3D[3];
		for (int i = 0; i < 3; i++) {
			this.vertices[i] = new Vector3D();
		}
	}

	/**
	 * Public constructor from array of vertices
	 * @param vertices
	 */
	public Triangle(Vector3D vertex0, Vector3D vertex1, Vector3D vertex2) {
		this.vertices = new Vector3D[3];
		this.vertices[0] = vertex0;
		this.vertices[1] = vertex1;
		this.vertices[2] = vertex2;
	}

	/**
	 * @return triangle vertices
	 */
	public Vector3D[] getVertices() {
		return vertices;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#fromBoundingVolumeWithTransformations(net.drewke.tdme.primitives.BoundingVolume, net.drewke.tdme.engine.Transformations)
	 */
	public void fromBoundingVolumeWithTransformations(BoundingVolume original, Transformations transformations) {
		// check for same type of original
		if (original instanceof Triangle == false) {
			System.out.println("Triangle::fromBoundingVolumeWithTransformations(): original is not of same type");
			return;
		}

		//
		Triangle triangle = (Triangle)original;
		Matrix3D transformationsMatrix = transformations.getMatrix();
		for (int i = 0; i < 3; i++) {
			vertices[i] = transformationsMatrix.multiply(triangle.vertices[i]);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#computeClosestPointOnBoundingVolume(net.drewke.tdme.math.Vector3D)
	 * 
	 * based on http://www.gamedev.net/topic/552906-closest-point-on-triangle/
	 */
	public Vector3D computeClosestPointOnBoundingVolume(Vector3D point) {
		Vector3D edge0 = vertices[1].clone().sub(vertices[0]);
		Vector3D edge1 = vertices[2].clone().sub(vertices[0]);
		Vector3D v0 = vertices[0].clone().sub(point);

		float a = Vector3D.computeDotProduct(edge0, edge0);
		float b = Vector3D.computeDotProduct(edge0, edge1);
		float c = Vector3D.computeDotProduct(edge1, edge1);
		float d = Vector3D.computeDotProduct(edge0, v0);
		float e = Vector3D.computeDotProduct(edge1, v0);

		float det = a * c - b * b;
		float s = b * e - c * d;
		float t = b * d - a * e;

		if (s + t < det) {
			if (s < 0.0f) {
				if (t < 0.0f) {
					if (d < 0.f) {
						s = MathTools.clamp(-d / a, 0.0f, 1.0f);
						t = 0.0f;
					} else {
						s = 0.0f;
						t = MathTools.clamp(-e / c, 0.0f, 1.0f);
					}
				} else {
					s = 0.0f;
					t = MathTools.clamp(-e / c, 0.0f, 1.0f);
				}
			} else if (t < 0.0f) {
				s = MathTools.clamp(-d / a, 0.0f, 1.0f);
				t = 0.0f;
			} else {
				float invDet = 1.0f / det;
				s *= invDet;
				t *= invDet;
			}
		} else {
			if (s < 0.0f) {
				float tmp0 = b + d;
				float tmp1 = c + e;
				if (tmp1 > tmp0) {
					float numer = tmp1 - tmp0;
					float denom = a - 2 * b + c;
					s = MathTools.clamp(numer / denom, 0.0f, 1.0f);
					t = 1 - s;
				} else {
					t = MathTools.clamp(-e / c, 0.0f, 1.0f);
					s = 0.0f;
				}
			} else if (t < 0.0f) {
				if (a + d > b + e) {
					float numer = c + e - b - d;
					float denom = a - 2 * b + c;
					s = MathTools.clamp(numer / denom, 0.0f, 1.f);
					t = 1 - s;
				} else {
					s = MathTools.clamp(-e / c, 0.0f, 1.0f);
					t = 0.f;
				}
			} else {
				float numer = c + e - b - d;
				float denom = a - 2 * b + c;
				s = MathTools.clamp(numer / denom, 0.0f, 1.f);
				t = 1.0f - s;
			}
		}

		return
			vertices[0].clone(). 
			add(edge0.clone().scale(s)).
			add(edge1.clone().scale(t));
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#containsPoint(net.drewke.tdme.math.Vector3D)
	 */
	public boolean containsPoint(Vector3D point) {
		// TODO: implement me
		System.out.println("Sphere::computeClosestPointOnBoundingVolume(): not implemented yet.");
		return false;
	}

	/*
	 * (non-Javadoc)
	 * @see net.drewke.tdme.primitives.BoundingVolume#doesCollideWith(net.drewke.tdme.primitives.BoundingVolume)
	 */
	public CollisionResponse doesCollideWith(BoundingVolume bv2) {
		if (bv2 instanceof BoundingBox) {
			return CollisionDetection.doCollide(this, (BoundingBox)bv2);
		} else
		if (bv2 instanceof OrientedBoundingBox) {
			return CollisionDetection.doCollide(this, (OrientedBoundingBox)bv2);
		} else
		if (bv2 instanceof Sphere) {
			return CollisionDetection.doCollide(this, (Sphere)bv2);
		} else
		if (bv2 instanceof Capsule) {
			return CollisionDetection.doCollide(this, (Capsule)bv2);
		} else
		if (bv2 instanceof Triangle) {
			return CollisionDetection.doCollide(this, (Triangle)bv2);
		} else {
			System.out.println("Triangle::doesCollideWith(): unsupported bounding volume 2");
			return null;
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public BoundingVolume clone() {
		return new Triangle(
			vertices[0].clone(),
			vertices[1].clone(),
			vertices[2].clone()
		);
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "Triangle [vertices=" + Arrays.toString(vertices) + "]";
	}

}
